// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"
#include "annotation.h"

#include "univout.h"

// From NCBI:
#include "subutil.h" // For definition of TOPOLOGY_LINEAR, TOPOLOGY_CIRCULAR, etc.
#include "accentr.h"
#include "accutils.h"
#include "sqnutils.h"
#include "pmfapi.h"
#include "ent2api.h"
#include "explore.h"

#include <fstream>
#include <iostream>
#include <sstream>
#include <algorithm>

using namespace std;

// Local functions:
void BuildBioseqList(SeqEntryPtr sep, Pointer data, Int4 index, Int2 indent);
void FindBioseq(SeqEntryPtr sep, Pointer data, Int4 index, Int2 indent);
const char* GetAsnFileType(const string &name);

NLM_EXTERN void GetFeatListCallback(SeqEntryPtr sep, Pointer data, Int4 index, Int2 indent);
NLM_EXTERN void ApplyOffsetCallback(SeqEntryPtr sep, Pointer data, Int4 index, Int2 indent);

static void touch_bioseq(BioseqPtr m_bsp, Pointer m_userdata);
static Boolean LIBCALLBACK load_bioseq_data(
   SeqLocPtr m_slp, SeqMgrSegmentContextPtr m_context);

void add_loc_offset(SeqLocPtr anp, const Int4 &m_offset);

bool match_seq_list(list<DNA3Seq>::const_iterator m_query_start, 
	list<DNA3Seq>::const_iterator m_query_stop,
	DNA3Seq::iterator target_start,
	const DNA3Seq &target,
	const bool &m_is_circular,
	list<unsigned int> &last_base);

////////////////////////////////////////////////////////////////////////////
// Global variables for use with NCBI Callback functions
SeqIdPtr global_search_id = NULL;
list<SeqEntryPtr> global_far_sep;
bool is_fetch_enabled = false;

// Turn a range or segment string into a start/stop and segment list
// Return false on failure and true on success
void GeneAnnotation::range(const string &m_range)
{
	// Clear any existing segments
	seg_list.clear();
	
	const unsigned int len = m_range.size();
	unsigned int i = 0;
	
	// Skip to the start of the range entry.
	while( (i < len) && isspace(m_range[i]) ){
		i++;
	}
	
	if(i == len){
		throw "Range: Empty range string";
	}
	
	while(i < len){
		
		list<char> number;
		
		if( !isdigit(m_range[i]) ){
			
			seg_list.clear();
			throw "Range: Could not read start";
		}
		
		while( (i < len) && isdigit(m_range[i]) ){
		
			number.push_back(m_range[i] - '0');
			i++;
		}
		
		const int start = list_to_int(number);
		
		while( (i < len) && isspace(m_range[i]) ){
			i++;
		}

		if( (i == len) || (m_range[i] != '.') ){
		
			seg_list.clear();
			throw "Range: Could not read \"..\"";
		}
		
		i++;
		
		if( (i == len) || (m_range[i] != '.') ){
		
			seg_list.clear();
			throw "Range: Could not read \"..\"";
		}
		
		i++;
				
		while( (i < len) && isspace(m_range[i]) ){
			i++;
		}
		
		if(i == len){
		
			seg_list.clear();
			throw "Range: Could not read stop";
		}
		
		
		if( !isdigit(m_range[i]) ){
		
			seg_list.clear();
			throw "Range: Could not read stop";
		}
		
		while( (i < len) && isdigit(m_range[i]) ){
		
			number.push_back(m_range[i] - '0');
			i++;
		}
		
		const int stop = list_to_int(number);
				
		while( (i < len) && ( isspace(m_range[i]) || (m_range[i] == ',') ) ){
			i++;
		}
		
		seg_list.push_back( make_pair(start, stop) );
	}
	
	seg_list.sort();
	
	start( seg_list.front().first );
	stop( seg_list.back().second );
	
	// Only keep the segment list around if it stores more than 1 element
	if(seg_list.size() == 1){
		seg_list.clear();
	}
}

// Using the start/stop and segment list, return a valid range string
string GeneAnnotation::range() const
{
	stringstream ss;
	
	if(is_segmented() == true){
		list< pair<unsigned int, unsigned int> >::const_iterator iter = 
			seg_list.begin();
		
		while( iter != seg_list.end() ){
			ss << iter->first << ".." << iter->second;
			
			iter++;
			
			if( iter != seg_list.end() ){
				ss << ", ";
			}
		}
	}
	else{
		ss << start() << ".." << stop();
	}
	
	return ss.str();
}


void DNAMol::update_annotation(const GeneAnnotation &m_annot, 
		const list<GeneAnnotation>::const_iterator &m_iter)
{
	typedef list<GeneAnnotation>::iterator I;

	I iter;
	
	for(iter = gene_list.begin();iter != gene_list.end();iter++){
	
		if( (list<GeneAnnotation>::const_iterator)(iter) == m_iter){
			break;
		}
	}
	
	if( iter == gene_list.end() ){
		
		gene_list.push_back(m_annot);
	}
	else{
		
		*iter = m_annot;
	}
}	

bool DNAMol::delete_annotation()
{
	typedef list<GeneAnnotation>::iterator I;

	I iter;

	list<I> reaper;

	for(iter = gene_list.begin();iter != gene_list.end();iter++){
		if(iter->select() == true){
			reaper.push_back(iter);
		}
	}

	if(reaper.empty() == true){
		return false;
	}

	list<I>::iterator reaper_iter;

	for(reaper_iter = reaper.begin();reaper_iter != reaper.end();reaper_iter++){
		gene_list.erase( *reaper_iter );
	}

	// Recompute the intergenic space
	processGeneList(true);
	
	return true;
}

void DNAMol::annotate_selection(const std::list<SeqHit> &m_hit_list)
{
	list<SeqHit>::const_iterator iter;

	for(iter = m_hit_list.begin();iter != m_hit_list.end();iter++){
		stringstream ss;
		stringstream ss_note;

		GeneAnnotation user_annot;

		user_annot.type(GeneAnnotation::USER);
		user_annot.start( iter->start() );
		user_annot.stop( iter->stop() );
		user_annot.is_complement( iter->complement() );

		switch( iter->type() ){
			case SeqHit::EXACT_MATCH:
				user_annot.info(GeneAnnotation::LOCUS, "Exact match");
				user_annot.info(GeneAnnotation::PRODUCT, "Exact match to search string");
				break;
			case SeqHit::PROBE_MATCH:
				user_annot.info(GeneAnnotation::LOCUS, "Probe");

				ss << "Probe Tm is " << iter->probe_tm()
				   << "C; exact matches = " 
					   << iter->probe_match() 
					   << "/" << iter->probe_len();;

				user_annot.info(GeneAnnotation::PRODUCT, ss.str() );
				
				break;
			case SeqHit::PRIMER_MATCH:
				user_annot.info(GeneAnnotation::LOCUS, "Amplicon");
				
				ss << "Primer " << iter->primer_id_1() 
				   << " Tm is " << iter->primer_tm_1()
				   << " C, Primer " << iter->primer_id_2() 
				   << " Tm is " << iter->primer_tm_2() << " C";

				user_annot.info(GeneAnnotation::PRODUCT, ss.str() );
				
				ss_note << "Primer " << iter->primer_id_1() 
						<< " 3' exact match = " << iter->primer_clamp_1()
						<< " bases, exact matches = " 
						<< iter->primer_match_1() << "/" 
				 		<< iter->primer_len_1() 
						<< ", Primer " << iter->primer_id_2() 
						<< " 3' exact match = " << iter->primer_clamp_2()
						<< " bases, exact matches = " 
						<< iter->primer_match_2() << "/" 
				 		<< iter->primer_len_2();

				user_annot.info(GeneAnnotation::NOTE, ss_note.str() );

				break;
			case SeqHit::MOL_PRIMER_MATCH:
				// user_annot.info(GeneAnnotation::LOCUS, "MOLigo");
				user_annot.info(GeneAnnotation::LOCUS, "Padlock");
				
				ss << "Primer " << iter->primer_id_1() << " Tm is " 
				   << iter->primer_tm_1()
				   << " C, Primer " << iter->primer_id_2() 
				   << " Tm is " << iter->primer_tm_2() << " C";

				user_annot.info(GeneAnnotation::PRODUCT, ss.str() );
				
				ss_note << "Primer " << iter->primer_id_1() << " 3' exact match = " 
						<< iter->primer_clamp_1()
						<< " bases, exact matches = " 
						<< iter->primer_match_1() << "/" 
				 		<< iter->primer_len_1() << ", Primer " << iter->primer_id_2() 
						<< " 5' exact match = " << iter->primer_clamp_2()
						<< " bases, exact matches = " 
						<< iter->primer_match_2() << "/" 
				 		<< iter->primer_len_2();

				user_annot.info(GeneAnnotation::NOTE, ss_note.str() );

				break;
			default: // Unknown
				user_annot.info(GeneAnnotation::PRODUCT, "Unknown");
				break;
		};

		gene_list.push_back(user_annot);
	}
}

bool DNAMol::load(const std::string &m_filename, const unsigned int &m_type, streampos &m_pos)
{
	// Return true if there are multiple genomes to read in this file
	bool ret = false;
	
	switch(m_type){
		case ASN_1:
			ret = loadASN(m_filename, m_pos);
			break;
		case FASTA:
			// Multiple genomes are not currently supported for ASN
			loadFASTA(m_filename);
			break;
		case PTT:
			// Multiple genomes are not currently supported for ASN
			loadPTT(m_filename);
			break;
		case GBK:
			ret = loadGBK(m_filename, m_pos);
			break;
		case EMBL:
			ret = loadEMBL(m_filename, m_pos);
			break;
		default:
			throw "Unknown file format!";
	};

	return ret;
}

// Load an entry from the NCBI database
bool DNAMol::load(const int &m_gi)
{
	return loadNetASN(m_gi);
}

// Load an entry from the NCBI database and save the entry to disk as well
bool DNAMol::load_and_save(const int &m_gi, const string &m_filename)
{
	return loadNetASN(m_gi, m_filename);
}


// This function is based on the code contained in ncbi/demo/seqtest.c
// Note that the calling function is responsible for freeing the memory
// associated with "SeqEntryPtr sep".
bool DNAMol::loadASN(SeqEntryPtr sep, SeqIdPtr sid)
{
	BioseqPtr genome_bioseq = NULL;
	SeqPortPtr     spp = NULL;
	Uint1          residue;
	
	// Handle ASN.1 data with external links
	const bool fetch_state = is_fetch_enabled;
		
	global_far_sep.clear();
		
	Uint2 entityID = ObjMgrGetEntityIDForChoice(sep);
	
	SeqMgrIndexFeatures(entityID, NULL);
	
	VisitBioseqsInSep(sep, NULL, touch_bioseq);
	
	if(!fetch_state && is_fetch_enabled){
		
		// If PubSeqFetchEnable started off disabled but finished enabled,
		// then we've used the web to download additional data to supplement an
		// ASN file. Turn off Seq-fetching ...
		
		PubSeqFetchDisable();
		is_fetch_enabled = false;
	}
	
	// Assume that SeqEntryLoad has already been called!
	//if (! SeqEntryLoad()) {
	//	throw __FILE__ ":DNAMol::loadASN: SeqEntryLoad failed";
	//}

	if(NULL == sep){
		throw "loadASN got a NULL SeqEntryPtr";
	}
			
	// Get taxanomic info about this DNAMol
	taxaInfo( SeqEntryGetSeqDescr(sep, Seq_descr_source, NULL) );
	
	global_search_id = sid;	
	BioseqExplore(sep, (Pointer) &genome_bioseq, FindBioseq);
	global_search_id = NULL;
		
	// The DNA molecule of interest is the first bioseq in the file
	if(NULL == genome_bioseq){
		throw "loadASN: Unable to find a single DNA molecule!";
	}

	// There is sequence data until we determine otherwise
	bool found_sequence_data = true;
	
	// Save the SeqId
	sip = SeqIdSetDup(genome_bioseq->id);
		
	ValNodePtr chunk_ptr = (Bioseq_repr(genome_bioseq) == Seq_repr_delta) ?
		(ValNodePtr)(genome_bioseq->seq_ext) : NULL;
	
	if( (chunk_ptr != NULL) && (chunk_ptr->choice != 2) ){
		
		// Don't try to read the segmented sequence
		found_sequence_data = false;
		
		if(BioseqGetLen(genome_bioseq) == 0){
			throw "Unable to determine sequence length";
		}
	}
	else{
	
		// Save the sequence data
		spp = SeqPortNew(genome_bioseq, 0, -1, Seq_strand_plus, Seq_code_iupacna);

		if(spp == NULL){

			throw "Didn't read any sequence data. "
				"If downloading, obtain the file directly from NCBI";
		}
	}
			
	seq = DNA3Seq(BioseqGetLen(genome_bioseq));
	
	DNA3Seq::iterator iter = seq.begin();
	
	uout << "Loading sequence data ..." << uout_endl;

	// Prepare to compute GC content and base composition
	gc_content = 0.0f;
	a_content = 0.0f;
	t_content = 0.0f;
	g_content = 0.0f;
	c_content = 0.0f;
	other_content = 0.0f;
	
	if(found_sequence_data == false){
		
		// Fill the sequence record will 'N'
		while( iter != seq.end() ){
		
			iter.base(DNA3Seq::UNKNOWN);
			iter++;
		}
	}
	else{
		while ((residue = SeqPortGetResidue(spp)) != SEQPORT_EOF) {

			if (IS_residue(residue)){

				// Make sure residue is upper case for %GC calculation
				residue = toupper(residue);

				iter.c_base(residue);

				switch(iter.base()){
					case DNA3Seq::A:
						a_content ++;
						break;
					case DNA3Seq::T:
						t_content ++;
						break;
					case DNA3Seq::G:
						g_content ++;
						gc_content ++;
						break;
					case DNA3Seq::C:
						c_content ++;
						gc_content ++;
						break;
					default:
						other_content ++;
						break;
				};

			}
			else{
				switch (residue)
				{
					case SEQPORT_VIRT:
						iter.base(DNA3Seq::GAP);
						break;
					default:
						iter.base(DNA3Seq::UNKNOWN);
						break;
				}
			}

			iter++;
		}
	}
	
	gc_content /= seq.size();
	a_content /= seq.size();
	t_content /= seq.size();
	g_content /= seq.size();
	c_content /= seq.size();
	other_content /= seq.size();
	
	if(spp != NULL){
		spp = SeqPortFree(spp);
	}
	
	// Is this genome linear or circular?
	is_circular = (genome_bioseq->topology == TOPOLOGY_CIRCULAR);
	
	list<SeqFeatPtr> feat_list[SEQFEAT_MAX];
	list<SeqFeatPtr>::iterator feat_iter;
		
	// Extract all of the relevant feature pointers in a single pass
	global_search_id = sid;
	SeqEntryExplore(sep, (Pointer)(feat_list), GetFeatListCallback);
	global_search_id = NULL;
	
	if(global_far_sep.empty() == false){
		
		list<SeqEntryPtr>::const_iterator far_iter;
		
		for(far_iter = global_far_sep.begin();far_iter != global_far_sep.end();far_iter++){
			
			SeqEntryExplore(*far_iter, (Pointer)(feat_list), GetFeatListCallback);
		}
		
		// We're done saving the feature pointers in the array of feature lists
		global_far_sep.clear();
	}
	
	uout << "Parsing open reading frame records ..." << uout_endl;
	
	for(feat_iter = feat_list[SEQFEAT_CDREGION].begin();
		feat_iter != feat_list[SEQFEAT_CDREGION].end();feat_iter++){

		GeneAnnotation tmp_gene;
		gene_list.push_back(tmp_gene);

		gene_list.back() = *feat_iter;

		// Did we parse the entry correctly?
		if(gene_list.back().type() == GeneAnnotation::NONE){
			gene_list.pop_back();
		}
	}
	
	uout << "Parsing gene records ..." << uout_endl;

	for(feat_iter = feat_list[SEQFEAT_GENE].begin();
		feat_iter != feat_list[SEQFEAT_GENE].end();feat_iter++){

		GeneAnnotation tmp_gene;
		
		tmp_gene = *feat_iter;
		
		// Check to see if this gene record is already present in the
		// form of a CDS region
		list<GeneAnnotation>::iterator iter = 
			find(gene_list.begin(), gene_list.end(), tmp_gene);
			
		if(iter == gene_list.end()){

			// Couldn't find the gene
			gene_list.push_back(tmp_gene);

			// Did we parse the entry correctly?
			if(gene_list.back().type() == GeneAnnotation::NONE){
				gene_list.pop_back();
			}
		}
		else{

			// Change the annotation type from CDS to GENE.
			// Don't change PSEUDO_GENES to GENES -- they stay pseudo!
			if(iter->type() == GeneAnnotation::CDS){
				iter->type(tmp_gene.type());
			}
			
			// Add the gene locus and/or locus tag
			iter->info(GeneAnnotation::LOCUS, tmp_gene.info(GeneAnnotation::LOCUS));
			iter->info(GeneAnnotation::LOCUS_TAG, tmp_gene.info(GeneAnnotation::LOCUS_TAG));
		}
	}
	
	uout << "Parsing RNA records ..." << uout_endl;
	
	for(feat_iter = feat_list[SEQFEAT_RNA].begin();
		feat_iter != feat_list[SEQFEAT_RNA].end();feat_iter++){

		GeneAnnotation tmp_gene;

		tmp_gene = *feat_iter;

		// Check to see if this gene record is already present in a 
		// previously recorded region
		list<GeneAnnotation>::iterator iter = 
			find(gene_list.begin(), gene_list.end(), tmp_gene);
			
		if(iter == gene_list.end()){

			// Couldn't find the gene
			gene_list.push_back(tmp_gene);

			// Did we parse the entry correctly?
			if(gene_list.back().type() == GeneAnnotation::NONE){
				gene_list.pop_back();
			}
		}
		else{

			// Change the annotation type to (t)RNA
			iter->type(tmp_gene.type());
			
			// DON'T add the RNA locus and/or locus tag (since it has none!)
		}
	}
	
	uout << "Parsing misc records ..." << uout_endl;

	for(feat_iter = feat_list[SEQFEAT_IMP].begin();
		feat_iter != feat_list[SEQFEAT_IMP].end();feat_iter++){

		GeneAnnotation tmp_gene;
		gene_list.push_back(tmp_gene);

		gene_list.back() = *feat_iter;

		// Did we parse the entry correctly?
		if(gene_list.back().type() == GeneAnnotation::NONE){
			gene_list.pop_back();
		}
	}
	
	uout << "Found " << (unsigned int)( gene_list.size() ) << " genes. Assigning protein annotation ..." 
		 << uout_endl;
	
	
	for(feat_iter = feat_list[SEQFEAT_PROT].begin();
		feat_iter != feat_list[SEQFEAT_PROT].end();feat_iter++){
	
		if(!(*feat_iter)->location){
			continue;
		}
		
		list<GeneAnnotation>::iterator g_iter = find(gene_list.begin(),
				gene_list.end(), SeqLocId((*feat_iter)->location));
				
		if(g_iter == gene_list.end()){
			continue;
		}
		
		// Assume that sfp->data.choice == SEQFEAT_PROT
		ProtRefPtr prp = (ProtRefPtr)(*feat_iter)->data.value.ptrvalue;
		
		// Set the product annotation
		if(prp->name && prp->name->data.ptrvalue){
			g_iter->info(GeneAnnotation::PRODUCT, (char*)(prp->name->data.ptrvalue));
		}
		else{
			if(prp->desc){
				g_iter->info(GeneAnnotation::PRODUCT, prp->desc);
			}
		}
		
		// Is there an ec value?
		if(prp->ec && prp->ec->data.ptrvalue){
			g_iter->info(GeneAnnotation::EC, (char*)(prp->ec->data.ptrvalue));
		}
	}
	
	processGeneList(true /* Loading this data for the first time */);
	
	// Clean up any objects stored by the NCBI toolkit
	ObjMgrFreeCache(0);
	
	return true;
}

void DNAMol::processGeneList(const bool &m_loading /* = false */)
{
	uout << "Sorting and computing gene depth ..." << uout_endl;
	
	list<GeneAnnotation>::iterator gene_iter;

	vector<unsigned int> length_dist;

	bool annot_overlaps_origin = false;
	
	const unsigned int num_annot = gene_list.size();
	const unsigned int seq_len = seq.size();
	
	unsigned int length_threshold = 0;

	if(scale_mode == SIZE_CENTRIC){
		length_dist.reserve(num_annot);
	}

	for(gene_iter = gene_list.begin();gene_iter != gene_list.end();gene_iter++){
		
		// Handle the special case of a gene that overlaps the genome start site
		if( gene_iter->handle_gene_start_overlap(seq_len) == true){
			annot_overlaps_origin = true;
		}
		
		if( (scale_mode == SIZE_CENTRIC) && (gene_iter->type() <= GeneAnnotation::PSEUDO_GENE) ){
			length_dist.push_back( gene_iter->length(seq_len) );
		}
	}
	
	if(scale_mode == SIZE_CENTRIC){
	
		// Compute the median length 
		sort( length_dist.begin(), length_dist.end() );

		// Don't attempt to access the length_dist vector if it is empty 
		// (i.e. there are no annotations).
		if(length_dist.empty() == false){
			length_threshold = length_dist[length_dist.size()/2];
		}
	}
	
	// Sort genes
	gene_list.sort();
		
	// Compute the annotation depth (the amount of overlap in the linear genome map)
	// It is important to backfill genes to gene the minimum depth!
	list<unsigned int> stop_top_list;
	list<unsigned int> stop_mid_list;
	list<unsigned int>::iterator stop_iter;
	unsigned int depth_count;

	// Reset the gene and annotation count
	num_top_level = num_mid_level = 0;

	if(scale_mode == GENE_CENTRIC){
	
		for(gene_iter = gene_list.begin();gene_iter != gene_list.end();gene_iter++){

			// By default, the rank is MID_LEVEL unless otherwise specified
			unsigned char rank = GeneAnnotation::MID_LEVEL_RANK;

			switch( gene_iter->type() ){
				case GeneAnnotation::GENE:
				case GeneAnnotation::PSEUDO_GENE:
				case GeneAnnotation::CDS:
				case GeneAnnotation::RNA:
				case GeneAnnotation::tRNA:
				
					// Compute the depth of gene-gene overlap
					for(depth_count = 0, stop_iter = stop_top_list.begin();
						stop_iter != stop_top_list.end();
						stop_iter++, depth_count++){
						
						if( (gene_iter->start() > *stop_iter) && (gene_iter->overlaps_origin() == false) ){
							*stop_iter = gene_iter->stop();
							break;
						}
					}
					
					if( stop_iter == stop_top_list.end() ){
						stop_top_list.push_back( gene_iter->stop() );
					}
					
					gene_iter->depth(depth_count);

					rank = GeneAnnotation::TOP_LEVEL_RANK;

					num_top_level ++;

					// Fall through
				case GeneAnnotation::IMP:
				case GeneAnnotation::PRIMER:
				case GeneAnnotation::USER:
				case GeneAnnotation::NONE:
					// Compute the depth of annotation-annotation overlap
					for(depth_count = 0, stop_iter = stop_mid_list.begin();
						stop_iter != stop_mid_list.end();
						stop_iter++, depth_count++){
						
						if( (gene_iter->start() > *stop_iter) && (gene_iter->overlaps_origin() == false) ){
							*stop_iter = gene_iter->stop();
							break;
						}
					}
					
					if( stop_iter == stop_mid_list.end() ){
						stop_mid_list.push_back(gene_iter->stop());
					}

					gene_iter->annot_depth(depth_count);
					
					if(rank == GeneAnnotation::MID_LEVEL_RANK){
						gene_iter->depth(0);
					}

					// Set the rank for *all* annotations here!
					gene_iter->annot_rank(rank);
					break;
			};
		}
	}
	else if(scale_mode == SIZE_CENTRIC){
	
		for(gene_iter = gene_list.begin();gene_iter != gene_list.end();gene_iter++){
			
			// By default, the rank is MID_LEVEL unless otherwise specified
			unsigned char rank = GeneAnnotation::MID_LEVEL_RANK;

			// Is this a top level annotation -- based on size alone?
			if(gene_iter->length(seq_len) >= length_threshold){
			
				// Compute the depth of top-level overlap
				for(depth_count = 0, stop_iter = stop_top_list.begin();
					stop_iter != stop_top_list.end();
					stop_iter++, depth_count++){
					
					if( (gene_iter->start() > *stop_iter) && (gene_iter->overlaps_origin() == false) ){
					
						*stop_iter = gene_iter->stop();
						break;
					}
				}
				
				if( stop_iter == stop_top_list.end() ){
				
					stop_top_list.push_back( gene_iter->stop() );
				}
					
				gene_iter->depth(depth_count);

				rank = GeneAnnotation::TOP_LEVEL_RANK;

				num_top_level ++;
			}

			// Compute the depth of mid-level overlap
			for(depth_count = 0, stop_iter = stop_mid_list.begin();
				stop_iter != stop_mid_list.end();
				stop_iter++, depth_count++){
				
				if( (gene_iter->start() > *stop_iter) && (gene_iter->overlaps_origin() == false) ){
					*stop_iter = gene_iter->stop();
					break;
				}
			}
			
			if( stop_iter == stop_mid_list.end() ){
				stop_mid_list.push_back( gene_iter->stop() );
			}
			
			if(rank == GeneAnnotation::MID_LEVEL_RANK){
				gene_iter->depth(0);
			}

			gene_iter->annot_depth(depth_count);

			// Set the rank for *all* annotations here!
			gene_iter->annot_rank(rank);
		}
	}

	// Handle the special case when both the first and last annotation overlap 
	// the origin
	if(gene_list.size() > 1){
			
		if( (gene_list.back().annot_depth() == 0) && 
			gene_list.front().overlaps_origin() && 
			( gene_list.back().stop() >= gene_list.front().start() ) ){
		
			gene_list.back().annot_depth(1);
		}
	}

	if(m_loading == true){
		uout << "Adding intergenic spaces ..." << uout_endl;
	}

	if( m_loading && gene_list.empty() ){
	
		// No annotations were found! Is there any sequence to be read?
		// If so, then it will be classified as intergenic space
		if(seq.size() != 0){
		
			GeneAnnotation inter_genic_space;
				
			inter_genic_space.start(0);
			inter_genic_space.stop(seq.size() - 1);

			// Since this record only has intergenic space, make this
			// space top level
			inter_genic_space.annot_rank(GeneAnnotation::TOP_LEVEL_RANK);

			gene_list.push_back(inter_genic_space);
			
			// Since there is only a single annotation now, it
			// will automatically be top-level.
			num_top_level ++;
		}
	}
	else if(m_loading == true){

		unsigned int last_stop_plus_1;
		unsigned int first_start;
		bool annot_is_entire_mol = false;

		gene_iter = gene_list.begin();

		if(is_circular){
		
			last_stop_plus_1 = wrap( gene_iter->stop() + 1, seq.size() );
			first_start = gene_iter->start();
			
			// if last_stop_plus_1 == first_start, then there is an
			// annotation element that encompasses the entire molecule
			// and intergenic space is not needed.
			if(last_stop_plus_1 == first_start){
				annot_is_entire_mol = true;
			}
				
			gene_iter++;
		}
		else{ // linear genome
			last_stop_plus_1 = 0;
			first_start = 0;
		}
		
		// Only add intergenic space if there is room -- i.e. no
		// molecule spanning single annotation!
		if(annot_is_entire_mol == false){
		
			// Add intergenic spaces to the gene list.
			for(/* Already initialized */;gene_iter != gene_list.end();gene_iter++){
				
				// Only consider annotations that do not overlap the origin
				// (i.e. only consider cases where start <= stop).
				if( ( gene_iter->start() <= gene_iter->stop() ) && 
				     (gene_iter->start() > last_stop_plus_1) ){
					
					GeneAnnotation inter_genic_space;
					
					inter_genic_space.start(last_stop_plus_1);
					inter_genic_space.stop(gene_iter->start() - 1);
					
					if( (scale_mode == SIZE_CENTRIC) && 
						(inter_genic_space.length(seq_len) >= length_threshold) ){
						inter_genic_space.annot_rank(GeneAnnotation::TOP_LEVEL_RANK);
					}
					else{
						inter_genic_space.annot_rank(GeneAnnotation::MID_LEVEL_RANK);
					}

					gene_list.insert(gene_iter, inter_genic_space);
				}
				
				last_stop_plus_1 = max(last_stop_plus_1, gene_iter->stop() + 1);
			}
			
			// Do we need to add intergenic space to the tail?
			last_stop_plus_1 = wrap(last_stop_plus_1, seq.size());

			first_start = (is_circular) ? wrap(first_start - 1, seq.size()) : seq.size() - 1;
			
			// Only add intergenic space if no annotation overlaps the origin.
			// Also, when last_stop_plus_1 == 0 and first_start == seq.size() - 1, there is
			// no intergenic space (these points are right next to each other on the
			// circle).
			if( !( (last_stop_plus_1 == 0) && (first_start == seq.size() - 1) ) &&
				(last_stop_plus_1 - 1 != first_start) && (annot_overlaps_origin == false)){
				GeneAnnotation inter_genic_space;
				
				if(last_stop_plus_1 <= first_start){
					inter_genic_space.start(last_stop_plus_1);
					inter_genic_space.stop(first_start);
				}
				else{
					inter_genic_space.start(last_stop_plus_1);
					inter_genic_space.stop(seq.size() + first_start);
				}
				
				if( (scale_mode == SIZE_CENTRIC) && 
					(inter_genic_space.length(seq_len) >= length_threshold) ){
					inter_genic_space.annot_rank(GeneAnnotation::TOP_LEVEL_RANK);
				}
				else{
					inter_genic_space.annot_rank(GeneAnnotation::MID_LEVEL_RANK);
				}
				
				gene_list.push_back(inter_genic_space);
			}
		}
	}

	num_mid_level = gene_list.size();

	uout << "Done" << uout_endl;
}

bool DNAMol::loadASN(const std::string &m_filename, streampos &m_pos)
{
	AsnIoPtr		aip;
	AsnTypePtr	atp_se;
	AsnTypePtr	atp;
	AsnModulePtr	amp;
	SeqEntryPtr	sep;

	// Load SeqEntry object loader and sequence alphabets
	if( !SeqEntryLoad() ){
		throw "loadASN: SeqEntryLoad failed";
	}
	
	// Determine the file type (binary or ASCII) by probing the file
	// with GetAsnFileType.
	if ( (aip = AsnIoOpen( (char*)m_filename.c_str(), (char*)GetAsnFileType(m_filename) ) ) == NULL){
		throw "loadASN: Unable to open input file";
	}
	
	uout << "Reading (in ASN.1) " << m_filename << uout_endl;
	
	// Jump to the specified location in the file
	AsnIoSeek(aip, m_pos);
	
	// We restrict ourselves to reading ASN.1 files that contain annotation and
	// sequence data!
	amp = AsnAllModPtr();
	atp_se = AsnFind("Seq-entry");
	atp = AsnReadId(aip, amp, NULL);

	// The following test only works with ASCII text files
	if( (aip->type & ASNIO_TEXT) && (atp != atp_se) ){
	
		aip = AsnIoClose(aip);

		// If we were in the process of reading mulitple genomes from a single
		// ASN file, then we have read the last genome. Return false to stop
		// checking this file for more data.
		if( (long)m_pos != 0 ){
			return false;
		}
	
		throw "This ASN.1 file does not contain a Seq-entry";
	}

	// Rewind the file
	AsnIoSeek(aip, m_pos);

	// Read in the whole entry into the Sequence Entry Pointer, sep.
	// Close the ASN stream, which in turn closes the input file.
	sep = SeqEntryAsnRead(aip, NULL);
	
	if(sep == NULL){
		// There is no genome to load
		return false;
	}
	
	// Get the current location
	m_pos = AsnIoTell(aip);
	
	aip = AsnIoClose(aip);
			
	loadASN(sep);
	
	sep = SeqEntryFree(sep);
	
	uout << "Finished loading " << m_filename << uout_endl;
	
	return true;
}

// This function is based on ncbi\demo\getfasta.c
bool DNAMol::loadNetASN(const unsigned int &m_gi)
{
	SeqEntryPtr sep = NULL;
	
	// Load SeqEntry object loader and sequence alphabets
	if (! SeqEntryLoad()) {
		throw "loadNetASN: SeqEntryLoad failed";
	}

	is_fetch_enabled = PubSeqFetchEnable();

	try{
		
		sep = PubSeqSynchronousQuery(m_gi, 0, 0);

		if (sep == NULL){
			throw "Error downloading (sep was NULL). Please download directly from NCBI";
		}
		
		// This SeqEntryPtr may contain multiple entires. Extract
		// the entry that corresponds to m_gi
		// Search for the desired entry
		SeqId sid;
		
		sid.choice = SEQID_GI;
		sid.data.intvalue = m_gi;
		sid.next = NULL;
		
		loadASN(sep, &sid);		
	}
	catch(const char *error){
	
		if(sep){
			sep = SeqEntryFree(sep);
		}

		PubSeqFetchDisable();
		is_fetch_enabled = false;
		
		throw error;
	}
	catch(...){
	
		if(sep){
			sep = SeqEntryFree(sep);
		}

		PubSeqFetchDisable();
		is_fetch_enabled = false;
		
		throw ":loadNetASN: Unknown error occured parsing ASN data";
	}

	if(sep){
		sep = SeqEntryFree(sep);
	}

	PubSeqFetchDisable();
	is_fetch_enabled = false;
	
	return true;
}


bool DNAMol::loadNetASN(const unsigned int &m_gi, const string &m_filename)
{
	SeqEntryPtr sep = NULL;
	
	// Load SeqEntry object loader and sequence alphabets
	if (! SeqEntryLoad()) {
		throw "loadNetASN: SeqEntryLoad failed";
	}

	is_fetch_enabled = PubSeqFetchEnable();

	try{
		
		sep = PubSeqSynchronousQuery(m_gi, 0, 0);
		
		if (sep == NULL){
			throw "Error downloading: sep was NULL";
		}
		
		// This SeqEntryPtr may contain multiple entires. Extract
		// the entry that corresponds to m_gi
		// Search for the desired entry
		SeqId sid;
		
		sid.choice = SEQID_GI;
		sid.data.intvalue = m_gi;
		sid.next = NULL;
		
		loadASN(sep, &sid);
	}
	catch(const char *error){
		if(sep){
			sep = SeqEntryFree(sep);
		}

		PubSeqFetchDisable();
		is_fetch_enabled = false;
		
		throw error;
	}
	catch(...){
		if(sep){
			sep = SeqEntryFree(sep);
		}

		PubSeqFetchDisable();
		is_fetch_enabled = false;
		
		throw "Unknown error occured parsing ASN data";
	}
			
	// Save this sep to disk
	if(sep == NULL){
		throw "Unable to save ASN data to disk";
	}

	// Save as a text file to allow user inspection (for now)
	AsnIoPtr aip = AsnIoOpen((char*)m_filename.c_str(), "w");
		
	if(aip == NULL){
	
		sep = SeqEntryFree(sep);
		throw "Unable to save file to disk. Please check file name.";
	}
		
	SeqEntryAsnWrite(sep, aip, NULL);
	
	if(aip){
		aip = AsnIoClose(aip);
	}
	
	if(sep){
		sep = SeqEntryFree(sep);
	}
	
	PubSeqFetchDisable();
	is_fetch_enabled = false;
	
	return true;
}

bool DNAMol::loadFASTA(const std::string &m_filename)
{
	// First, convert the Fasta file into a SeqEntryPtr
	FILE *fin;

	fin = fopen(m_filename.c_str(), "r");

	if(!fin){
		throw ":loadFASTA: Unable to open FASTA file";
	}

	SeqEntryPtr	sep = NULL;

	uout << "Reading (in FASTA) " << m_filename << uout_endl;

	list<BioseqPtr> bsp_list;
	list<SeqEntryPtr> sep_list;

	unsigned int total_len = 0;

	info_map[TAXA_NAME] = m_filename;

	// Read each record, one at a time
	while( (sep = FastaToSeqEntry(fin, true /* is Nucleic acid */)) ){

		Int4 seqnum = BioseqCount(sep);
		BioseqPtr bsp;

		// The DNA moclecule of interest is the first bioseq in the file
		if(seqnum != 1){
			//throw __FILE__ ":DNAMol::loadASN: Unable to find a single DNA molecule!";

			// Try to keep reading
			continue;
		}

		BioseqExplore(sep, (Pointer) &bsp, BuildBioseqList);

		bsp_list.push_back(bsp);

		// Assume that each bioseq record corresponds to a gene. Not true,
		// but not a lot of info to go on.
		GeneAnnotation tmp_gene;

		tmp_gene.type(GeneAnnotation::GENE);

		tmp_gene.seqid(bsp->id);

		// Create a fake gene locus that provides a helpful name that
		// will be displayed on the gene
		string tmp_locus = tmp_gene.seq_id_str();
		
		string::size_type loc = tmp_locus.find('|');

		if(loc != string::npos){
			loc += 1;

			tmp_gene.info( GeneAnnotation::LOCUS, tmp_locus.substr(loc, tmp_locus.size() - loc) );
		}
		else{
			tmp_gene.info( GeneAnnotation::LOCUS, tmp_locus );
		}
		
		// Gene boundaries are zero based
		tmp_gene.start(total_len);

		total_len += BioseqGetLen(bsp);
		
		tmp_gene.stop(total_len - 1);

		tmp_gene.info(GeneAnnotation::NOTE, BioseqGetTitle(bsp));

		gene_list.push_back(tmp_gene);

		sep_list.push_back(sep);
	}

	fclose(fin);

	list<BioseqPtr>::iterator bsp_iter;
	
	// No longer needed
	// num_top_level = num_mid_level = gene_list.size();

	seq = DNA3Seq(total_len);

	DNA3Seq::iterator iter = seq.begin();

	// Prepare to compute GC content and base composition
	gc_content = 0.0f;
	a_content = 0.0f;
	t_content = 0.0f;
	g_content = 0.0f;
	c_content = 0.0f;
	other_content = 0.0f;

	// Store the DNA sequence
	for(bsp_iter = bsp_list.begin();bsp_iter != bsp_list.end();bsp_iter++){
		Uint1 residue;
		
		// Save the sequence data
		SeqPortPtr spp = SeqPortNew(*bsp_iter, FIRST_RESIDUE, LAST_RESIDUE, 
			Seq_strand_plus, Seq_code_iupacna);

		while ((residue = SeqPortGetResidue(spp)) != SEQPORT_EOF) {

			if (IS_residue(residue)){
				iter.c_base(residue);

				// Update the %GC content
				switch(iter.base()){
					case DNA3Seq::A:
						a_content ++;
						break;
					case DNA3Seq::T:
						t_content ++;
						break;
					case DNA3Seq::G:
						g_content ++;
						gc_content ++;
						break;
					case DNA3Seq::C:
						c_content ++;
						gc_content ++;
						break;
					default:
						other_content ++;
						break;
				};
					
			}
			else{
				switch (residue)
				{
					case SEQPORT_VIRT:
						iter.base(DNA3Seq::GAP);
						break;
					default:
						iter.base(DNA3Seq::UNKNOWN);
						break;
				}
			}
			
			iter++;
		}

		spp = SeqPortFree(spp);
	}


	gc_content /= seq.size();
	a_content /= seq.size();
	t_content /= seq.size();
	g_content /= seq.size();
	c_content /= seq.size();
	other_content /= seq.size();

	// Is this genome linear or circular? There is no way to known, so assume not
	is_circular = false;

	// Clean up
	list<SeqEntryPtr>::iterator sep_iter;
	for(sep_iter = sep_list.begin();sep_iter != sep_list.end();sep_iter++){
		SeqEntryFree(*sep_iter);
	}

	processGeneList(true /* Loading this data for the first time */);

	uout << "Ready" << uout_endl;
	
	return true;
}

bool DNAMol::import(const DNAMol &m_mol, const unsigned int &m_type)
{
	// Import all annotations from m_mol that are within the sequence bounds of 
	// mol.
	
	// Delete all intergenic space prior to the addition of new annotation
	list<GeneAnnotation>::iterator iter;

	list<list<GeneAnnotation>::iterator> reaper;
	
	for(iter = gene_list.begin();iter != gene_list.end();iter++){
		if(iter->is_intergenic() == true){
			reaper.push_back(iter);
		}
	}
	
	while(reaper.empty() == false){
		gene_list.erase( reaper.back() );
		reaper.pop_back();
	}
		
	const unsigned int num_bases = seq.size();
	
	list<GeneAnnotation>::const_iterator m_iter;
	
	for(m_iter = m_mol.gene_list.begin();m_iter != m_mol.gene_list.end();m_iter++){
		// If this annotation *starts* before the sequence end, then we will include
		// it (to allow for annotations that overlap the oriC).
		if(m_iter->start() < num_bases){
		
			gene_list.push_back(*m_iter);
			
			if(m_iter->stop() > num_bases){
			
				// This annotation needs to wrap around the oriC
				gene_list.back().stop(m_iter->stop() - num_bases);
			}
		}
	}
	
	processGeneList(true /* recomputing intergenic space */);
	
	return true;
}

void DNAMol::taxaInfo(ValNodePtr m_vnp)
{	
	if(!m_vnp){
		return;
	}
	
	BioSourcePtr bsp;
	OrgRefPtr orp;
	OrgNamePtr onp;

	bsp = (BioSourcePtr)m_vnp->data.ptrvalue;
	
	if(!bsp){
		return;
	}

	switch(bsp->genome){
		case 0:
			info_map[SOURCE] = "Unknown";
			break;
		case 1:
			info_map[SOURCE] = "genomic";
			break;
		case 2:
			info_map[SOURCE] = "chloroplast";
			break;
		case 3:
			info_map[SOURCE] = "chromoplast";
			break;
		case 4:
			info_map[SOURCE] = "kinetoplast";
			break;
		case 5:
			info_map[SOURCE] = "mitochondrion";
			break;
		case 6:
			info_map[SOURCE] = "plastid";
			break;
		case 7:
			info_map[SOURCE] = "macronuclear";
			break;
		case 8:
			info_map[SOURCE] = "extrachrom";
			break;
		case 9:
			info_map[SOURCE] = "plasmid";
			break;
		case 10:
			info_map[SOURCE] = "transposon";
			break;
		case 11:
			info_map[SOURCE] = "insertion-seq";
			break;
		case 12:
			info_map[SOURCE] = "cyanelle";
			break;
		case 13:
			info_map[SOURCE] = "proviral";
			break;
		case 14:
			info_map[SOURCE] = "viral";
			break;
		case 15:
			info_map[SOURCE] = "nucleomorph";
			break;
		case 16:
			info_map[SOURCE] = "apicoplast";
			break;
		case 17:
			info_map[SOURCE] = "leucoplast";
			break;
		case 18:
			info_map[SOURCE] = "proplastid";
			break;
		case 19:
			info_map[SOURCE] = "endogenous_virus";
			break;
		default:
			throw "taxaInfo: Unknown biosource!";
	};
	
	orp = bsp->org;

	if(orp->taxname){
		info_map[TAXA_NAME] = orp->taxname;
	}
	else{
		if(orp->common){
			info_map[TAXA_NAME] = orp->taxname;
		}
	}

	onp = orp->orgname;

	if(!onp){
		return;
	}

	if(onp){
		if(onp->lineage){
			info_map[LINEAGE] = onp->lineage;
		}
		
		BinomialOrgNamePtr bonp;
		
		switch(onp->choice){
			case 1: // binomial
				bonp = (BinomialOrgNamePtr)onp->data;
				
				if(!bonp){
					return;
				}
				
				if(bonp->genus){
					info_map[GENUS] = bonp->genus;
				}
				
				if(bonp->species){
					info_map[SPECIES] = bonp->species;
				}
				
				if(bonp->subspecies){
					info_map[SUBSPECIES] = bonp->subspecies;
				}
				
				break;
		}
	}
}

// Allow the user to search for all annotations that
// contain or partially contain the given DNA sequence.
list< list<GeneAnnotation>::iterator > DNAMol::find_seq(const list<DNA3Seq> &m_query_list, 
														list<SeqHit> &hit_list)
{
	list< list<GeneAnnotation>::iterator > annot_iter_list;

	if( m_query_list.empty() ){
		return annot_iter_list;
	}

	DNA3Seq::iterator iter = seq.begin();
	
	while( (iter = seq.find(m_query_list.front(), iter, is_circular) ) != seq.end() ){
	
		list<unsigned int> last_base;

		list<DNA3Seq>::const_iterator m_query_next = m_query_list.begin();

		m_query_next ++;

		if( m_query_next == m_query_list.end() ){
			last_base.push_back(iter.index() + m_query_list.front().size() - 1);			
		}
		else{
			match_seq_list(m_query_next, 
				m_query_list.end(),
				iter + m_query_list.front().size(),
				seq,
				is_circular,
				last_base);
		}
		
		if(last_base.empty() == true){
			break;
		}

#define FIND_ALL_OCCURANCES

#ifdef FIND_ALL_OCCURANCES
		list<unsigned int>::iterator last_iter;

		for(last_iter = last_base.begin();last_iter != last_base.end();last_iter++){
		
			// index() is 0 based and the NCBI ASN.1 entries are also 
			// 0 based.
			unsigned int match_start = iter.index();
			unsigned int match_stop = *last_iter;

			// Save the location of the matching sequence
			hit_list.push_back(SeqHit(match_start, match_stop, false, SeqHit::EXACT_MATCH));

			list<GeneAnnotation>::iterator a_iter;
			
			for(a_iter = gene_list.begin();a_iter != gene_list.end();a_iter++){
				
				if(overlap(a_iter->start(), a_iter->stop(), match_start, match_stop, seq.size()) == true){
					// We found the product! Save the annotation that
					// matched.
					annot_iter_list.push_back(a_iter);
				}
			}
		}
		
#else // Find the most compact occurance
		// index() is 0 based and the NCBI ASN.1 entries are also 
		// 0 based.
		unsigned int match_start = iter.index();
		unsigned int match_stop = last_base.front();

		// Save the location of the matching sequence
		hit_list.push_back(SeqHit(match_start, match_stop, false));

		list<GeneAnnotation>::iterator a_iter;
			
		for(a_iter = gene_list.begin();a_iter != gene_list.end();a_iter++){
				
			if(overlap(a_iter->start(), a_iter->stop(), match_start, match_stop, seq.size()) == true){
				// We found the product! Save the annotation that
				// matched.
				annot_iter_list.push_back(a_iter);
			}
		}
#endif // FIND_ALL_OCCURANCES
		
		iter++;		
	}

	// Now search for the complement of the search seaquences
	list<DNA3Seq> query_complement_list;
	
	list<DNA3Seq>::const_iterator query_iter;

	// The order of the entries in the list must be inverted
	// to be a valid complement.
	for(query_iter = m_query_list.begin();query_iter != m_query_list.end();query_iter++){
		query_complement_list.push_front(query_iter->complement());
	}
	
	// Start at the begining of the target sequence (again).
	iter = seq.begin();
	
	while( (iter = seq.find( query_complement_list.front(), iter, is_circular) ) != seq.end() ){
	
		list<unsigned int> last_base;

		list<DNA3Seq>::const_iterator m_query_next = query_complement_list.begin();

		m_query_next ++;

		if(m_query_next == query_complement_list.end()){
			last_base.push_back(iter.index() + query_complement_list.front().size() - 1);
		}
		else{
			match_seq_list(m_query_next, 
				query_complement_list.end(),
				iter + query_complement_list.front().size(),
				seq,
				is_circular,
				last_base);
		}
		
		if(last_base.empty() == true){
			break;
		}

#ifdef FIND_ALL_OCCURANCES
		list<unsigned int>::iterator last_iter;

		for(last_iter = last_base.begin();last_iter != last_base.end();last_iter++){
			// index() is 0 based and the NCBI ASN.1 entries are also 
			// 0 based.
			unsigned int match_start = iter.index();
			unsigned int match_stop = *last_iter;

			// Save the location of the matching sequence
			hit_list.push_back(SeqHit(match_start, match_stop, true /* Complement */, SeqHit::EXACT_MATCH));

			list<GeneAnnotation>::iterator a_iter;
			
			for(a_iter = gene_list.begin();a_iter != gene_list.end();a_iter++){
				
				if(overlap(a_iter->start(), a_iter->stop(), match_start, match_stop, seq.size()) == true){
					// We found the product! Save the annotation that
					// matched.
					annot_iter_list.push_back(a_iter);
				}
			}
		}
#else // Find the most compact occurance
		unsigned int match_start = iter.index();
		unsigned int match_stop = last_base.front();

		// Save the location of the matching sequence
		hit_list.push_back(SeqHit(match_start, match_stop, true /* Complement */));

		list<GeneAnnotation>::iterator a_iter;
			
		for(a_iter = gene_list.begin();a_iter != gene_list.end();a_iter++){
				
			if(overlap(a_iter->start(), a_iter->stop(), match_start, match_stop, seq.size()) == true){
				// We found the product! Save the annotation that
				// matched.
				annot_iter_list.push_back(a_iter);
			}
		}
#endif // FIND_ALL_OCCURANCES

		iter++;
	}

	// Use template specialization to define a comparison
	// operator for GeneAnnotation *. Other (non-WIN32) implementations
	// of the STL may use std::less instead of std::greater -- be carefull!
	annot_iter_list.sort(greater< list<GeneAnnotation>::iterator >());

	annot_iter_list.unique();
	
	return annot_iter_list;
}

// Allow the user to search for all annotations that
// contain or partially contain the given DNA sequence.
bool match_seq_list(list<DNA3Seq>::const_iterator m_query_start, 
	list<DNA3Seq>::const_iterator m_query_stop,
	DNA3Seq::iterator target_start,
	const DNA3Seq &target,
	const bool &m_is_circular,
	list<unsigned int> &last_base)
{
	DNA3Seq::iterator iter = target_start;
	
	if(m_query_start == m_query_stop){
		return false;
	}
	
	while( (iter = target.find(*m_query_start, iter, m_is_circular)) != target.end() ){
		
		list<DNA3Seq>::const_iterator m_query_next = m_query_start;
		
		// The next search sequence
		m_query_next++;

		if(m_query_next == m_query_stop){
			last_base.push_back(iter.index() + m_query_start->size() - 1);
			return true;
		}
		else{
			bool ret = match_seq_list(m_query_next, m_query_stop, target_start + m_query_start->size(), 
				target, m_is_circular, last_base);

			if(ret == false){
				return false;
			}

			// Keep searching!
			iter++;
		}
	}

	return false;
}

// The string m_name must be lower case
list< list<GeneAnnotation>::iterator > DNAMol::find_name(const string &m_name)
{
	list< list<GeneAnnotation>::iterator > annot_iter_list;
	
	list<GeneAnnotation>::iterator iter;

	for(iter = gene_list.begin();iter != gene_list.end();iter++){
		
		string gene_locus = iter->info(GeneAnnotation::LOCUS);
		string gene_tag = iter->info(GeneAnnotation::LOCUS_TAG);

		if(gene_locus.empty() && gene_tag.empty()){
			continue;
		}
		
		unsigned int j;

		// First try to find the gene name
		for(j = 0;j < gene_locus.size();j++){
			gene_locus[j] = tolower(gene_locus[j]);
		}
		
		// Allow substring matches instead of just exact matches!
		//if(m_name == gene_locus){
		if(gene_locus.find(m_name) != string::npos){
			annot_iter_list.push_back(iter);
			continue;
		}
		
		// Couldn't find the gene name. Try to find the gene locus tag
		for(j = 0;j < gene_tag.size();j++){
			gene_tag[j] = tolower(gene_tag[j]);
		}
		
		// Allow substring matches instead of just exact matches!
		//if(m_name == gene_tag){
		if(gene_tag.find(m_name) != string::npos){
			annot_iter_list.push_back(iter);
			continue;
		}
	}

	// Use template specialization to define a comparison
	// operator for GeneAnnotation *. Other (non-WIN32) implementations
	// of the STL may use std::less instead of std::greater -- be carefull!
	annot_iter_list.sort(greater< list<GeneAnnotation>::iterator >());

	annot_iter_list.unique();
	
	return annot_iter_list;
}

list< list<GeneAnnotation>::iterator > DNAMol::find_gi(const unsigned int &m_gi)
{
	list< list<GeneAnnotation>::iterator > annot_iter_list;
	
	// Turn the gi into a valid SeqId
	SeqId sid;
	
	sid.choice = SEQID_GI;
	sid.data.intvalue = m_gi;

	// There should only be a single gi value, so use the 
	// find function
	list<GeneAnnotation>::iterator iter = find(gene_list.begin(), gene_list.end(), &sid);

	if(iter != gene_list.end()){
		annot_iter_list.push_back(iter);
	}
	
	return annot_iter_list;
}

list< list<GeneAnnotation>::iterator > DNAMol::find_base(const unsigned int &m_base)
{
	list< list<GeneAnnotation>::iterator > annot_iter_list;

	list<GeneAnnotation>::iterator iter;

	for(iter = gene_list.begin();iter != gene_list.end();iter++){
		// The annotations are sorted
		if(iter->start() > m_base){
			break;
		}

		if(iter->stop() >= m_base){
			// There can be multiple annotations that contain a given base
			annot_iter_list.push_back(iter);
		}
	}
	
	return annot_iter_list;
}

list< list<GeneAnnotation>::iterator > DNAMol::find_text(const list<string> &m_text)
{
	list< list<GeneAnnotation>::iterator > annot_iter_list;

	list<GeneAnnotation>::iterator iter;

	for(iter = gene_list.begin();iter != gene_list.end();iter++){
		
		string annot_text;
		
		annot_text = iter->info(GeneAnnotation::PRODUCT) + " " + 
					 iter->info(GeneAnnotation::NOTE);
		
		for(unsigned int j = 0;j < annot_text.size();j++){
			annot_text[j] = tolower(annot_text[j]);
		}

		list<string>::const_iterator key_iter;

		for(key_iter = m_text.begin();key_iter != m_text.end();key_iter++){
			if(annot_text.find(*key_iter) == string::npos){
				break;
			}
		}
		
		if( key_iter == m_text.end() ){
			// If we get here we found a match to all of the keywords
			annot_iter_list.push_back(iter);
		}
	}
	
	// Use template specialization to define a comparison
	// operator for GeneAnnotation *. Other (non-WIN32) implementations
	// of the STL may use std::less instead of std::greater -- be carefull!
	annot_iter_list.sort(greater< list<GeneAnnotation>::iterator >());

	annot_iter_list.unique();

	return annot_iter_list;
}

void DNAMol::clear_select()
{
	list<GeneAnnotation>::iterator iter;

	for(iter = gene_list.begin();iter != gene_list.end();iter++){
		iter->select(false);
	}
}

// Change from 5'->3' to 3'->5'
void DNAMol::invert()
{
	list<GeneAnnotation>::iterator iter;

	const unsigned int num_base = seq.size();

	for(iter = gene_list.begin();iter != gene_list.end();iter++){

		// From the NCBI sdk.doc file:
		// "Locations of regions on Bioseqs are always given as 
		//  integer offsets, also described in the Biological 
		//  Sequences chapter. So the first residue is always 0 and 
		//  the last residue is always (length - 1)." 
		
		//     0  .........................  (N-1)
		//     5' -----> ---> -->   ------>  3'
		//
		//     3' -----> ---> -->   ------>  5'
		// (N-1)  .........................  0
		
		// start' = N - 1 - stop
		// stop' = N - 1 - start
		const unsigned int tmp_start = iter->start();

		if(iter->stop() >= num_base){
			iter->start(2*num_base - 1 - iter->stop());
		}
		else{
			iter->start(num_base - 1 - iter->stop());
		}
		
		if(tmp_start >= num_base){
			iter->stop(2*num_base - 1 - tmp_start);
		}
		else{
			iter->stop(num_base - 1 - tmp_start);
		}
		
		// Switch all segments
		iter->invert_seg_list(num_base);

		// Switch the strand
		if(iter->strand() == Seq_strand_plus){
			iter->strand(Seq_strand_minus); 
		}
		else{
			iter->strand(Seq_strand_plus); 
		}
	}

	// Resort the gene list to place in 5'->3' order
	gene_list.sort();

	seq = seq.complement();

	complement = !complement;
}

GeneAnnotation& GeneAnnotation::operator=(const SeqFeatPtr &sfp)
{

	if(!sfp){
		throw "GeneAnnotation::=: NULL sfp";
	}

	if(!sfp->location){
		throw "GeneAnnotation::=: NULL sfp->location";
	}
	
	int tmp_start = SeqLocStart(sfp->location);
	int tmp_stop = SeqLocStop(sfp->location);

	if( (tmp_start < 0) || (tmp_stop < 0)){
		gene_start = 0;
		gene_stop = 0;
	}
	else{
		gene_start = (unsigned int)tmp_start;
		gene_stop = (unsigned int)tmp_stop;
	}

	// Is this one contiguous SeqLoc entry ?
	if(IS_one_loc(sfp->location, true) == false){

		// Look through the list of SeqLocs to extract the info
		// we need!
		SeqLocPtr slp = NULL;

		while( (slp = SeqLocFindNext(sfp->location, slp)) != NULL){
			// Save all of the segments
			tmp_start = SeqLocStart(slp);
			tmp_stop = SeqLocStop(slp);
			
			if( (tmp_start >= 0) && (tmp_stop >= 0)){
				// There is a bug in SeqLocStart/SeqLocStop that can cause
				// incorrect gene_start/gene_stop values (see NC_006151.asn for an example;
				// Note the NULL value in the seqloc).
				gene_start = min((int)gene_start, tmp_start);
				gene_stop = max((int)gene_stop, tmp_stop);

				seg_list.push_back( make_pair((unsigned int)tmp_start, (unsigned int)tmp_stop) );
			}
		}

		// Sort the seg list (important for annotations that span the origin)
		seg_list.sort();
	}

	complement = (SeqLocStrand(sfp->location) == Seq_strand_minus);
	
	if(sip){
		sip = SeqIdSetFree(sip);
	}
	
	if(sfp->product){
		sip = SeqIdSetDup(SeqLocId(sfp->product));
	}
	
	if(sfp->comment){
		info_map[NOTE] = sfp->comment;
	}
	
	GeneRefPtr grp;
	RnaRefPtr rrp;
	ImpFeatPtr ifp;
	CdRegionPtr crp;

	switch(sfp->data.choice){
		case SEQFEAT_GENE: // Gene
			grp = (GeneRefPtr)sfp->data.value.ptrvalue;
			
			// Is this a gene or pseudo gene?
			gene_type = (sfp->pseudo) ? PSEUDO_GENE : GENE;

			if(grp->locus){
				info_map[LOCUS] = grp->locus;
			}

			if(grp->locus_tag){
				info_map[LOCUS_TAG] = grp->locus_tag;
			}

			break;
		case SEQFEAT_CDREGION: // CD-region
			crp = (CdRegionPtr)sfp->data.value.ptrvalue;

			// Is this a CDS or pseudo gene?
			gene_type = (sfp->pseudo) ? PSEUDO_GENE : CDS;
			
			{
				// Create a fake gene locus that provides a helpful name that
				// will be displayed on the CDS
				info_map[LOCUS] = seq_id_str();
				
				string::size_type loc = info_map[LOCUS].find('|');

				if(loc != string::npos){
					loc += 1;

					info_map[LOCUS] = info_map[LOCUS].substr(loc, info_map[LOCUS].size() - loc);
				}
			}

			break;
		case SEQFEAT_PROT: // Prot-ref
			throw "GeneAnnotation::=: Protein record assigned to gene!";
			
			break;
		case SEQFEAT_RNA: // RNA
			rrp = (RnaRefPtr)sfp->data.value.ptrvalue;
			
			if(rrp == NULL){
				throw "GeneAnnotation::=: NULL rrp";
			}
			
			// Note that there is also an mRNA type (that has been 
			// associated with packed-int SeqLoc structures). The non-local
			// nature of the packed-int SeqLoc throws off the parsing scheme, so
			// just skip these entries (they will have gene types == NONE and hence
			// be removed).
			switch(rrp->ext.choice){
				case 0: // Fall through
				case 1:
					gene_type = RNA;

					if(rrp->ext.value.ptrvalue != NULL){
						info_map[PRODUCT] = (char*)rrp->ext.value.ptrvalue;
					}
					break;
				case 2:
					// Codon is stored in:
					// tRNAPtr trp = (tRNAPtr)rrp->ext.value.ptrvalue;
					// trp->aa
					gene_type = tRNA;
					info_map[PRODUCT] = "Codon";
					break;
			};
			
			break;
		case SEQFEAT_IMP: // Imp
			ifp = (ImpFeatPtr)sfp->data.value.ptrvalue;
			
			gene_type = IMP;
			
			if(ifp == NULL){
				throw "GeneAnnotation::gene: NULL ifp";
			}
			
			info_map[PRODUCT] = ifp->key;
			
			// If there is a string stored in the qual pointer, add it to the
			// annotation note (this is where the label for an STS is stored)
			if(sfp->qual && (sfp->qual->val) ){
				
				info_map[NOTE] += sfp->qual->val;
			}
			
			break;
	}
	
	return *this;
}

// The invert_seg_list is a public helper function for DNAMol::invert()
void GeneAnnotation::invert_seg_list(const unsigned int &m_num_base)
{

	list< pair<unsigned int, unsigned int> >::iterator seg_list_iter;

	for(seg_list_iter = seg_list.begin();seg_list_iter != seg_list.end();seg_list_iter++){
		// start' = N - 1 - stop
		// stop' = N - 1 - start
		const unsigned int tmp_start = seg_list_iter->first;

		if(seg_list_iter->second >= m_num_base){
			seg_list_iter->first = 2*m_num_base - 1 - seg_list_iter->second;
		}
		else{
			seg_list_iter->first = m_num_base - 1 - seg_list_iter->second;
		}
		
		if(tmp_start >= m_num_base){
			seg_list_iter->second = 2*m_num_base - 1 - tmp_start;
		}
		else{
			seg_list_iter->second = m_num_base - 1 - tmp_start;
		}
	}
}	
		
// This SeqEntry exploration function copies the current pointer position in the
// the Bioseq entry to a list of Bioseq pointers
void BuildBioseqList(SeqEntryPtr sep, Pointer data, Int4 index, Int2 indent)
{
	((BioseqPtr PNTR) data)[index] = (BioseqPtr)sep->data.ptrvalue;
	return;
}

// For use with BioseqExplore: Only copy the location of the first bioseq
void FindBioseq(SeqEntryPtr sep, Pointer data, Int4 index, Int2 indent)
{
	// If global_search_id is NULL, return the first bioseq in the record
	// (note that global_search_id is a global variable!)
	if(global_search_id == NULL){
	
		if(0 == index){
			*((BioseqPtr PNTR) data) = (BioseqPtr)sep->data.ptrvalue;
		}
	}
	else{
	
		// If global_search_id is *not* NULL, return the bioseq record that
		// matches global_search_id.
		if( SeqIdIn(global_search_id, BioseqPtr(sep->data.ptrvalue)->id) ){
		
			// We found a match!
			*((BioseqPtr PNTR) data) = (BioseqPtr)sep->data.ptrvalue;
		}
	}
}

ostream& operator << (ostream &s, const pair<unsigned int, GeneAnnotation> &m_annot)
{
	switch(m_annot.second.gene_type){
		case GeneAnnotation::CDS:
			s << "CDS: ";
			break;
		case GeneAnnotation::GENE:
			s << "GENE: ";
			break;
		case GeneAnnotation::PSEUDO_GENE:
			s << "PSEUDO-GENE: ";
			break;
		case GeneAnnotation::RNA:
			s << "RNA: ";
			break;
		case GeneAnnotation::tRNA:
			s << "tRNA: ";
			break;
		case GeneAnnotation::NONE:
			s << "Intergenic space: ";
			break;
		case GeneAnnotation::PRIMER:
			s << "PCR Primer: ";
			break;
		case GeneAnnotation::USER:
			s << "User defined: ";
			break;
		default:
			s << "Misc: ";
			break;
	};
	
	if(m_annot.second.info(GeneAnnotation::LOCUS) != ""){
	
		s << m_annot.second.info(GeneAnnotation::LOCUS) << " ";
	}
	else{
		if(m_annot.second.info(GeneAnnotation::LOCUS_TAG) != ""){
		
			s << m_annot.second.info(GeneAnnotation::LOCUS_TAG) << " ";
		}
	}
		
	if(m_annot.second.sip){
		char buffer[256];
		
		SeqIdPtr sip = m_annot.second.sip;

		while(sip){
			SeqIdPrint(sip, buffer, '|');
			s << buffer;

			sip = sip->next;
		}

		s << ' ';
	}
	
	// Handle annotations that wrap around the origin
	const unsigned int gene_stop = (m_annot.first > m_annot.second.gene_stop) ?
		m_annot.second.gene_stop : (m_annot.second.gene_stop - m_annot.first);
		
	if(m_annot.second.complement){
		s << "complement(" << m_annot.second.gene_start << ".." 
			<< gene_stop << ")" << ENDL;
	}
	else{
		s << "[" << m_annot.second.gene_start << ".." 
			<< gene_stop << "]" << ENDL;
	}
	
	s << "Length = " << m_annot.second.length(m_annot.first) << " bases" << ENDL;

	if(m_annot.second.info(GeneAnnotation::PRODUCT) != ""){
		s << "Product = " << m_annot.second.info(GeneAnnotation::PRODUCT) << ENDL;
	}
	
	if(m_annot.second.info(GeneAnnotation::EC) != ""){
		s << "EC = " << m_annot.second.info(GeneAnnotation::EC) << ENDL;
	}
	
	if(m_annot.second.info(GeneAnnotation::NOTE) != ""){
		// Wrap long lines
		//s << "Note = " << m_annot.info(GeneAnnotation::NOTE) << ENDL;
		const string& str_ref = m_annot.second.info(GeneAnnotation::NOTE);

		s << "Note = ";
		
		const unsigned int max_line_len = 50;

		unsigned int line_count = 1;

		for(unsigned int i = 0;i < str_ref.size();i++, line_count++){
		
			if( (line_count >= max_line_len) && isspace(str_ref[i]) ){
				s << ENDL;
				s << "  ";
				line_count = 1;
			}
			else{
				s << str_ref[i];
			}
		}
	}

	if(m_annot.second.type() == GeneAnnotation::PRIMER){
		if(m_annot.second.info(GeneAnnotation::PRIMER_TM) != ""){
			s << m_annot.second.info(GeneAnnotation::PRIMER_TM) << ENDL;
		}

		if(m_annot.second.info(GeneAnnotation::PRIMER_GC) != ""){
			s << m_annot.second.info(GeneAnnotation::PRIMER_GC) << ENDL;
		}

		if(m_annot.second.info(GeneAnnotation::PRIMER_HAIRPIN_TM) != ""){
			s << m_annot.second.info(GeneAnnotation::PRIMER_HAIRPIN_TM) << ENDL;
		}

		if(m_annot.second.info(GeneAnnotation::PRIMER_DIMER_TM) != ""){
			s << m_annot.second.info(GeneAnnotation::PRIMER_DIMER_TM) << ENDL;
		}
	}
	
	return s;
}

ostream& operator << (ostream &s, const DNAMol &m_mol)
{
	if(m_mol.info(DNAMol::TAXA_NAME) != ""){
		s << m_mol.info(DNAMol::TAXA_NAME) << " ";
	}
	else{
		if(m_mol.info(DNAMol::GENUS) != ""){
			s << m_mol.info(DNAMol::GENUS) << " ";
		}
		
		if(m_mol.info(DNAMol::SPECIES) != ""){
			s << m_mol.info(DNAMol::SPECIES) << " ";
		}
		
		if(m_mol.info(DNAMol::SUBSPECIES) != ""){
			s << m_mol.info(DNAMol::SUBSPECIES) << " ";
		}
	}
	
	if(m_mol.info(DNAMol::SOURCE) != ""){
		s << "[" << m_mol.info(DNAMol::SOURCE) << "] ";
	}
	
	if(m_mol.sip){
		char buffer[256];
		SeqIdPtr tmp_id = m_mol.sip;
		
		while(tmp_id){
			SeqIdPrint(tmp_id, buffer, '|');
			s << buffer;
			
			tmp_id = tmp_id->next;
		}
		
		s << ENDL;
	}
	
	// Summarize the number of genes and other annotations
	vector<unsigned int> annot_count(GeneAnnotation::IMP + 1);

	list<GeneAnnotation>::const_iterator iter;
	unsigned int total_annot_count = 0;

	for(iter = m_mol.begin();iter != m_mol.end();iter++){
		if(iter->type() <= GeneAnnotation::IMP){
			annot_count[iter->type()] ++;
		}
	}
	
	s << "Annotations:" << ENDL;

	// The number of genes
	total_annot_count += annot_count[GeneAnnotation::GENE];
	s << '\t' << annot_count[GeneAnnotation::GENE] << " gene, ";

	// The number of pseudo genes
	total_annot_count += annot_count[GeneAnnotation::PSEUDO_GENE];
	s << annot_count[GeneAnnotation::PSEUDO_GENE] << " pseudo gene, ";

	// The number of CDS
	total_annot_count += annot_count[GeneAnnotation::CDS];
	s << annot_count[GeneAnnotation::CDS] << " CDS," << ENDL;

	// The number of RNA
	total_annot_count += annot_count[GeneAnnotation::RNA];
	s << '\t' << annot_count[GeneAnnotation::RNA] << " RNA, ";

	// The number of tRNA
	total_annot_count += annot_count[GeneAnnotation::tRNA];
	s << annot_count[GeneAnnotation::tRNA] << " tRNA, ";

	// The number of IMP
	total_annot_count += annot_count[GeneAnnotation::IMP];
	s << annot_count[GeneAnnotation::IMP] << " Misc" << ENDL;

	s << "\tTotal = " << total_annot_count << ENDL;

	return s;
}

int file_type(const std::string &m_filename)
{
	// Open the file and read unitl we can determine the file type
	ifstream fin(m_filename.c_str(), ios::binary);

	int ret = -1; // return -1 on error

	if(!fin){
		return ret;
	}
	
	const unsigned int buffer_size = 512;
	char buffer[buffer_size];

	// Clear the array of bytes
	memset(buffer, 0, buffer_size);

	fin.read(buffer, buffer_size);

	if(fin){
		fin.close();
	}

	bool GBK_hint = false;
	unsigned int ptt_hint = 0;
	
	int first_non_space_char = -1;

	for(unsigned int i = 0;i < buffer_size;i++){
		if( isspace(buffer[i]) ){
			continue;
		}

		if(first_non_space_char < 0){
			first_non_space_char = i;
		}

		if(buffer[i] == '>'){
			ret = DNAMol::FASTA;
			break;
		}
		
		if( (buffer[i] == '{') || (buffer[i] == '}') ){
			// ASCII Asn.1
			ret = DNAMol::ASN_1;
			break;
		}

		if(buffer[i] < 0){
			// Binary Asn.1
			ret = DNAMol::ASN_1;
			break;
		}

		if(isupper(buffer[i])){
			// Genbank flat files typically start with upper case string,
			// i.e. "LOCUS".
			GBK_hint = true;
			
			// This is only a hint, so don't break -- keep reading!
		}
		
		if(buffer[i] == '.'){
			if( (i != 0) && (buffer[i - 1] == '.') ){
				ptt_hint++;
			}
		}
	}

	// No valid characters detected! This is an error!
	if(first_non_space_char == -1){
		return ret;
	}
	
	// Quick test for GBK file
	if( strstr(buffer, "LOCUS") && strstr(buffer, "DEFINITION") ){
		return DNAMol::GBK;
	}
	
	// Use the arbitrary threshold of two ".." occurances
	// to indicate a PTT file.
	if(ptt_hint >= 2){
		return DNAMol::PTT;
	}
	
	if((ret == -1) && GBK_hint){

		// Does this file start with the string "ID"?
		if(first_non_space_char < (int)buffer_size - 1){
			if( ( (buffer[first_non_space_char] == 'I') && 
				(buffer[first_non_space_char + 1] == 'D') ) || 
				strstr(buffer, "\nFT") ){

				return DNAMol::EMBL;
			}
		}

		ret = DNAMol::GBK;
	}

	return ret;
}

NLM_EXTERN void GetFeatListCallback(SeqEntryPtr sep, Pointer data, Int4 index, Int2 indent)
{
	
	if(sep == NULL){
		return;
	}
	
	BioseqPtr bsp = NULL;
	BioseqSetPtr bssp = NULL;
	SeqAnnotPtr sap = NULL;
	SeqFeatPtr sfp = NULL;
	SeqIdPtr sid = NULL;
	list<SeqFeatPtr> *feat_list = (list<SeqFeatPtr> *)data;

	if (IS_Bioseq(sep))
	{
		bsp = (BioseqPtr)(sep->data.ptrvalue);
		sid = bsp->id;
		sap = bsp->annot;
	}
	else
	{
		bssp = (BioseqSetPtr)(sep->data.ptrvalue);
		sap = bssp->annot;
	}

	while (sap != NULL)
	{
		if (sap->type == 1)  /* feature table */
		{
			// If the global variable global_search_id is NULL, load all
			// entires. If global_search_id == the current SeqId, load
			// the corresponding features
			if( (global_search_id == NULL) || SeqIdIn(global_search_id, sid) ){
			
				for (sfp = (SeqFeatPtr)(sap->data); sfp != NULL; sfp = sfp->next){
					feat_list[sfp->data.choice].push_back(sfp);
				}
			}
			else{
				
				if(global_search_id != NULL){
				
					for (sfp = (SeqFeatPtr)(sap->data); sfp != NULL; sfp = sfp->next){
						
						if( SeqIdIn( global_search_id, SeqLocId(sfp->location) ) ){
							feat_list[sfp->data.choice].push_back(sfp);
						}
					}
				}
			}
		}

		sap = sap->next;
	}
}

NLM_EXTERN void ApplyOffsetCallback(SeqEntryPtr sep, Pointer data, Int4 index, Int2 indent)
{
	
	if(sep == NULL){
		return;
	}
	
	BioseqPtr bsp = NULL;
	SeqAnnotPtr sap = NULL;
	SeqFeatPtr sfp = NULL;
	Int4 from = *((Int4 *)data);
	
	if (IS_Bioseq(sep) )
	{
		bsp = (BioseqPtr)(sep->data.ptrvalue);
		sap = bsp->annot;		
	}
	else
	{
		return;
	}

	while (sap != NULL)
	{
		if (sap->type == 1)  /* feature table */
		{
			for (sfp = (SeqFeatPtr)(sap->data); sfp != NULL; sfp = sfp->next){
				
				// Add an offset to the feature
				add_loc_offset(sfp->location, from);
			}
		}

		sap = sap->next;
	}
}

void add_loc_offset(SeqLocPtr anp, const Int4 &m_offset)
{
	SeqIdPtr sip = NULL;
	
	if(anp == NULL){
		return;
	}

	switch(anp->choice){
		case SEQLOC_INT:
			
			if((SeqIntPtr)anp->data.ptrvalue == NULL){
				throw "Unable to apply annotation offset";
			}
			
			( (SeqIntPtr)anp->data.ptrvalue )->from += m_offset;
			( (SeqIntPtr)anp->data.ptrvalue )->to += m_offset;
			break;
		case SEQLOC_PNT:
			
			if((SeqIntPtr)anp->data.ptrvalue == NULL){
				throw "Unable to apply annotation offset";
			}
			
			( (SeqPntPtr)anp->data.ptrvalue )->point += m_offset;
			break;
		case SEQLOC_MIX:    /* mix -- more than one seq */
		case SEQLOC_EQUIV:    /* equiv -- ditto */
        	case SEQLOC_PACKED_INT:    /* packed int */

			sip = SeqLocId(anp);

			if (sip != NULL){      /* all on one Bioseq */

				SeqLocPtr slp = (SeqLocPtr)anp->data.ptrvalue;

				while (slp != NULL){
				
					add_loc_offset(slp, m_offset);	
					slp = slp->next;
				}
			}

			break;
	};
}

// Test the input file to determine if it is in either binary or ASCII format.
const char* GetAsnFileType(const string &name)
{
        ifstream fin(name.c_str());
 
        if(!fin.is_open()){
                throw "GetAsnFileType:Unable to open file for testing.";
        }
 
        const int sample_size = 15;
        char sample[sample_size];
 
        fin.read(sample, sample_size);
 
        if(fin.gcount() != sample_size){
	   
                throw "GetAsnFileType: Unable to read the required numnber of samples";
        }
 
        for(int i = 0;i < sample_size;i++){
	   
                // The method of detecting a binary file is not very sophisticated: look for
                // any negative 8 bit values (i.e. those values that do not map to ASCII
                // characters).
                if(sample[i] < 0){
			 
                        // We found a non-ASCII character
                        fin.close();
 
                        return "rb";
                }
        }
 
        fin.close();
 
        return "r";
}

string wrap_string(const string &m_str, const unsigned int &m_len)
{
	if(m_len == 0){
		throw __FILE__ ":wrap_string: m_len == 0";
	}
	
	stringstream sout;
	const unsigned int len = m_str.size();
	unsigned int count = 1;
	
	for(unsigned int i = 0;i < len;i++, count++){

		if( (count >= m_len) && isspace(m_str[i]) ){
		
			sout << ENDL;
			count = 1;
		}
		else{
			sout << m_str[i];
		}
	}
	
	return sout.str();
}

bool integer(const string &m_int)
{
	for(string::const_iterator iter = m_int.begin();iter != m_int.end();iter++){
		if( !isdigit(*iter) ){
			return false;
		}
	}
	
	return true;
}

string trim_space(const string &m_str)
{
	int len = m_str.size();
	int start = 0;
	
	if(len == 0){
		return "";
	}
	
	int stop = len - 1;
	
	while( (start < len) && isspace(m_str[start]) ){
		start++;
	}
	
	while( (start <= stop) && isspace(m_str[stop]) ){
		stop--;
	}
	
	return m_str.substr(start, stop - start + 1);
}

// Remove any end of line characters
string pack_string(const string &m_str)
{
	// Push back onto a list of char (instead of string) to retain
	// compatability with the old windows version of std::string
	// [which does not allow string::push_back()].
	list<char> buffer;
	char last_char = -1;
	
	for(string::const_iterator iter = m_str.begin();iter != m_str.end();iter++){
	
		if(*iter == '\n'){
		
			if( (last_char != '\n') && (last_char != '\r') ){
				buffer.push_back(' ');
			}
			
			last_char = *iter;
			continue;
		}
		
		if(*iter == '\r'){
		
			if( (last_char != '\n') && (last_char != '\r') ){
				buffer.push_back(' ');
			}
			
			last_char = *iter;
			continue;
		}

		buffer.push_back(*iter);
		
		last_char = *iter;
	}

	const unsigned int len = buffer.size();
	
	string tmp(len, 'A');
	unsigned int j = 0;
	
	for(list<char>::const_iterator i = buffer.begin();i != buffer.end();i++, j++){
		tmp[j] = *i;
	}
	
	return tmp;
}

static void touch_bioseq(BioseqPtr m_bsp, Pointer m_userdata)
{
	if(m_bsp == NULL){
		return;
	}
	
	if( (m_bsp->repr == Seq_repr_delta) || (m_bsp->repr == Seq_repr_seg) ){
		
		
		// If we're loading an ASN file from disk, we may need to go to NCBI over the 
		// net to get additional data
		if(is_fetch_enabled == false){
		
			is_fetch_enabled = PubSeqFetchEnable();
		}
		
		SeqMgrExploreSegments(m_bsp, NULL, load_bioseq_data);
   	} 
}

static Boolean LIBCALLBACK load_bioseq_data(
   SeqLocPtr m_slp, SeqMgrSegmentContextPtr m_context)
{
	BioseqPtr  bsp = NULL;
   	Uint2      entityID;
   	SeqLocPtr  loc;
   	SeqIdPtr   sip;
	Int4 	 from;
	Int4 	 to;
	SeqEntryPtr sep = NULL;
  
   	if( (m_slp == NULL) || (m_context == NULL) ){
		return FALSE;
	}

   	from = m_context->cumOffset;
   	to = from + m_context->to - m_context->from;

   	sip = SeqLocId(m_slp);
	
   	if(sip == NULL){
	
		loc = SeqLocFindNext(m_slp, NULL);
     	
		if(loc != NULL){
       		sip = SeqLocId(loc);
     	}
   	}
	
	if(sip == NULL){
   		return TRUE;
	}
	
	if(sip->choice != SEQID_GI){
		return TRUE;
	}
	
	// Remote fetch genome component if not already in memory
	sep = PubSeqSynchronousQuery(sip->data.intvalue, 0, 0);

	if( (sep == NULL) || (IS_Bioseq(sep) == false) ){
		return TRUE;
	}
	
	// Save this gi in a global list so we know to extract features from it
	global_far_sep.push_back(sep);
	
	//bsp = (BioseqPtr)sep->data.ptrvalue;
	
	bsp = BioseqLockById(sip);

   	if(bsp == NULL){
		return TRUE;
	}

   	entityID = ObjMgrGetEntityIDForPointer(bsp);

   	if(entityID != m_context->entityID){
		
		//if segment not packaged in record, may need to feature index it
		if(SeqMgrFeaturesAreIndexed(entityID) == 0){
			
			SeqMgrIndexFeatures(entityID, NULL);
		}
   	}
	
	// Correct all sequence features by adding the cumulative offset calculated above
	SeqEntryExplore(sep, (Pointer)(&from), ApplyOffsetCallback);	
		
   	return TRUE;
}

