// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "nuc_cruc.h"
#include <math.h>
#include <iostream>
#include <algorithm>
#include <vector>

using namespace std;

#if defined(WIN32) && !defined(max)
// needed to cope with bug in MS library:
// it fails to define min/max

template <class T> inline T max(const T& a, const T& b)
{
      return (a > b) ? a : b;
}

template <class T> inline T min(const T& a, const T& b)
{
      return (a < b) ? a : b;
}

#endif // WIN32

////////////////////////// Free energy parameterization //////////////////////////

float NucCruc::param_H[25][25];
float NucCruc::param_S[25][25];

float NucCruc::param_nucleation_H[25];
float NucCruc::param_nucleation_S[25];
		
bool NucCruc::param_init = false;
unsigned int NucCruc::parameter_set = NucCruc::SANTA_LUCIA;

// Enumerate all possible nearest-neighboor base stacking arrangements
#define	AA	0
#define	AC	1
#define	AG	2
#define	AT	3
#define	A_	4

#define	CA	5
#define	CC	6
#define	CG	7
#define	CT	8
#define	C_	9

#define	GA	10
#define	GC	11
#define	GG	12
#define	GT	13
#define	G_	14

#define	TA	15
#define	TC	16
#define	TG	17
#define	TT	18
#define	T_	19

// OS X has macros names _A, _C and _G
#ifdef _A
#undef _A
#endif

#ifdef _C
#undef _C
#endif

#ifdef _G
#undef _G
#endif

// Windows has a macro named _T
#ifdef _T
#undef _T
#endif

#define	_A	20
#define	_C	21
#define	_G	22

#define	_T	23
#define	__	24

#define	ALLAWI_SANTALUCIA_SYMMETRY_DS			(-1.4e-3f)
#define	ALLAWI_SANTALUCIA_SALT				(0.368e-3f)

// Note that as of 4/11/04 there are no published double mismatch
// parameters for DNA/DNA duplexes.
// Mismatch parameters have been estimated by the "HyTher (tm)" web site:
// http://ozone2.chem.wayne.edu
// and these could be extracted by feeding the site test sequences. For now,
// assume that double mismatches do not enthalpically contribute to Delta G.
#define	ALLAWI_SANTALUCIA_DOUBLE_MISMATCH_H 	0.0f		// <-- No enthalpic penalty
#define	ALLAWI_SANTALUCIA_DOUBLE_MISMATCH_S 	-5.0e-3f  // <-- Small entropic penalty

NucCruc::NucCruc(const unsigned int &m_param_set /*= SANTA_LUCIA*/)
{
	switch(m_param_set){
		case SANTA_LUCIA:
			init_param_AS();
			break;
		case PRIMER3:
			init_param_Primer3();
			break;
		default:
			throw "Unknowm parameter set!";
	};
	
	min_len = 2;
	max_len = 75;
	
	// The number of bases required to form a hairpin loop
	hairpin_loop_size = 3;
	
	na_concentration = NC_Na;
	strand_concentration = NC_CT;
	
	// By default, always compute Tm
	identity_cutoff = -1.0f;
}

void NucCruc::init_param_AS()
{
	// Have the parameters already been initialized and
	// is the parameter set == SANTA_LUCIA? If not, we
	// need to initialize
	if( param_init && (parameter_set == SANTA_LUCIA) ){
		return;
	}
	
	unsigned int i;
	unsigned int j;
	
	// Initialize all parameter to default values
	for(i = 0;i < NUM_BASE_PAIR;i++){
	
		param_nucleation_H[i] = param_nucleation_S[i] = 0.0f;
		
		for(j = 0;j < NUM_BASE_PAIR;j++){
			param_H[i][j] = ALLAWI_SANTALUCIA_DOUBLE_MISMATCH_H;
			param_S[i][j] = ALLAWI_SANTALUCIA_DOUBLE_MISMATCH_S;
		}
	}
	
	// Exact nearest-neighboor matches
	// SantaLucia, PNAS 95: 1460-1465
	param_H[AT][AT] = -7.9f; 	// AA/TT		Enthalpic
	param_H[AT][CG] = -8.4f; 	// AC/TG
	param_H[AT][GC] = -7.8f; 	// AG/TC
	param_H[AT][TA] = -7.2f; 	// AT/TA
	param_H[CG][AT] = -8.5f; 	// CA/GT
	param_H[CG][CG] = -8.0f; 	// CC/GG
	param_H[CG][GC] = -10.6f;	// CG/GC
	param_H[CG][TA] = -7.8f; 	// CT/GA
	param_H[GC][AT] = -8.2f;		// GA/CT
	param_H[GC][CG] = -9.8f;		// GC/CG
	param_H[GC][GC] = -8.0f;		// GG/CC
	param_H[GC][TA] = -8.4f;		// GT/CA
	param_H[TA][AT] = -7.2f;		// TA/AT
	param_H[TA][CG] = -8.2f;		// TC/AG
	param_H[TA][GC] = -8.5f;		// TG/AC
	param_H[TA][TA] = -7.9f;		// TT/AA
	
	param_H[A_][AT] = param_H[TA][_A] = 0.2f;	// AA/_T
	param_H[A_][CG] = param_H[GC][_A] = -6.3f;	// AC/_G
	param_H[A_][GC] = param_H[CG][_A] = -3.7f;	// AG/_C
	param_H[A_][TA] = param_H[AT][_A] = -2.9f;	// AT/_A
	param_H[C_][AT] = param_H[TA][_C] = 0.6f;	// CA/_T
	param_H[C_][CG] = param_H[GC][_C] = -4.4f;	// CC/_G
	param_H[C_][GC] = param_H[CG][_C] = -4.0f;	// CG/_C
	param_H[C_][TA] = param_H[AT][_C] = -4.1f;	// CT/_A
	param_H[G_][AT] = param_H[TA][_G] = -1.1f;	// GA/_T
	param_H[G_][CG] = param_H[GC][_G] = -5.1f;	// GC/_G
	param_H[G_][GC] = param_H[CG][_G] = -3.9f;	// GG/_C
	param_H[G_][TA] = param_H[AT][_G] = -4.2f;	// GT/_A
	param_H[T_][AT] = param_H[TA][_T] = -6.9f;	// TA/_T
	param_H[T_][CG] = param_H[GC][_T] = -4.0f;	// TC/_G
	param_H[T_][GC] = param_H[CG][_T] = -4.9f;	// TG/_C
	param_H[T_][TA] = param_H[AT][_T] = -0.2f;	// TT/_A
	
	param_H[_A][AT] = param_H[TA][A_] = -0.7f;	// _A/AT
	param_H[_A][CG] = param_H[GC][A_] = -2.1f;	// _C/AG
	param_H[_A][GC] = param_H[CG][A_] = -5.9f;	// _G/AC
	param_H[_A][TA] = param_H[AT][A_] = -0.5f;	// _T/AA
	param_H[_C][AT] = param_H[TA][C_] = 4.4f;	// _A/CT
	param_H[_C][CG] = param_H[GC][C_] = -0.2f;	// _C/CG
	param_H[_C][GC] = param_H[CG][C_] = -2.6f;	// _G/CC
	param_H[_C][TA] = param_H[AT][C_] = 4.7f;	// _T/CA
	param_H[_G][AT] = param_H[TA][G_] = -1.6f;	// _A/GT
	param_H[_G][CG] = param_H[GC][G_] = -3.9f;	// _C/GG
	param_H[_G][GC] = param_H[CG][G_] = -3.2f;	// _G/GC
	param_H[_G][TA] = param_H[AT][G_] = -4.1f;	// _T/GA
	param_H[_T][AT] = param_H[TA][A_] = 2.9f;	// _A/TT
	param_H[_T][CG] = param_H[GC][T_] = -4.4f;	// _C/TG
	param_H[_T][GC] = param_H[CG][T_] = -5.2f;	// _G/TC
	param_H[_T][TA] = param_H[AT][T_] = -3.8f;	// _T/TA
	
	param_S[A_][AT] = param_S[TA][_A] = 2.3e-3f;		// AA/_T
	param_S[A_][CG] = param_S[GC][_A] = -17.1e-3f;	// AC/_G
	param_S[A_][GC] = param_S[CG][_A] = -10.0e-3f;	// AG/_C
	param_S[A_][TA] = param_S[AT][_A] = -7.6e-3f;	// AT/_A
	param_S[C_][AT] = param_S[TA][_C] = 3.3e-3f;		// CA/_T
	param_S[C_][CG] = param_S[GC][_C] = -12.6e-3f;	// CC/_G
	param_S[C_][GC] = param_S[CG][_C] = -11.9e-3f;	// CG/_C
	param_S[C_][TA] = param_S[AT][_C] = -13.0e-3f;	// CT/_A
	param_S[G_][AT] = param_S[TA][_G] = -1.6e-3f;	// GA/_T
	param_S[G_][CG] = param_S[GC][_G] = -14.0e-3f;	// GC/_G
	param_S[G_][GC] = param_S[CG][_G] = -10.9e-3f;	// GG/_C
	param_S[G_][TA] = param_S[AT][_G] = -15.0e-3f;	// GT/_A
	param_S[T_][AT] = param_S[TA][_T] = -20.0e-3f;	// TA/_T
	param_S[T_][CG] = param_S[GC][_T] = -10.9e-3f;	// TC/_G
	param_S[T_][GC] = param_S[CG][_T] = -13.8e-3f;	// TG/_C
	param_S[T_][TA] = param_S[AT][_T] = -0.5e-3f;	// TT/_A
	
	param_S[_A][AT] = param_S[TA][A_] = -0.8e-3f;	// _A/AT
	param_S[_A][CG] = param_S[GC][A_] = -3.9e-3f;	// _C/AG
	param_S[_A][GC] = param_S[CG][A_] = -16.5e-3f;	// _G/AC
	param_S[_A][TA] = param_S[AT][A_] = -1.1e-3f;	// _T/AA
	param_S[_C][AT] = param_S[TA][C_] = 14.9e-3f;	// _A/CT
	param_S[_C][CG] = param_S[GC][C_] = -0.1e-3f;	// _C/CG
	param_S[_C][GC] = param_S[CG][C_] = -7.4e-3f;	// _G/CC
	param_S[_C][TA] = param_S[AT][C_] = 14.2e-3f;	// _T/CA
	param_S[_G][AT] = param_S[TA][G_] = -3.6e-3f;	// _A/GT
	param_S[_G][CG] = param_S[GC][G_] = -11.2e-3f;	// _C/GG
	param_S[_G][GC] = param_S[CG][G_] = -10.4e-3f;	// _G/GC
	param_S[_G][TA] = param_S[AT][G_] = -13.1e-3f;	// _T/GA
	param_S[_T][AT] = param_S[TA][A_] = 10.4e-3f;	// _A/TT
	param_S[_T][CG] = param_S[GC][T_] = -13.1e-3f;	// _C/TG
	param_S[_T][GC] = param_S[CG][T_] = -15.0e-3f;	// _G/TC
	param_S[_T][TA] = param_S[AT][T_] = -12.6e-3f;	// _T/TA
	
	param_S[AT][AT] = -22.2e-3f;	// AA/TT		Entropic
	param_S[AT][CG] = -22.4e-3f;	// AC/TG
	param_S[AT][GC] = -21.0e-3f;	// AG/TC
	param_S[AT][TA] = -20.4e-3f;	// AT/TA
	param_S[CG][AT] = -22.7e-3f;	// CA/GT
	param_S[CG][CG] = -19.9e-3f;	// CC/GG
	param_S[CG][GC] = -27.2e-3f;	// CG/GC
	param_S[CG][TA] = -21.0e-3f;	// CT/GA
	param_S[GC][AT] = -22.2e-3f;	// GA/CT
	param_S[GC][CG] = -24.4e-3f;	// GC/CG
	param_S[GC][GC] = -19.9e-3f;	// GG/CC
	param_S[GC][TA] = -22.4e-3f;	// GT/CA
	param_S[TA][AT] = -21.3e-3f;	// TA/AT
	param_S[TA][CG] = -22.2e-3f;	// TC/AG
	param_S[TA][GC] = -22.7e-3f;	// TG/AC
	param_S[TA][TA] = -22.2e-3f;	// TT/AA
	
	// Single G-A mismatch
	// Allawi et. al. Biochemistry 1998, 37, 2170-2179
	param_H[AT][AG] = param_H[GA][TA] = -0.6f;	// Aa/Tg		Enthalpic
	param_H[AT][GA] = param_H[AG][TA] = -0.7f;	// Ag/Ta
	param_H[CG][AG] = param_H[GA][GC] = -0.7f;	// Ca/Gg
	param_H[CG][GA] = param_H[AG][GC] = -4.0f;	// Cg/Ga
	param_H[GC][AG] = param_H[GA][CG] = -0.6f;	// Ga/Cg
	param_H[GC][GA] = param_H[AG][CG] = 0.5f;	// Gg/Ca
	param_H[TA][AG] = param_H[GA][AT] = 0.7f;	// Ta/Ag
	param_H[TA][GA] = param_H[AG][AT] = 3.0f;	// Tg/Aa
	
	param_S[AT][AG] = param_S[GA][TA] = -2.3e-3f;	// Aa/Tg		Entropic
	param_S[AT][GA] = param_S[AG][TA] = -2.3e-3f;	// Ag/Ta
	param_S[CG][AG] = param_S[GA][GC] = -2.3e-3f;	// Ca/Gg
	param_S[CG][GA] = param_S[AG][GC] = -13.2e-3f;	// Cg/Ga
	param_S[GC][AG] = param_S[GA][CG] = -1.0e-3f;	// Ga/Cg
	param_S[GC][GA] = param_S[AG][CG] = 3.2e-3f;		// Gg/Ca
	param_S[TA][AG] = param_S[GA][AT] = 0.7e-3f;		// Ta/Ag
	param_S[TA][GA] = param_S[AG][AT] = 7.4e-3f;		// Tg/Aa
	
	// Single C-T mismatch
	// Allawi et. al. Nucleic Acids Research, 1998, 26, 11, 2694-2701
	param_H[AT][CT] = param_H[TC][TA] = 0.7f;	// Ac/Tt		Enthalpic
	param_H[AT][TC] = param_H[CT][TA] = -1.2f;	// At/Tc
	param_H[CG][CT] = param_H[TC][GC] = -0.8f;	// Cc/Gt
	param_H[CG][TC] = param_H[CT][GC] = -1.5f;	// Ct/Gc
	param_H[GC][CT] = param_H[TC][CG] = 2.3f;	// Gc/Ct
	param_H[GC][TC] = param_H[CT][CG] = 5.2f;	// Gt/Cc
	param_H[TA][CT] = param_H[TC][AT] = 1.2f;	// Tc/At
	param_H[TA][TC] = param_H[CT][AT] = 1.0f;	// Tt/Ac
	
	param_S[AT][CT] = param_S[TC][TA] = 0.2e-3f;	// Ac/Tt		Entropic
	param_S[AT][TC] = param_S[CT][TA] = -6.2e-3f;// At/Tc
	param_S[CG][CT] = param_S[TC][GC] = -4.5e-3f;// Cc/Gt
	param_S[CG][TC] = param_S[CT][GC] = -6.1e-3f;// Ct/Gc
	param_S[GC][CT] = param_S[TC][CG] = 5.4e-3f;	// Gc/Ct
	param_S[GC][TC] = param_S[CT][CG] = 13.5e-3f;// Gt/Cc
	param_S[TA][CT] = param_S[TC][AT] = 0.7e-3f;	// Tc/At
	param_S[TA][TC] = param_S[CT][AT] = 0.7e-3f;	// Tt/Ac
	
	// Single A-C mismatch
	// Allawi et. al. Biochemistry 1998, 37, 9435-9444
	param_H[AT][AC] = param_H[CA][TA] = 2.3f; 	// Aa/Tc		Enthalpic
	param_H[AT][CA] = param_H[AC][TA] = 5.3f;	// Ac/Ta
	param_H[CG][AC] = param_H[CA][GC] = 1.9f;	// Ca/Gc
	param_H[CG][CA] = param_H[AC][GC] = 0.6f;	// Cc/Ga
	param_H[GC][AC] = param_H[CA][CG] = 5.2f;	// Ga/Cc
	param_H[GC][CA] = param_H[AC][CG] = -0.7f;	// Gc/Ca
	param_H[TA][AC] = param_H[CA][AT] = 3.4f;	// Ta/Ac
	param_H[TA][CA] = param_H[AC][AT] = 7.6f;	// Tc/Aa
	
	param_S[AT][AC] = param_S[CA][TA] = 4.6e-3f; // Aa/Tc		Entropic
	param_S[AT][CA] = param_S[AC][TA] = 14.6e-3f;// Ac/Ta
	param_S[CG][AC] = param_S[CA][GC] = 3.7e-3f;	// Ca/Gc
	param_S[CG][CA] = param_S[AC][GC] = -0.6e-3f;// Cc/Ga
	param_S[GC][AC] = param_S[CA][CG] = 14.2e-3f;// Ga/Cc
	param_S[GC][CA] = param_S[AC][CG] = -3.8e-3f;// Gc/Ca
	param_S[TA][AC] = param_S[CA][AT] = 8.0e-3f;	// Ta/Ac
	param_S[TA][CA] = param_S[AC][AT] = 20.2e-3f;// Tc/Aa
	
	// Single G-T mismatch
	// Allawi et. al. Biochemistry 1997, 36, 10581-10594
	param_H[AT][GT] = param_H[TG][TA] = 1.0f;	// Ag/Tt		Enthalpic
	param_H[AT][TG] = param_H[GT][TA] = -2.5f;	// At/Tg
	param_H[CG][GT] = param_H[TG][GC] = -4.1f;	// Cg/Gt
	param_H[CG][TG] = param_H[GT][GC] = -2.8f;	// Ct/Gg
	param_H[GC][GT] = param_H[TG][CG] = 3.3f;	// Gg/Ct
	param_H[GT][GT] = param_H[TG][TG] = 5.8f;	// gg/tt <-- double mismatch!
	param_H[GC][TG] = param_H[GT][CG] = -4.4f;	// Gt/Cg
	param_H[GT][TG] 			    = 4.1f;	// gt/tg <-- double mismatch!
	param_H[TA][GT] = param_H[TG][AT] = -0.1f;	// Tg/At
	param_H[TG][GT] 			    = -1.4f;	// tg/gt <-- double mismatch!
	param_H[TA][TG] = param_H[GT][AT] = -1.3f;	// Tt/Ag
	
	param_S[AT][GT] = param_S[TG][TA] = 0.9e-3f;		// Ag/Tt		Entropic
	param_S[AT][TG] = param_S[GT][TA] = -8.3e-3f;	// At/Tg
	param_S[CG][GT] = param_S[TG][GC] = -11.7e-3f;	// Cg/Gt
	param_S[CG][TG] = param_S[GT][GC] = -8.0e-3f;	// Ct/Gg
	param_S[GC][GT] = param_S[TG][CG] = 10.4e-3f;	// Gg/Ct
	param_S[GT][GT] = param_S[TG][TG] = 16.3e-3f;	// gg/tt <-- double mismatch!
	param_S[GC][TG] = param_S[GT][CG] = -12.3e-3f;	// Gt/Cg
	param_S[GT][TG] 			    = 9.5e-3f;		// gt/tg <-- double mismatch!
	param_S[TA][GT] = param_S[TG][AT] = -1.7e-3f;	// Tg/At
	param_S[TG][GT] 			    = -6.2e-3f;	// tg/gt <-- double mismatch!
	param_S[TA][TG] = param_S[GT][AT] = -5.3e-3f;	// Tt/Ag
	
	// Single A-A, C-C, G-G or T-T mismatch
	// Peyret et. al. Biochemistry 1999, 38, 3468-3477
	param_H[AT][AA] = param_H[AA][TA] = 1.2f;	// Aa/Ta		Enthalpic
	param_H[CG][AA] = param_H[AA][GC] = -0.9f;	// Ca/Ga
	param_H[GC][AA] = param_H[AA][CG] = -2.9f;	// Ga/Ca
	param_H[TA][AA] = param_H[AA][AT] = 4.7f;	// Ta/Aa
	
	param_H[AT][CC] = param_H[CC][TA] = 0.0f;	// Ac/Tc
	param_H[CG][CC] = param_H[CC][GC] = -1.5f;	// Cc/Gc
	param_H[GC][CC] = param_H[CC][CG] = 3.6f;	// Gc/Cc
	param_H[TA][CC] = param_H[CC][AT] = 6.1f;	// Tc/Ac
	
	param_H[AT][GG] = param_H[GG][TA] = -3.1f;	// Ag/Tg
	param_H[CG][GG] = param_H[GG][GC] = -4.9f;	// Cg/Gg
	param_H[GC][GG] = param_H[GG][CG] = -6.0f;	// Gg/Cg
	param_H[TA][GG] = param_H[GG][AT] = 1.6f;	// Tg/Ag
	
	param_H[AT][TT] = param_H[TT][TA] = -2.7f;	// At/Tt
	param_H[CG][TT] = param_H[TT][GC] = -5.0f;	// Ct/Gt
	param_H[GC][TT] = param_H[TT][CG] = -2.2f;	// Gt/Ct
	param_H[TA][TT] = param_H[TT][AT] = 0.2f;	// Tt/At
	
	param_S[AT][AA] = param_S[AA][TA] = 1.7e-3f;	// Aa/Ta		Entropic
	param_S[CG][AA] = param_S[AA][GC] = -4.2e-3f;	// Ca/Ga
	param_S[GC][AA] = param_S[AA][CG] = -9.8e-3f;	// Ga/Ca
	param_S[TA][AA] = param_S[AA][AT] = 12.9e-3f;	// Ta/Aa
	
	param_S[AT][CC] = param_S[CC][TA] = -4.4e-3f;	// Ac/Tc
	param_S[CG][CC] = param_S[CC][GC] = -7.2e-3f;	// Cc/Gc
	param_S[GC][CC] = param_S[CC][CG] = 8.9e-3f;	// Gc/Cc
	param_S[TA][CC] = param_S[CC][AT] = 16.4e-3f;	// Tc/Ac
	
	param_S[AT][GG] = param_S[GG][TA] = -9.5e-3f;	// Ag/Tg
	param_S[CG][GG] = param_S[GG][GC] = -15.3e-3f;	// Cg/Gg
	param_S[GC][GG] = param_S[GG][CG] = -15.8e-3f;	// Gg/Cg
	param_S[TA][GG] = param_S[GG][AT] = 3.6e-3f;	// Tg/Ag
	
	param_S[AT][TT] = param_S[TT][TA] = -10.8e-3f;	// At/Tt
	param_S[CG][TT] = param_S[TT][GC] = -15.8e-3f;	// Ct/Gt
	param_S[GC][TT] = param_S[TT][CG] = -8.4e-3f;	// Gt/Ct
	param_S[TA][TT] = param_S[TT][AT] = -1.5e-3f;	// Tt/At
	
	// Nucleation parameters
	param_nucleation_H[AT] = param_nucleation_H[TA] = 2.3f;
	param_nucleation_S[AT] = param_nucleation_S[TA] = 4.1e-3f;
	
	param_nucleation_H[GC] = param_nucleation_H[CG] = 0.1f;
	param_nucleation_S[GC] = param_nucleation_S[CG] = -2.8e-3f;
	
	param_init = true;
	
	parameter_set = SANTA_LUCIA;
}

void NucCruc::init_param_Primer3()
{
	// Have the parameters already been initialized and
	// is the parameter set == PRIMER3? If not, we
	// need to initialize
	if( param_init && (parameter_set == PRIMER3) ){
		return;
	}
	
	// Start with the SantaLucia parameter set (since primer3
	// does not provide mismatch paramters).
	init_param_AS();
	
	// Compute a correction factor to map the SantaLucia parameters
	// to Primer3 parameters
	float c_h = 0.0;
	float c_s = 0.0;
	
	const float H_A_A = -9.1f;
	const float H_A_C = -6.5f;
	const float H_A_G = -7.8f;
	const float H_A_T = -8.6f;

	const float H_C_A = -5.8f;
	const float H_C_C = -11.0f;
	const float H_C_G = -11.9f;
	const float H_C_T = -7.8f;

	const float H_G_A = -5.6f;
	const float H_G_C = -11.1f;
	const float H_G_G = -11.0f;
	const float H_G_T = -6.5f;

	const float H_T_A = -6.0f;
	const float H_T_C = -5.6f;
	const float H_T_G = -5.8f;
	const float H_T_T = -9.1f;

	c_h += param_H[AT][AT]/H_A_A; // AA/TT		Enthalpic
	c_h += param_H[AT][CG]/H_A_C; // AC/TG
	c_h += param_H[AT][GC]/H_A_G; // AG/TC
	c_h += param_H[AT][TA]/H_A_T; // AT/TA
	c_h += param_H[CG][AT]/H_C_A; // CA/GT
	c_h += param_H[CG][CG]/H_C_C; // CC/GG
	c_h += param_H[CG][GC]/H_C_G;	// CG/GC
	c_h += param_H[CG][TA]/H_C_T; // CT/GA
	c_h += param_H[GC][AT]/H_G_A;	// GA/CT
	c_h += param_H[GC][CG]/H_G_C;	// GC/CG
	c_h += param_H[GC][GC]/H_G_G;	// GG/CC
	c_h += param_H[GC][TA]/H_G_T;	// GT/CA
	c_h += param_H[TA][AT]/H_T_A;	// TA/AT
	c_h += param_H[TA][CG]/H_T_C;	// TC/AG
	c_h += param_H[TA][GC]/H_T_G;	// TG/AC
	c_h += param_H[TA][TA]/H_T_T;	// TT/AA
	
	c_h /= 16.0f;
	
	const float S_A_A = -24.0e-3f;
	const float S_A_C = -17.3e-3f;
	const float S_A_G = -20.8e-3f;
	const float S_A_T = -23.9e-3f;

	const float S_C_A = -12.9e-3f;
	const float S_C_C = -26.6e-3f;
	const float S_C_G = -27.8e-3f;
	const float S_C_T = -20.8e-3f;

	const float S_G_A = -13.5e-3f;
	const float S_G_C = -26.7e-3f;
	const float S_G_G = -26.6e-3f;
	const float S_G_T = -17.3e-3f;

	const float S_T_A = -16.9e-3f;
	const float S_T_C = -13.5e-3f;
	const float S_T_G = -12.9e-3f;
	const float S_T_T = -24.0e-3f;

	c_s += param_S[AT][AT]/S_A_A; // AA/TT		Entropic
	c_s += param_S[AT][CG]/S_A_C; // AC/TG
	c_s += param_S[AT][GC]/S_A_G; // AG/TC
	c_s += param_S[AT][TA]/S_A_T; // AT/TA
	c_s += param_S[CG][AT]/S_C_A; // CA/GT
	c_s += param_S[CG][CG]/S_C_C; // CC/GG
	c_s += param_S[CG][GC]/S_C_G;	// CG/GC
	c_s += param_S[CG][TA]/S_C_T; // CT/GA
	c_s += param_S[GC][AT]/S_G_A;	// GA/CT
	c_s += param_S[GC][CG]/S_G_C;	// GC/CG
	c_s += param_S[GC][GC]/S_G_G;	// GG/CC
	c_s += param_S[GC][TA]/S_G_T;	// GT/CA
	c_s += param_S[TA][AT]/S_T_A;	// TA/AT
	c_s += param_S[TA][CG]/S_T_C;	// TC/AG
	c_s += param_S[TA][GC]/S_T_G;	// TG/AC
	c_s += param_S[TA][TA]/S_T_T;	// TT/AA
	
	c_s /= 16.0f;
	
	unsigned int i;
	unsigned int j;
	
	// Initialize all parameter to default values
	for(i = 0;i < NUM_BASE_PAIR;i++){
	
		param_nucleation_H[i] = param_nucleation_S[i] = 0.0f;
		
		for(j = 0;j < NUM_BASE_PAIR;j++){
			param_H[i][j] *= c_h;
			param_S[i][j] *= c_s;
		}
	}
	
	// Make sure that we agree with primer3 for exact DNA matches
	
	param_H[AT][AT] = H_A_A; // AA/TT		Enthalpic
	param_H[AT][CG] = H_A_C; // AC/TG
	param_H[AT][GC] = H_A_G; // AG/TC
	param_H[AT][TA] = H_A_T; // AT/TA
	param_H[CG][AT] = H_C_A; // CA/GT
	param_H[CG][CG] = H_C_C; // CC/GG
	param_H[CG][GC] = H_C_G;	// CG/GC
	param_H[CG][TA] = H_C_T; // CT/GA
	param_H[GC][AT] = H_G_A;	// GA/CT
	param_H[GC][CG] = H_G_C;	// GC/CG
	param_H[GC][GC] = H_G_G;	// GG/CC
	param_H[GC][TA] = H_G_T;	// GT/CA
	param_H[TA][AT] = H_T_A;	// TA/AT
	param_H[TA][CG] = H_T_C;	// TC/AG
	param_H[TA][GC] = H_T_G;	// TG/AC
	param_H[TA][TA] = H_T_T;	// TT/AA
	
	param_S[AT][AT] = S_A_A; // AA/TT		Entropic
	param_S[AT][CG] = S_A_C; // AC/TG
	param_S[AT][GC] = S_A_G; // AG/TC
	param_S[AT][TA] = S_A_T; // AT/TA
	param_S[CG][AT] = S_C_A; // CA/GT
	param_S[CG][CG] = S_C_C; // CC/GG
	param_S[CG][GC] = S_C_G;	// CG/GC
	param_S[CG][TA] = S_C_T; // CT/GA
	param_S[GC][AT] = S_G_A;	// GA/CT
	param_S[GC][CG] = S_G_C;	// GC/CG
	param_S[GC][GC] = S_G_G;	// GG/CC
	param_S[GC][TA] = S_G_T;	// GT/CA
	param_S[TA][AT] = S_T_A;	// TA/AT
	param_S[TA][CG] = S_T_C;	// TC/AG
	param_S[TA][GC] = S_T_G;	// TG/AC
	param_S[TA][TA] = S_T_T;	// TT/AA
	
	param_init = true;
	parameter_set = PRIMER3;
}

bool NucCruc::is_exact_match(const unsigned int &m_base_pair) const
{
	switch(m_base_pair){
		case AT:
		case TA:
		case GC:
		case CG:
		
		// The following pairs are considered exact matches for the sake
		// starting and stopping the binding interaction (in the sense that
		// there are parameters defining the interaction).
		case _A: 
		case _T:
		case _G:
		case _C:
		case A_:
		case T_:
		case G_:
		case C_:
		
			return true;
	};
	
	return false;
}

float NucCruc::Tm(const string &m_seq, pair<unsigned int, unsigned int> *m_align,
	const float &m_DH_bias, const float &m_DS_bias) const
{
	unsigned int len = m_seq.size();
	
	// Pack the duplex
	list< pair<unsigned char, unsigned char> > duplex;
	
	for(unsigned int i = 0;i < len;i++){
		unsigned char base = CharToBase(m_seq[i]);
		
		if(base >= GAP){
			throw __FILE__ ":NucCruc::Tm: Illegal base!";
		}
		
		duplex.push_back( pair<unsigned char, unsigned char>(base, complement(base)) );
	}
	
	len = duplex.size();
	
	if(len < min_len){
		throw __FILE__ ":NucCruc::Tm: Input sequence is less than minimum length";
	}

	if(len > max_len){
		throw __FILE__ ":NucCruc::Tm: Input sequence is greater than maximum length";
	}
	
	return Tm(duplex, m_align, m_DH_bias, m_DS_bias);
}

float NucCruc::Tm(const string &m_seq_a, const string &m_seq_b, const bool &m_complement_strand_b,
	pair<unsigned int, unsigned int> *m_align,
	const float &m_DH_bias, const float &m_DS_bias) const
{
	unsigned int len = m_seq_a.size();
	
	if(m_seq_b.size() != len){
		throw __FILE__ ":NucCruc::Tm: Sequence lengths are not equal!";
	}
		
	// Pack the duplex
	list< pair<unsigned char, unsigned char> > duplex;
	
	for(unsigned int i = 0;i < len;i++){
		unsigned char base_a = CharToBase(m_seq_a[i]);
		unsigned char base_b;
		
		if(base_a > GAP){
			throw __FILE__ ":NucCruc::Tm: Illegal base!";
		}
		
		if(m_complement_strand_b){
			base_b = CharToBaseComplement(m_seq_b[i]);
		}
		else{
			base_b = CharToBase(m_seq_b[i]);
		}
		
		if(base_b > GAP){
			throw __FILE__ ":NucCruc::Tm: Illegal base!";
		}
		
		duplex.push_back( pair<unsigned char, unsigned char>(base_a, base_b) );
	}
	
	len = duplex.size();
	
	if(len < min_len){
		throw __FILE__ ":NucCruc::Tm: Input sequence is less than minimum length";
	}

	if(len > max_len){
		throw __FILE__ ":NucCruc::Tm: Input sequence is greater than maximum length";
	}
	
	return Tm(duplex, m_align, m_DH_bias, m_DS_bias);
}

// Tm = (dH_d + dH_n + dH_s)/(Rln{[C_T]/alpha} + 
//	dS_d + dS_n + dS_s + dS_salt)
float NucCruc::Tm(const list< pair<unsigned char, unsigned char> > &m_duplex,
	pair<unsigned int, unsigned int> *m_align,
	const float &m_DH_bias, const float &m_DS_bias) const
{		
	// Find the duplex substring that has the best free energy (most stable). This
	// is equivalent to dynamic programming on a diagonal matrix (since we do not currently
	// allow gaps, there is no point to incurring the cost of the full matrix dynamic
	// programming rigmarol).
	list< pair<unsigned char, unsigned char> >::const_iterator base_iter;
	list< pair<unsigned char, unsigned char> >::const_iterator left_iter;
	list< pair<unsigned char, unsigned char> >::const_iterator right_iter;
	
	int last_base_pair = -1;
	int cur_base_pair = -1;
	
	// Compute the cumulative free energy using a reference temperature of 37 degrees.
	// This is equivalent to finding the best LOCAL alignment on the diagonal element 
	// of the dynamic programming matrix.
	float c_dG = 0.0f;
	
	// Also save the value of the min c_dG and its location
	float min_c_dG = 99999999.9f;
	
	// Compute all di-nucleotide free energy parameters. We will then extract the
	// the most stable subset for actual Tm calculation. Note that this caluclation
	// does not take into account the initiation energies due to terminal A/T or G/C
	// base pairs.
	
	for(base_iter = m_duplex.begin();base_iter != m_duplex.end();base_iter++){
		
		last_base_pair = cur_base_pair;
		cur_base_pair = base_iter->first*5 + base_iter->second;
		
		if(last_base_pair < 0){
			continue;
		}
		
		c_dG += param_H[last_base_pair][cur_base_pair]
				- (37.0f + NC_ZERO_C)*param_S[last_base_pair][cur_base_pair];
		
		// Use a local, not global alignment
		c_dG = min(c_dG, 0.0f);
		
		if(c_dG < min_c_dG){
			min_c_dG = c_dG;
			right_iter = base_iter;			
		}		
	}
	
	// Once we have identified the location and value of the minimum cumulative
	// delta G, back track to find the location of the minimum starting from the right.
	
	// Reset the current base pair
	cur_base_pair = -1;
	c_dG = 0.0f;
	
	base_iter = right_iter;
	
	min_c_dG = 99999999.9f;
	
	// Count the number of duplex pairs
	unsigned int num_pair = 0;
	unsigned int count = 0;
	
	// Do we need to compute the sequence identity?
	const bool compute_identity = (identity_cutoff > 0.0f);
	unsigned int num_ident = 0;
	
	while(true){
		
		count ++;
		
		last_base_pair = cur_base_pair;
		cur_base_pair = base_iter->first*5 + base_iter->second;
		
		if(compute_identity){
			// Compute the sequence identity
			num_ident += is_exact_match(cur_base_pair);
		}
		
		if(last_base_pair < 0){
			if(base_iter == m_duplex.begin()){
				// Too few bases to compute Tm
				return -NC_ZERO_C;
			}
		
			base_iter --;
			continue;
		}
		
		// ************************** NOTE! **************************
		// Since we are traversing the duplex in reverse, we must reverse the
		// order of the array indexing:
		// [last_base_pair][cur_base_pair] -- [cur_base_pair][last_base_pair]
		c_dG += param_H[cur_base_pair][last_base_pair]
				- (37.0f + NC_ZERO_C)*param_S[cur_base_pair][last_base_pair];
		
		if(c_dG < min_c_dG){
			min_c_dG = c_dG;
			left_iter = base_iter;
			num_pair = count;
		}
		
		if(base_iter == m_duplex.begin()){
			break;
		}
		
		base_iter --;
	}
	
	// Are we prefiltering the Tm calculation by sequence identity?
	if( compute_identity && (num_ident < num_pair*identity_cutoff) ){
		// Insufficent sequence identity to warrent a full Tm calculation
		return -NC_ZERO_C;
	}
		
	// The parameterization in terms of di-nucleotide interactions
	// doesn't make sense for less than 3 bases since the flanking bases could
	// be mismatches -- which would cause trouble when trying to compute the 
	// nucleation energy.
	if(num_pair <= 3){
		return -NC_ZERO_C;
	}
	
	// The di-nulceotide contribution to enthalpy and entropy.
	// Incorporate user supplied biases
	float dH_d = m_DH_bias;
	float dS_d = m_DS_bias + (parameter_set == PRIMER3) ? -10.8e-3f : 0.0f;
		
	if(m_align){
		m_align->second = m_align->first + num_pair - 1;
	}
		
	// Track the number of each kind of base
	bool is_self_complement = true;
	
	bool initiation_53 = false;
	bool initiation_35 = false;
	
	unsigned int num_base = 0;
	
	// Reset the current base pair
	cur_base_pair = -1;
	
	for(unsigned int i = 0;i < num_pair;i++, left_iter++, right_iter--){
		// Is this sequence self-complementary? By default, it is until proven
		// otherwise
		if(is_self_complement){
			
			is_self_complement = (left_iter->first == right_iter->second) && 
				(left_iter->second == right_iter->first);
		}
		
		last_base_pair = cur_base_pair;
		cur_base_pair = left_iter->first*5 + left_iter->second;
		
		// Compute the initiation energy. Consider the first terminal base pairs
		// that make valid A/T or G/C pairs (mismatches and gaps are ignored).
		if(!initiation_53 && is_exact_match(cur_base_pair)){			
			dH_d += param_nucleation_H[cur_base_pair];
			dS_d += param_nucleation_S[cur_base_pair];
			
			initiation_53 = true;
		}
		
		if(!initiation_35){
			int tmp_index = right_iter->second*5 + right_iter->first;
			
			if(is_exact_match(tmp_index)){		
				dH_d += param_nucleation_H[tmp_index];
				dS_d += param_nucleation_S[tmp_index];
				
				initiation_35 = true;
			}
		}
		
		// Count the number of bases
		num_base += (left_iter->first < GAP);
		num_base += (left_iter->second < GAP);
		
		if(last_base_pair < 0){
			continue;
		}
		
		dH_d += param_H[last_base_pair][cur_base_pair];
		dS_d += param_S[last_base_pair][cur_base_pair];		
	}
		
	// If we are computing a primer3 Tm, return immediately (i.e. no self similarity
	// correction is used).
	if(parameter_set == PRIMER3){
	
		// Binding must be enthalpically driven
		if(dH_d >= 0.0f){
			return -NC_ZERO_C;
		}
	
		// This formula is based on oligotm.c
		return dH_d*1000.0f / (dS_d*1000.0f + 1.987f * log(strand_concentration/4.0f) )
			- NC_ZERO_C + 16.6f * log10(na_concentration);
	}
	
	// alpha, the self similarity correction to the
	// strand concentration term and the symmetry corrections
	// to H and D
	float alpha = 4.0f;
	float dH_s = 0.0f;
	float dS_s = 0.0f;
		
	if(is_self_complement){
		alpha = 1.0f;
		dS_s = ALLAWI_SANTALUCIA_SYMMETRY_DS;
	}
	
	// Entropic correction for salt concentration. The correction
	// is proportional to the total number of phosphates in the 
	// duplex divided by 2. [An 8bp duplex would have N = 7, not 8]
	float dS_salt = ALLAWI_SANTALUCIA_SALT * (0.5f*num_base - 1) * log(na_concentration);
	
	const float dH_total = dH_d + dH_s;
	const float dS_total = dS_d + dS_s + dS_salt;
	
	#ifdef _DEBUG
	cerr << endl;
	
	cerr << "num_base = " << num_base << endl;
	cerr << "[Na+] = " << na_concentration << endl;
	cerr << "[Ct] = " << strand_concentration << endl;
	cerr << "R*log([Ct]/alpha) = " << NC_R*log(strand_concentration/alpha) << endl;
	cerr << "alpha = " << alpha << endl;
	cerr << "DH_s = " << dH_s << endl;
	cerr << "DS_s = " << dS_s << endl;
	cerr << "dS_salt = " << dS_salt << endl;
	
	cerr << "DH = " << dH_d + dH_s << endl;
	cerr << "DS = " << dS_d + dS_s + dS_salt << endl;
	
	cerr << "dG(37 C) = " << dH_d + dH_s - (37.0f + NC_ZERO_C)*(dS_d + dS_s + dS_salt) << endl;
	#endif // _DEBUG
	
	// Binding must be enthalpically driven
	if(dH_total >= 0.0f){
		return -NC_ZERO_C;
	}
	
	// Tm = (dH_d + dH_n + dH_s)/(Rln{[C_T]/alpha} + 
	//	dS_d + dS_n + dS_s + dS_salt)
	//
	// Note that the nucleation contributions has already been included
	// in dH_d and dS_d
	const float tm = dH_total/(NC_R*log(strand_concentration/alpha) + dS_total);
	
	return tm - NC_ZERO_C;
}

// Use the internal reference/target buffer
float NucCruc::Tm_reference(const float &m_DH_bias /*= 0.0f*/, const float &m_DS_bias /*= 0.0f*/) const
{
	// Unlike the more general Tm calculation, the Tm_referenece requires that 
	// both reference and target sequences have the same length.
	if( reference_buffer.size() != target_buffer.size() ){
		throw __FILE__ ":Tm_reference: Reference/Target buffer size mismatch";
	}
	
	// Find the duplex substring that has the best free energy (most stable). This
	// is equivalent to dynamic programming on a diagonal matrix (since we do not currently
	// allow gaps, there is no point to incurring the cost of the full matrix dynamic
	// programming rigmarol).
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator ref_iter;
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator tar_iter;
		
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator left_ref_iter;
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator right_ref_iter;
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator left_tar_iter;
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator right_tar_iter;
	
	int last_base_pair = -1;
	int cur_base_pair = -1;
	
	// Compute the cumulative free energy using a reference temperature of 37 degrees.
	// This is equivalent to finding the best LOCAL alignment on the diagonal element 
	// of the dynamic programming matrix.
	float c_dG = 0.0f;
	
	// Also save the value of the min c_dG and its location
	float min_c_dG = 99999999.9f;
	
	// Compute all di-nucleotide free energy parameters. We will then extract the
	// the most stable subset for actual Tm calculation. Note that this caluclation
	// does not take into account the initiation energies due to terminal A/T or G/C
	// base pairs.
	
	for(ref_iter = reference_buffer.begin(), tar_iter = target_buffer.begin();
	    ref_iter != reference_buffer.end(); // Reference and target have the same lengths (so only test ref)
	    ref_iter++,tar_iter++){
		
		last_base_pair = cur_base_pair;
		cur_base_pair = (*ref_iter)*5 + (*tar_iter);
		
		if(last_base_pair < 0){
			continue;
		}
		
		c_dG += param_H[last_base_pair][cur_base_pair]
				- (37.0f + NC_ZERO_C)*param_S[last_base_pair][cur_base_pair];
		
		// Use a local, not global alignment
		c_dG = min(c_dG, 0.0f);
		
		if(c_dG < min_c_dG){
			min_c_dG = c_dG;
			right_ref_iter = ref_iter;
			right_tar_iter = tar_iter;
		}
	}
	
	// Once we have identified the location and value of the minimum cumulative
	// delta G, back track to find the location of the minimum starting from the right.
	
	// Reset the current base pair
	cur_base_pair = -1;
	c_dG = 0.0f;
	
	ref_iter = right_ref_iter;
	tar_iter = right_tar_iter;
	
	min_c_dG = 99999999.9f;
	
	// Count the number of duplex pairs
	unsigned int num_pair = 0;
	unsigned int count = 0;
	
	// Do we need to compute the sequence identity?
	const bool compute_identity = (identity_cutoff > 0.0f);
	unsigned int num_ident = 0;
		
	while(true){
		
		count ++;
		
		last_base_pair = cur_base_pair;
		cur_base_pair = (*ref_iter)*5 + (*tar_iter);;
		
		if(compute_identity){
			// Compute the sequence identity
			num_ident += is_exact_match(cur_base_pair);
		}
		
		if(last_base_pair < 0){
		
			// Both reference and target buffers have the same length
			// so we only need to test the target iterator
			if( ref_iter == reference_buffer.begin() ){
				// Too few bases to compute Tm
				return -NC_ZERO_C;
			}
		
			ref_iter --;
			tar_iter --;
			continue;
		}
		
		// ************************** NOTE! **************************
		// Since we are traversing the duplex in reverse, we must reverse the
		// order of the array indexing:
		// [last_base_pair][cur_base_pair] -- [cur_base_pair][last_base_pair]
		c_dG += param_H[cur_base_pair][last_base_pair]
				- (37.0f + NC_ZERO_C)*param_S[cur_base_pair][last_base_pair];
		
		if(c_dG < min_c_dG){
			min_c_dG = c_dG;
			left_ref_iter = ref_iter;
			left_tar_iter = tar_iter;
			num_pair = count;
		}
		
		if( ref_iter == reference_buffer.begin() ){
			break;
		}
		
		ref_iter --;
		tar_iter --;
	}
		
	// Are we prefiltering the Tm calculation by sequence identity?
	if( compute_identity && (num_ident < num_pair*identity_cutoff) ){
		// Insufficent sequence identity to warrent a full Tm calculation
		return -NC_ZERO_C;
	}
	
	// The parameterization in terms of di-nucleotide interactions
	// doesn't make sense for less than 3 bases since the flanking bases could
	// be mismatches -- which would cause trouble when trying to compute the 
	// nucleation energy.
	if(num_pair <= 3){
		return -NC_ZERO_C;
	}
	
	// The di-nulceotide contribution to enthalpy and entropy.
	// Incorporate user supplied biases
	float dH_d = m_DH_bias;
	float dS_d = m_DS_bias + (parameter_set == PRIMER3) ? -10.8e-3f : 0.0f;
		
	// Track the number of each kind of base
	bool is_self_complement = true;
	
	bool initiation_53 = false;
	bool initiation_35 = false;
	
	unsigned int num_base = 0;
	
	// Reset the current base pair
	cur_base_pair = -1;
	
	for(unsigned int i = 0;i < num_pair;i++, 
		left_ref_iter++, left_tar_iter++,
		right_ref_iter--, right_tar_iter--){
		
		// Is this sequence self-complementary? By default, it is until proven
		// otherwise
		if(is_self_complement){
			
			is_self_complement = (*left_ref_iter == *right_tar_iter) && 
				(*left_tar_iter == *right_ref_iter);
		}
		
		last_base_pair = cur_base_pair;
		cur_base_pair = (*left_ref_iter)*5 + *left_tar_iter;
		
		// Compute the initiation energy. Consider the first terminal base pairs
		// that make valid A/T or G/C pairs (mismatches and gaps are ignored).
		if(!initiation_53 && is_exact_match(cur_base_pair)){
			dH_d += param_nucleation_H[cur_base_pair];
			dS_d += param_nucleation_S[cur_base_pair];

			initiation_53 = true;
		}
		
		if(!initiation_35){
			int tmp_index = (*right_tar_iter)*5 + (*right_ref_iter);
			
			if(is_exact_match(tmp_index)){
				dH_d += param_nucleation_H[tmp_index];
				dS_d += param_nucleation_S[tmp_index];

				initiation_35 = true;
			}
		}
				
		// Count the number of bases
		num_base += (*left_ref_iter < GAP);
		num_base += (*left_tar_iter < GAP);
		
		if(last_base_pair < 0){
			continue;
		}
		
		dH_d += param_H[last_base_pair][cur_base_pair];
		dS_d += param_S[last_base_pair][cur_base_pair];
	}
	
	// If we are computing a primer3 Tm, return immediately (i.e. no self similarity
	// correction is used).
	if(parameter_set == PRIMER3){
		// Binding must be enthalpically driven
		if(dH_d >= 0.0f){
			return -NC_ZERO_C;
		}
		
		// This formula is based on oligotm.c
		return dH_d*1000.0f / (dS_d*1000.0f + 1.987f * log(strand_concentration/4.0f) )
			- NC_ZERO_C + 16.6f * log10(na_concentration);
	}
	
	// alpha, the self similarity correction to the
	// strand concentration term and the symmetry corrections
	// to H and D
	float alpha = 4.0f;
	float dH_s = 0.0f;
	float dS_s = 0.0f;
	
	if(is_self_complement){
		alpha = 1.0f;
		dS_s = ALLAWI_SANTALUCIA_SYMMETRY_DS;
	}
	
	// Entropic correction for salt concentration. The correction
	// is proportional to the total number of phosphates in the 
	// duplex divided by 2. [An 8bp duplex would have N = 7, not 8]
	float dS_salt = ALLAWI_SANTALUCIA_SALT * (0.5f*num_base - 1) * 
		log(na_concentration);
	
	const float dH_total = dH_d + dH_s;
	const float dS_total = dS_d + dS_s + dS_salt;
	
	#ifdef _DEBUG
	cerr << endl;
	
	cerr << "num_base = " << num_base << endl;
	cerr << "[Na+] = " << na_concentration << endl;
	cerr << "[Ct] = " << strand_concentration << endl;
	cerr << "R*log([Ct]/alpha) = " << NC_R*log(strand_concentration/alpha) << endl;
	cerr << "alpha = " << alpha << endl;
	cerr << "DH_s = " << dH_s << endl;
	cerr << "DS_s = " << dS_s << endl;
	cerr << "dS_salt = " << dS_salt << endl;
	
	cerr << "DH = " << dH_d + dH_s << endl;
	cerr << "DS = " << dS_d + dS_s + dS_salt << endl;
	
	cerr << "dG(37 C) = " << dH_d + dH_s - (37.0f + NC_ZERO_C)*(dS_d + dS_s + dS_salt) << endl;
	#endif // _DEBUG
	
	// Binding must be enthalpically driven
	if(dH_total >= 0.0f){
		return -NC_ZERO_C;
	}
	
	// Tm = (dH_d + dH_n + dH_s)/(Rln{[C_T]/alpha} + 
	//	dS_d + dS_n + dS_s + dS_salt)
	//
	// Note that the nucleation contributions has already been included
	// in dH_d and dS_d
	const float tm = dH_total/(NC_R*log(strand_concentration/alpha) + dS_total);
		
	return tm - NC_ZERO_C;
}

// How many 3' bases of m_seq_a, exactly match m_seq_b?
unsigned int NucCruc::anchor(const std::string &m_seq_a, const std::string &m_seq_b, 
		const bool &m_complement_strand_b) const
{
	unsigned int len = m_seq_a.size();
	
	if(m_seq_b.size() != len){
		throw __FILE__ ":NucCruc::Tm: Sequence lengths are not equal!";
	}
		
	unsigned int anchor = 0;
		
	for(unsigned int i = len - 1;i > 0;i--, anchor++){
		unsigned char base_a = CharToBase(m_seq_a[i]);
		unsigned char base_b;
		
		if(m_complement_strand_b){
			base_b = CharToBaseComplement(m_seq_b[i]);
		}
		else{
			base_b = CharToBase(m_seq_b[i]);
		}
				
		if(is_exact_match(base_a*5 + base_b) == false){
			return anchor;
		}
	}
	
	return anchor;
}		

// How many 5' bases of the reference, exactly match the target?
unsigned int NucCruc::anchor5_reference() const
{
	// Compute the number of 5' reference bases that exactly match
	// the target sequence  
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator 
		reference_base_iter;
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator 
		target_base_iter;
	
	unsigned int anchor = 0;
	
	reference_base_iter = reference_buffer.begin();
	target_base_iter = target_buffer.begin();
	
	while(reference_base_iter != reference_buffer.end()){
		if(is_exact_match(*reference_base_iter, *target_base_iter) ){
			anchor++;
			
			reference_base_iter++;
			target_base_iter++;
		}
		else{
			return anchor;
		}
	}
	
	return anchor;
}


// How many 3' bases of the reference, exactly match the target?
unsigned int NucCruc::anchor3_reference() const
{
	// Compute the number of 3' reference bases that exactly match
	// the target sequence  
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_reverse_iterator 
		r_reference_base_iter;
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_reverse_iterator 
		r_target_base_iter;
	
	unsigned int anchor = 0;
	
	r_reference_base_iter = reference_buffer.rbegin();
	r_target_base_iter = target_buffer.rbegin();
	
	while(r_reference_base_iter != reference_buffer.rend()){
		if(is_exact_match(*r_reference_base_iter, *r_target_base_iter) ){
			anchor++;
			
			r_reference_base_iter++;
			r_target_base_iter++;
		}
		else{
			return anchor;
		}
	}
	
	return anchor;
}

// This function provided for backwards compatibility
unsigned int NucCruc::anchor_reference() const
{
	return anchor3_reference();
}

// Do at LEAST m_anchor bases of the reference match the target
// (this should be more efficient than anchor_reference() is only a
// yes/no answer is required).
bool NucCruc::anchor_reference(const unsigned int &m_anchor) const
{
	// An anchor of zero automatically passes
	if(m_anchor == 0){
		return true;
	}

	// Compute the number of 3' reference bases that exactly match
	// the target sequence  
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_reverse_iterator 
		r_reference_base_iter( reference_buffer.rbegin() );
	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_reverse_iterator 
		r_target_base_iter( target_buffer.rbegin() );
	
	unsigned int anchor = 0;
	
	while(r_reference_base_iter != reference_buffer.rend()){
		if( is_exact_match( (*r_reference_base_iter)*5 + (*r_target_base_iter) ) ){
			anchor++;
			
			if(m_anchor <= anchor){
				return true;
			}

			r_reference_base_iter++;
			r_target_base_iter++;
		}
		else{
			// If we get here, we did not find a match
			return false;
		}
	}
	
	// If we get here, we did not find a match
	return false;
}

	
float NucCruc::MaxDimerTm(const std::string &m_seq) const
{
	// Include dangling end parameters by appending "-"
	return MaxDimerTm("-" + m_seq + "-", "-" + m_seq + "-", true);
}

float NucCruc::MaxDimerTm(const string &m_seq_a, const string &m_seq_b, 
	const bool &m_reverse_strand_b) const
{
	float max_tm = -NC_ZERO_C;
	
	const unsigned int len_a = m_seq_a.size();
	const unsigned int len_b = m_seq_b.size();
	
	if((m_seq_a.size() < min_len) || (len_b < min_len)){
		return max_tm;
	}
	
	string m_seq_reverse;
	
	const string *ptr_b;
	
	// Do we need to reverse the b oligo?
	if(m_reverse_strand_b){
		m_seq_reverse = m_seq_b;
		reverse(m_seq_reverse.begin(), m_seq_reverse.end());
		
		ptr_b = &m_seq_reverse;
	}
	else{
		ptr_b = &m_seq_b;
	}
	
	// Slide sequence a against sequence b and record the best duplex Tm.
	int start_a;
	int stop_a;
	int start_b;
	int stop_b;
	
	stop_a = len_a - 1;
	start_a = stop_a - min_len + 1;
	
	start_b = 0;
	stop_b = min_len - 1;
	
	while(true){
		const int align_len = stop_a - start_a + 1;
		
		#ifdef _DEBUG
		cerr << "align_len = " << align_len << endl;
		cerr << "len_a = " << len_a << endl;
		cerr << "start_a = " << start_a << endl;
		cerr << "stop_a = " << stop_a << endl;
		
		cerr << "len_b = " << len_b << endl;
		cerr << "start_b = " << start_b << endl;
		cerr << "stop_b = " << stop_b << endl;
		
		cerr << endl;
		#endif // _DEBUG
		
		if(align_len < (int)min_len){
			break;
		}
		
		// Compute Tm for this alignment
		//max_tm = std::max(max_tm, Tm(m_seq_a.substr(start_a, align_len), ptr_b->substr(start_b, align_len), false));
		max_tm = max(max_tm, Tm(m_seq_a.substr(start_a, align_len), ptr_b->substr(start_b, align_len), false));
		
		#ifdef _DEBUG
		cerr << "max_tm = " << max_tm << endl;
		cerr << "A = " << m_seq_a.substr(start_a, align_len) << endl;
		cerr << "B = " << ptr_b->substr(start_b, align_len) << endl;
		#endif // _DEBUG
		
		if(stop_b < (int)len_b - 1){
			// We extended b
			stop_b ++;
		}
		else{
			// We didn't extend b, so truncate a
			stop_a --;
		}
		
		if(start_a > 0){
			// We extended a
			start_a --;
		}
		else{
			// We didn't extend a, so truncate b
			start_b ++;
		}
	}
	
	return max_tm;
}

float NucCruc::MaxHairPinTm(const std::string &m_seq) const
{
	// What modifications to the duplex Delta G calculation are required to account
	// for increased Hairpin stability?
	//const float DH_hair_pin = 0.0f;
	//const float DS_hair_pin = 0.0f;
	
	// Until we determine the above modifications, just return the highest Dimer-Dimer
	// Tm. Include asymmetric dangling ends by appending "-".
	return MaxDimerTm("-" + m_seq, m_seq + "-", true);
}

// Use the first strand of the internal duplex buffer
float NucCruc::MaxHairPinTm()
{
	return MaxDimerTm_duplex(true /* compute hairpin */);
}

// Use the first strand of the internal duplex buffer
float NucCruc::MaxDimerTm()
{
	return MaxDimerTm_duplex(false /* compute dimer */);
}

// Use the internal duplex buffer for the dimer/hairpin 
// calculation. Note that this function will overwrite the 
// reference and target buffers!!!
// m_configuration controls the dangling base
// topology:
// m_configuration == true ... compute hairpin
// m_configuration == false ... compute dimer
float NucCruc::MaxDimerTm_duplex(const bool &m_configuration)
{
	float max_tm = -NC_ZERO_C;

	// Clear the reference and target buffers for writing
	reference_buffer.clear();
	target_buffer.clear();

	// Both hairpins and dimers have a terminal gap on the target strand
	target_buffer.push_back(GAP);

	list< pair<unsigned char, unsigned char> >::reverse_iterator r_duplex_iter;
	
	if(m_configuration){
		// Hairpin
		unsigned char loop_base = GAP;
		
		r_duplex_iter = duplex_buffer.rbegin();

		while( r_duplex_iter != duplex_buffer.rend() ){
			reference_buffer.push_front(r_duplex_iter->first);
			
			// We need at least 3 base pairs to compute Tm
			if(reference_buffer.size() > 3){
				max_tm = max( max_tm, Tm_reference() );				
			}
			
			// Transfer the last base of the reference buffer to the
			// end of the target buffer or the loop
			if(loop_base == GAP){
				loop_base = reference_buffer.back();
				reference_buffer.pop_back();
			}
			else{
				target_buffer.push_back(loop_base);
				loop_base = GAP;
			}
			
			r_duplex_iter ++;
		}
		
		// Add leading gap to reference strand
		reference_buffer.push_front(GAP);

		while(reference_buffer.size() > 3){
			max_tm = max( max_tm, Tm_reference() );
						
			if(loop_base == GAP){
				loop_base = reference_buffer.back();
				reference_buffer.pop_back();
			}
			else{
				target_buffer.push_back(loop_base);
				loop_base = GAP;
			}
			
			target_buffer.pop_front();
		}
	}
	else{
		// Dimer
		reference_buffer.push_front(GAP);

		// Both reference and target strands grow and shrink together
		for(r_duplex_iter = duplex_buffer.rbegin();r_duplex_iter != duplex_buffer.rend();r_duplex_iter++){
			reference_buffer.push_front(r_duplex_iter->first);
			target_buffer.push_back(r_duplex_iter->first);

			// We need at least 3 base pairs to compute Tm
			if(reference_buffer.size() > 3){
				max_tm = max( max_tm, Tm_reference() );
			}
		}

		reference_buffer.push_front(GAP);
		target_buffer.push_back(GAP);

		max_tm = max( max_tm, Tm_reference() );

		while(reference_buffer.size() > 3){
			reference_buffer.pop_back();
			target_buffer.pop_front();

			max_tm = max( max_tm, Tm_reference() );
		}
	}

	// Clear the reference and target buffers to leave a "clean slate"
	reference_buffer.clear();
	target_buffer.clear();

	return max_tm;
}

// Compute the fractional GC content of the internal duplex buffer
float NucCruc::gc_content() const
{
	if( duplex_buffer.empty() ){
		throw "Empty duplex_buffer in gc_content()";
	}

	float gc = 0.0f;

	std::list< std::pair<unsigned char, unsigned char> >::const_iterator iter;

	for(iter = duplex_buffer.begin();iter != duplex_buffer.end();iter++){
		if((iter->first == G) || (iter->first == C)){
			gc ++;
		}

		if((iter->second == G) || (iter->second == C)){
			gc ++;
		}
	}

	// Divide the fractional gc by 2 to normalize for a duplex!
	return ( gc/(2*duplex_buffer.size()) );
};

// Compute the fractional GC content of the internal duplex buffer
float NucCruc::gc_content_reference() const
{
	float gc = 0.0f;

	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator iter;

	for(iter = reference_buffer.begin();iter != reference_buffer.end();iter++){
		if((*iter == G) || (*iter == C)){
			gc ++;
		}
	}

	for(iter = target_buffer.begin();iter != target_buffer.end();iter++){
		if((*iter == G) || (*iter == C)){
			gc ++;
		}
	}

	// Divide the fractional gc by 2 to normalize for a duplex!
	return (gc/(reference_buffer.size() + target_buffer.size()));
};

// The fractional identity of the sequences stored in the reference
// buffers
float NucCruc::identity_reference() const
{
	const unsigned int ref_size = reference_buffer.size();

	// Both reference and target sequences have the same length.
	if( ref_size != target_buffer.size() ){
		throw __FILE__ ":identity_reference: Reference/Target buffer size mismatch";
	}

	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator ref_iter, tar_iter;

	float ident = 0.0f;

	for(ref_iter = reference_buffer.begin(), tar_iter = target_buffer.begin();
	    ref_iter != reference_buffer.end(); // Reference and target have the same lengths (so only test ref)
	    ref_iter++,tar_iter++){

		ident += is_exact_match(*ref_iter, *tar_iter);
	}

	return ident/ref_size;
};

unsigned int NucCruc::num_identity_reference() const
{
	const unsigned int ref_size = reference_buffer.size();

	// Both reference and target sequences have the same length.
	if( ref_size != target_buffer.size() ){
		throw __FILE__ ":identity_reference: Reference/Target buffer size mismatch";
	}

	CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator ref_iter, tar_iter;

	unsigned int num_match = 0;

	for(ref_iter = reference_buffer.begin(), tar_iter = target_buffer.begin();
	    ref_iter != reference_buffer.end(); // Reference and target have the same lengths (so only test ref)
	    ref_iter++,tar_iter++){

		num_match += is_exact_match(*ref_iter, *tar_iter);
	}

	return num_match;
};

ostream& operator << (ostream &s, const NucCruc &m_melt)
{

	// Are the reference and target buffers full?
	if(!m_melt.reference_buffer.empty() && (m_melt.reference_buffer.size() == m_melt.target_buffer.size())){
		CircleBuffer<unsigned char, CBUFFER_SIZE>::const_iterator iter;
		
		// Print the reference and target buffers
		s << "5'-";
		for(iter = m_melt.reference_buffer.begin();iter != m_melt.reference_buffer.end();iter++){
			switch(*iter){
				case NucCruc::A:
					s << 'A';
					break;
				case NucCruc::T:
					s << 'T';
					break;
				case NucCruc::G:
					s << 'G';
					break;
				case NucCruc::C:
					s << 'C';
					break;
				case NucCruc::GAP:
					s << '-';
					break;
				case NucCruc::UNKNOWN:
					s << '?';
					break;
				default:
					throw __FILE__ "Unknown base!";
			};
		}

		s << "-3'" << endl;

		s << "3'-";
		for(iter = m_melt.target_buffer.begin();iter != m_melt.target_buffer.end();iter++){
			switch(*iter){
				case NucCruc::A:
					s << 'A';
					break;
				case NucCruc::T:
					s << 'T';
					break;
				case NucCruc::G:
					s << 'G';
					break;
				case NucCruc::C:
					s << 'C';
					break;
				case NucCruc::GAP:
					s << '-';
					break;
				case NucCruc::UNKNOWN:
					s << '?';
					break;
				default:
					throw __FILE__ "Unknown base!";
			}
		}
		s << "-5'";
		return s;
	}
	
	unsigned int len = m_melt.size();
	
	// Is there any thing in the internal duplex buffer for us to print?
	if(len == 0){
		return s;
	}
	
	list< pair<unsigned char, unsigned char> >::const_iterator iter;
	
	s << "5'-";
	for(iter = m_melt.duplex_buffer.begin();iter != m_melt.duplex_buffer.end();iter++){
		switch(iter->first){
			case NucCruc::A:
				s << 'A';
				break;
			case NucCruc::T:
				s << 'T';
				break;
			case NucCruc::G:
				s << 'G';
				break;
			case NucCruc::C:
				s << 'C';
				break;
			case NucCruc::GAP:
				s << '-';
				break;
			case NucCruc::UNKNOWN:
				s << '?';
				break;
			default:
				throw __FILE__ "Unknown base!";
		};
	}
	
	s << "-3'" << endl;

	s << "3'-";
	for(iter = m_melt.duplex_buffer.begin();iter != m_melt.duplex_buffer.end();iter++){
		switch(iter->second){
			case NucCruc::A:
				s << 'A';
				break;
			case NucCruc::T:
				s << 'T';
				break;
			case NucCruc::G:
				s << 'G';
				break;
			case NucCruc::C:
				s << 'C';
				break;
			case NucCruc::GAP:
				s << '-';
				break;
			case NucCruc::UNKNOWN:
				s << '?';
				break;
			default:
				throw __FILE__ "Unknown base!";
		}
	}
	s << "-5'";
	
	return s;
}

string complement(const string &m_seq)
{
	string comp = m_seq;
	
	string::const_iterator iter;
	string::reverse_iterator comp_iter = comp.rbegin();
	
	for(iter = m_seq.begin();iter != m_seq.end();iter++, comp_iter++){
		switch(*iter){
			case 'A':
			case 'a':
				*comp_iter = 'T';
				break;
			case 'T':
			case 't':
				*comp_iter = 'A';
				break;
			case 'G':
			case 'g':
				*comp_iter = 'C';
				break;
			case 'C':
			case 'c':
				*comp_iter = 'G';
				break;
			default:
				throw ":complement: Unknown base";
		};
	}
	
	return comp;
}
