// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#ifndef __PCR_PRIMER_HEURISTIC
#define	__PCR_PRIMER_HEURISTIC


#include "circle_buffer.h"

// This class encapsulates heuristic rules for PCR primer design
// There are currently 5 heuristic criteria that are considered:
//
//	i) Avoid runs of 3 or more G's or C's at the 3' prime end.		POLY_3_GC
//	ii) The 5 bases at the 5' end should contain no more than		MULTI_5_GC
//			3 G's or C's if no two pyrimidines (T|C) are adjacent
//			2 G's or C's otherwise
//	iii) Polyprimidine (T|C) and polypurine (A|G) runs should be	NO_POLY_RUNS
//		 avoided.
//	iv) Avoid 3' terminal T's								NO_3_T
//	v)  GC_min <= %GC content <= GC_max						GC_CONTENT
// ------------------------------------------------------------------------------
// 	TaqMan (tm) Probe Design rules
//	vi)  No 5' G 											NO_5_G
//	vii) Avoid runs of identical nuleotides						NO_IDENTICAL_RUNS
//	



#define	BAD_BASE				-1
#define	PCR_VALID				0
#define	POLY_3_GC				1
#define	MULTI_5_GC			2
#define	NO_POLY_RUNS			4
#define	NO_3_T				8
#define	GC_CONTENT			16
#define	NO_5_G				32
#define	NO_IDENTICAL_RUNS		64

#define	HEURISTIC_PCR_GC_MIN	0.40f
#define	HEURISTIC_PCR_GC_MAX	0.60f
#define	HEURISTIC_PCR_RUN		4

#define	MAX_PCR_LENGTH			64

class PCRPrimer {
private:
		// What options are we testing?
		unsigned int mask;

		// Allowed ranges for GC conent
		float gc_min;
		float gc_max;

		// How many bases constitute a "run" for NO_POLY_RUNS ?
		unsigned int run_len;

		CircleBuffer<unsigned char, MAX_PCR_LENGTH> primer_buffer;

		int operator() (const CircleBuffer<unsigned char, MAX_PCR_LENGTH> &m_primer) const;
		
		bool report_verbose;
public:
	enum {A = 0, T, G, C};

	PCRPrimer(const unsigned int &m_mask, 
		const float &m_min_gc = HEURISTIC_PCR_GC_MIN, 
		const float &m_max_gc = HEURISTIC_PCR_GC_MAX) :
		mask(m_mask)
	{
		run_len = HEURISTIC_PCR_RUN;
		gc_range(m_min_gc, m_max_gc);
		report_verbose = true;
	};

	inline void verbose(const bool &m_verbose)
	{
		report_verbose = m_verbose;
	};
	
	inline unsigned int options() const
	{
		return mask;
	};

	inline void options(const unsigned int &m_mask)
	{
		mask = m_mask;
	};

	inline void run(const unsigned int &m_run_len)
	{
		run_len = m_run_len;
	};

	inline void gc_range(const float &m_min_gc, const float &m_max_gc)
	{
		if(m_min_gc > 1.0f){
			throw "GC min out of bounds (> 1.0)";
		}

		if(m_min_gc < 0.0f){
			throw "GC min out of bounds (< 0.0)";
		}
		
		if(m_max_gc > 1.0f){
			throw "GC max out of bounds (> 1.0)";
		}

		if(m_max_gc < 0.0f){
			throw "GC max out of bounds (< 0.0)";
		}
		
		if(m_max_gc < m_min_gc){
			throw "GC range out of bounds (max < min)";
		}

		gc_min = m_min_gc;
		gc_max = m_max_gc;
	};
	
	int operator() (const std::string &m_primer) const;

	// Use the internal buffer
	int operator() () const;

	inline void push_back(const char &m_base)
	{
		switch(m_base){
			case 'A':
			case 'a':
				primer_buffer.push_back(A);
				break;
			case 'T':
			case 't':
				primer_buffer.push_back(T);
				break;
			case 'G':
			case 'g':
				primer_buffer.push_back(G);
				break;
			case 'C':
			case 'c':
				primer_buffer.push_back(C);
				break;
			default:
				// This is an illegal base!
				primer_buffer.clear();
		};
	};
	
	inline void push_back_native(const unsigned char &m_base)
	{
		primer_buffer.push_back(m_base);
	};

	inline void push_front(const char &m_base)
	{
		switch(m_base){
			case 'A':
			case 'a':
				primer_buffer.push_front(A);
				break;
			case 'T':
			case 't':
				primer_buffer.push_front(T);
				break;
			case 'G':
			case 'g':
				primer_buffer.push_front(G);
				break;
			case 'C':
			case 'c':
				primer_buffer.push_front(C);
				break;
			default:
				// This is an illegal base!
				primer_buffer.clear();
		};
	};
	
	inline void push_front_native(const unsigned char &m_base)
	{
		primer_buffer.push_front(m_base);
	};

	inline void pop_front()
	{
		if(primer_buffer.empty() == false){
			primer_buffer.pop_front();
		}
	};

	inline void pop_back()
	{
		if(primer_buffer.empty() == false){
			primer_buffer.pop_back();
		}
	};

	inline void clear()
	{
		primer_buffer.clear();
	};

	inline unsigned int size() const
	{
		return primer_buffer.size();
	};

	std::string str() const;
	
	std::string error(const int &m_code) const;
};


#endif // __PCR_PRIMER_HEURISTIC
