// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

// An efficent biological sequence container
// J. D. Gans 
// Los Alamos National Laboratory, B-5
// Version 0.1
//
// Version 0.2 2/19/04
//	- Added iterators
//	- Added find functionality
//	- Added complement()
//	- Added interoperability between DNA2 and DNA3
//
// Version 0.3 4/12/04
//	- Cleaned up "const" keyword usage (mostly)
//	- Many bug fixes
//	- Added "[]" operator
//	- Added "subseq" member function (to mirror string functionality)
//
// Version 0.4 1/25/05
//	- Fixed subseq to perform bounds checking
//	- Added a default argument to the size construtor
//	- Overloaded '+' to provide string concatination
//	- Added empty()
//
// Version 0.5 1/28/06
//	- Fixed _find() bug that allowed searches on periodic genomes to 
//	  return incorrectly.
// Version 0.6 8/7/06
//	- Fixed bug in DNA3Seq::find() that missed exact matches at 3' end
//	  of target sequence.

#ifndef __BIO_SEQUENCE
#define __BIO_SEQUENCE

#include <string>
#include <iostream>
#include <vector>
#include <list>

// To turn on debugging code, uncomment the following line, or
// define SEQ_DEBUG in the makefile.
// #define	SEQ_DEBUG

// How many characters across for output with "<<"?
#define	DNA_OUTPUT_LINE_LENGTH		70u

// Should we wrap output?
extern bool wrap_seq_output;

#if defined(WIN32) && !defined(_WINDOWS)
template <class A>
A min(const A &m_a, const A &m_b);
#endif // WIN32

//////////////////////////// DNA 3 bit packing ////////////////////////////////////
#define	DNA3UnitSizeInBase	8
#define	DNA3UnitSizeInByte	3

#define	BITMASK3			7
#define	BITMASK2			3

// A forward definition of
class DNA2Seq;

// The 3 bit, binary DNA base dictionary. The first four entires
// must be consistent with the DNA2Seq structure for interoperability.
// A = 000
// C = 001
// G = 010
// T = 011
// ? = 100	(unknown base)
// - = 101	(gap)
// NONE = 110	(no base)

// Implementation notes
// 1) To save space, bases are store in groups of 8
// 2) The number of bases (num_base) is stored to make the size()
//	 function efficient (at the cost of additional memory overhead).
class DNA3Seq{
	friend class DNA2Seq;
private:
	unsigned char *seq;
	unsigned int len;		// Number of DNAUnits (i.e. NOT length of array).
	unsigned int num_base;	// Number of DNA bases

	// Provide efficient access to DNA3Seq data
	class __iterator{
		public:
			unsigned char *seq;
			unsigned int num_base;
			int base_index;
			unsigned int byte_index;
			unsigned char offset;

			__iterator()
			{
				seq = NULL;
				num_base = 0;
				base_index = -1;
				byte_index = 0;
				offset = 0;
			};
			
			virtual ~__iterator()
			{
				// Do nothing!
			};
			
			inline bool operator==(const __iterator &m_iter) const
			{
				return ((seq == m_iter.seq) && 
					   (base_index == m_iter.base_index));
			};
			
			inline bool operator!=(const __iterator &m_iter) const
			{
				return ((seq != m_iter.seq) || 
					   (base_index != m_iter.base_index));
			};
			
			inline __iterator& operator=(const __iterator &m_iter)
			{
				seq = m_iter.seq;
				num_base = m_iter.num_base;
				base_index = m_iter.base_index;
				byte_index = m_iter.byte_index;
				offset = m_iter.offset;
				
				return (*this);
			};
			
			inline int index() const
			{
				return base_index;
			};
			
			inline bool is_valid()
			{
				return ((base_index >= 0) && (base_index < (int)num_base));
			};
			
			// postfix
			//virtual __iterator& operator++(int) = 0;

			//virtual __iterator operator+(const unsigned int &m_i) = 0;
			//virtual __iterator operator-(const unsigned int &m_i) = 0;
			virtual void rewind() = 0;
			virtual unsigned char base() = 0;
	};
	
public:
	class iterator: public __iterator{
		public:
		
		// Provide a shortcut for iter = DNA3Seq::begin()
		inline void rewind()
		{
			offset = 0;
			base_index = 0;
			byte_index = 0;
		};
		
		// Return the iterator + m_i bases
		inline iterator operator+(const unsigned int &m_i)
		{
			iterator tmp_iter;

			// Did we go off the end of the sequence?
			if( (tmp_iter.base_index = base_index + m_i) > (int)num_base){
				throw __FILE__ ":iterator: Iterator overflow in (+)";
			}

			tmp_iter.seq = seq;
			tmp_iter.num_base = num_base;

			tmp_iter.offset = tmp_iter.base_index % DNA3UnitSizeInBase;
                
			tmp_iter.byte_index = (tmp_iter.base_index/DNA3UnitSizeInBase)*DNA3UnitSizeInByte;
					    			
			return tmp_iter;
		};
		
		inline iterator operator-(const unsigned int &m_i)
		{
			iterator tmp_iter;
			
			// Did we go off the end of the sequence?
			if( (tmp_iter.base_index = base_index - m_i) < 0){
				throw __FILE__ ":iterator: Iterator under flow in (-)";

			}
			
			tmp_iter.seq = seq;
			tmp_iter.num_base = num_base;
			
			tmp_iter.offset = tmp_iter.base_index % DNA3UnitSizeInBase;
			
			tmp_iter.byte_index = (tmp_iter.base_index/DNA3UnitSizeInBase)*DNA3UnitSizeInByte;
							  
			return tmp_iter;
		};
		
		inline int operator-(const iterator &m_iter) const
		{
			return (base_index - m_iter.base_index);
		};
			
		inline iterator& operator++(int) // postfix
		{
			base_index++;
			
			#ifdef SEQ_DEBUG
			if(base_index > (int)num_base){
				throw __FILE__ ":iterator ++: iterator overflow";
			}
			#endif // SEQ_DEBUG

			if( (++offset) % DNA3UnitSizeInBase == 0){
				byte_index += DNA3UnitSizeInByte;
				offset = 0;
			}
			
			return (*this);
		};
		
		inline unsigned char base()
		{
			#ifdef SEQ_DEBUG
			if(seq == NULL){
				throw __FILE__ ":DNA3Seq::iterator:base: NULL seq ptr";
			}

			if((base_index < 0) || (base_index >= (int)num_base)){
				throw __FILE__ ":DNA3Seq::iterator:base: Invalid iterator";
			}
			#endif // SEQ_DEBUG

			// There does not seem to be any time savings by replacing
			// the switch statement with a simgle computed return
			// [i.e. the bits to shift right is (21 - 3*offset)%8 for
			// offset = {0,1,3,4,6,7}]
			
			switch(offset){
				case 0:
					return seq[byte_index] >> 5;
				case 1:
					return (seq[byte_index] >> 2) & BITMASK3;
				case 2:
					return ((seq[byte_index] << 1) + 
						(seq[byte_index + 1] >> 7)) & BITMASK3;
				case 3:
					return (seq[byte_index + 1] >> 4) & BITMASK3;
				case 4:
					return (seq[byte_index + 1] >> 1) & BITMASK3;
				case 5:
					return ((seq[byte_index + 1] << 2) + 
						(seq[byte_index + 2] >> 6)) & BITMASK3;
				case 6:
					return (seq[byte_index + 2] >> 3) & BITMASK3;
				//case 7:
				//	break;
			};

			return seq[byte_index + 2] & BITMASK3;
		};
		
		// Here is a place where the iterator differs from the 
		// complement_iterator (i.e. we can set bases as well as
		// read them).
		inline void base(const unsigned char &c)
		{
			#ifdef SEQ_DEBUG
			if(seq == NULL){
				throw __FILE__ ":DNA3Seq::iterator:base: NULL seq ptr";
			}

			if((base_index < 0) || (base_index >= (int)num_base)){
				throw __FILE__ ":DNA3Seq::iterator:base: Invalid iterator";
			}

			#endif // SEQ_DEBUG
			
			unsigned char *ptr = seq + byte_index;

			switch(offset){
				case 0:
					ptr[0] = (ptr[0] & ~(BITMASK3 << 5)) + (c << 5);
					break;
				case 1:
					ptr[0] = (ptr[0] & ~(BITMASK3 << 2)) + (c << 2);
					break;
				case 2:
					ptr[0] = (ptr[0] & ~(BITMASK3 >> 1)) + (c >> 1);
					ptr[1] = (ptr[1] & ~(BITMASK3 << 7)) + (c << 7);
					break;
				case 3:
					ptr[1] = (ptr[1] & ~(BITMASK3 << 4)) + (c << 4);
					break;
				case 4:
					ptr[1] = (ptr[1] & ~(BITMASK3 << 1)) + (c << 1);
					break;
				case 5:
					ptr[1] = (ptr[1] & ~(BITMASK3 >> 2)) + (c >> 2);
					ptr[2] = (ptr[2] & ~(BITMASK3 << 6)) + (c << 6);
					break;
				case 6:
					ptr[2] = (ptr[2] & ~(BITMASK3 << 3)) + (c << 3);
					break;
				case 7:
					ptr[2] = (ptr[2] & ~(BITMASK3)) + c;
					break;
			};
		};
		
		inline char c_base()
		{
			switch(base()){
				case DNA3Seq::A:
					return 'A';
				case DNA3Seq::T:
					return 'T';
				case  DNA3Seq::G:
					return 'G';
				case DNA3Seq::C:
					return 'C';
				case DNA3Seq::GAP:
					return '-';
				case DNA3Seq::UNKNOWN:
					return 'N';
			};

			return DNA3Seq::NONE;
		};

		inline void c_base(const char &c)
		{
			switch(c){
				case 'A':
				case 'a':
					base(DNA3Seq::A);
					break;
				case 'T':
				case 't':
					base(DNA3Seq::T);
					break;
				case 'G':
				case 'g':
					base(DNA3Seq::G);
					break;
				case 'C':
				case 'c':
					base(DNA3Seq::C);
					break;
				case '-':
					base(DNA3Seq::GAP);
					break;
				default:
					base(DNA3Seq::UNKNOWN);
					break;
			};
		};
	};
	
	class complement_iterator: public __iterator{
		public:
		
		// Note that rewinding a complement_iterator is not
		// going to be efficient thanks to the "len" calculation.
		inline void rewind()
		{
			offset = num_base % DNA3UnitSizeInBase;
			base_index = num_base - 1;
			
			byte_index = (base_index/DNA3UnitSizeInBase)*DNA3UnitSizeInByte;
		};
		
		inline complement_iterator operator+(const unsigned int &m_i)
		{
			complement_iterator tmp_iter;
			
			///////////////////////////////////////////////////////
			// Since this is the complement strand, adding m_i
			// is really subtracting m_i from the current position.
			///////////////////////////////////////////////////////
			// Did we go off the end of the sequence?
			if( (tmp_iter.base_index = base_index - m_i) < -1){
				throw __FILE__ ":complement_iterator: Iterator overflow in (+)";
			}
			
			tmp_iter.seq = seq;
			tmp_iter.num_base = num_base;
			tmp_iter.offset = tmp_iter.base_index % DNA3UnitSizeInBase;
			
			tmp_iter.byte_index = (tmp_iter.base_index/DNA3UnitSizeInBase)*DNA3UnitSizeInByte;
					    			
			return tmp_iter;
		};
		
		inline complement_iterator operator-(const int &m_i)
		{
			complement_iterator tmp_iter;
			
			///////////////////////////////////////////////////////
			// Since this is the complement strand, subtracting m_i
			// is really adding m_i to the current position.
			///////////////////////////////////////////////////////
			// Did we go off the end of the sequence?
			if( (tmp_iter.base_index = base_index + m_i) >= (int)num_base){
				throw __FILE__ ":complement_iterator: Iterator underflow in (-)";
			}
			
			tmp_iter.seq = seq;
			tmp_iter.num_base = num_base;
			
			tmp_iter.offset = tmp_iter.base_index % DNA3UnitSizeInBase;
			
			tmp_iter.byte_index = (tmp_iter.base_index/DNA3UnitSizeInBase)*DNA3UnitSizeInByte;
							  
			return tmp_iter;
		};
		
		inline complement_iterator& operator++(int) // postfix
		{	
			base_index--;
			
			#ifdef SEQ_DEBUG
			if(base_index < -1){
				throw __FILE__ ":complement_iterator ++: iterator overflow";
			}
			#endif // SEQ_DEBUG

			if(offset == 0){
				byte_index -= DNA3UnitSizeInByte;
				offset = DNA3UnitSizeInBase - 1;
			}
			else{
				offset--;
			}
			
			return (*this);
		};
		
		inline unsigned char base()
		{
			#ifdef SEQ_DEBUG
			if(seq == NULL){
				throw __FILE__ ":DNA3Seq::complement_iterator:base: NULL seq ptr";
			}

			if((base_index < 0) || (base_index >= (int)num_base)){
				throw __FILE__ ":DNA3Seq::complement_iterator:base: Invalid iterator";
			}

			#endif // SEQ_DEBUG
			
			unsigned char c = DNA3Seq::UNKNOWN;
			
			switch(offset){
				case 0:
					c = seq[byte_index] >> 5;
					break;
				case 1:
					c = (seq[byte_index] >> 2) & BITMASK3;
					break;
				case 2:
					c = ((seq[byte_index] << 1) + 
						(seq[byte_index + 1] >> 7)) & BITMASK3;
					break;
				case 3:
					c = (seq[byte_index + 1] >> 4) & BITMASK3;
					break;
				case 4:
					c = (seq[byte_index + 1] >> 1) & BITMASK3;
					break;
				case 5:
					c = ((seq[byte_index + 1] << 2) + 
						(seq[byte_index + 2] >> 6)) & BITMASK3;
					break;
				case 6:
					c = (seq[byte_index + 2] >> 3) & BITMASK3;
					break;
				case 7:
					c = seq[byte_index + 2] & BITMASK3;
					break;
			};

			switch(c){
				case DNA3Seq::A:
					return DNA3Seq::T;
				case DNA3Seq::T:
					return DNA3Seq::A;
				case DNA3Seq::G:
					return DNA3Seq::C;
				case DNA3Seq::C:
					return DNA3Seq::G;
			};
			
			return DNA3Seq::UNKNOWN;
		};
		
		inline char c_base()
		{
			switch(base()){
				case DNA3Seq::A:
					return 'A';
				case DNA3Seq::T:
					return 'T';
				case  DNA3Seq::G:
					return 'G';
				case DNA3Seq::C:
					return 'C';
				case DNA3Seq::GAP:
					return '-';
				case DNA3Seq::UNKNOWN:
					return 'N';
			};

			return DNA3Seq::NONE;
		};
	};
	
	inline iterator begin() const
	{
		iterator dummy;
		
		dummy.seq = seq;
		dummy.num_base = num_base;
		dummy.base_index = 0;
		
		return dummy;
	};
	
	inline iterator end() const
	{
		iterator dummy;
		
		dummy.seq = seq;
		dummy.num_base = num_base;

		// This is the end()
		dummy.base_index = num_base;

		return dummy;
	};
	
	inline complement_iterator complement_begin() const
	{
		complement_iterator dummy;
		
		dummy.seq = seq;
		dummy.num_base = num_base;
		dummy.base_index = num_base - 1; // [0 ... num_base - 1]
		dummy.byte_index = (len - 1)*DNA3UnitSizeInByte; // [0 ... (len-1)*DNA3UnitSizeInByte]
		dummy.offset = (dummy.base_index) % DNA3UnitSizeInBase; // [0 ... DNA3UnitSizeInBase - 1]
		
		return dummy;
	};
	
	inline complement_iterator complement_end() const
	{
		complement_iterator dummy;
		
		dummy.seq = seq;
		dummy.num_base = num_base;

		// This is the end
		dummy.base_index = -1;

		return dummy;
	};
	
	// Note that A, C, G and T must have the same values for both
	// DNA2Seq and DNA3Seq.
	enum {A = 0, C = 1, G = 2, T = 3, UNKNOWN = 4, GAP = 5, NONE = 6};
	
	friend std::ostream& operator << (std::ostream &s, const DNA3Seq &m_seq);
	
	DNA3Seq();
	DNA3Seq(const std::string &m_seq);
	DNA3Seq(const std::string &m_seq, const int &m_from, const int &m_len,
		const char &m_strand);
	DNA3Seq(const DNA3Seq &m_seq, const int &m_from, const int &m_len,
		const char &m_strand);
	DNA3Seq(const char *m_seq);
	DNA3Seq(const DNA3Seq &m_seq);
	DNA3Seq(const DNA2Seq &m_seq);
	DNA3Seq(const int &m_size, const unsigned char &m_default_base = NONE);
	~DNA3Seq();
	
	// Make a copy of the given sequence
	DNA3Seq& operator=(const std::string &m_seq);
	DNA3Seq& operator=(const char *m_seq);
	DNA3Seq& operator=(const DNA3Seq &m_seq);
	DNA3Seq& operator=(const DNA2Seq &m_seq);
	
	// Concatinate sequences
	DNA3Seq operator+(const DNA3Seq &m_seq);
	
	// Read only access (for efficient sequential access, please use the iterator)
	inline const unsigned char operator[](const unsigned int &m_i) const
	{
		iterator iter = begin() + m_i;
		
		// Return base, not c_base
		return iter.base();
	};
	
	DNA3Seq subseq(const unsigned int &m_start, const unsigned int &m_len, 
		const bool &m_circular = false) const;
	
	// Set the current sequence equal the complement of the given sequence
	void complement(const char *m_seq);
	void complement(const std::string &m_seq);
	void complement(const DNA3Seq &m_seq);
	
	// Return the complement of the given sequence
	// Not efficient! Please don't use if speed is critical !!!
	DNA3Seq complement() const;

	// Still need a complement function that takes a DNA3Seq
	// argument! Use the complement_iterator
	
	// Find the first occurance of the given sequence as a 
	// substring in the current sequence.
#ifndef SEQ_DEBUG	
private:
#endif // SEQ_DEBUG
	// Private functions using brute force base comparison
	iterator __find(const DNA3Seq &m_seq,
		const bool &m_periodic = false) const;
	iterator __find(const DNA3Seq &m_seq,
		const iterator &m_begin,
		const bool &m_periodic = false) const;
	iterator __find(const DNA3Seq &m_seq, 
		const iterator &m_begin, 
		const iterator &m_end,
		const bool &m_periodic = false) const;
	iterator __find(const DNA2Seq &m_seq) const;
	complement_iterator __find_complement(const DNA3Seq &m_seq) const;
	complement_iterator __find_complement(const DNA2Seq &m_seq) const;
public:
	// Public functions using an implementation of the Rabin-Karp string matching
	// algorithm.
	iterator find(const DNA3Seq &m_seq, 
		const bool &m_periodic = false) const;
	iterator find(const DNA3Seq &m_seq,
		const iterator &m_begin, 
		const bool &m_periodic = false) const;
	iterator find(const DNA3Seq &m_seq,
		const iterator &m_begin, 
		const iterator &m_end, 
		const bool &m_periodic = false) const;
	
	iterator find(const DNA2Seq &m_seq) const;
	
	complement_iterator find_complement(const DNA3Seq &m_seq) const;
	complement_iterator find_complement(const DNA2Seq &m_seq) const;
	
	// Find the first occurance of either m_seq_a or m_seq_b (i.e. 
	// search for a sequence and its complement). Note that the
	// list-based form of this function is slightly less efficient than
	// the vector-based form.
	iterator find_any(const DNA3Seq &m_seq_a, const DNA3Seq &m_seq_b);
	iterator find_any(const DNA2Seq &m_seq_a, const DNA2Seq &m_seq_b);
	unsigned int find_any(const std::vector<DNA3Seq> &m_seq,
		std::list<unsigned int> &m_match);
	//iterator find_any(const std::vector<DNA2Seq> &m_seq);
	unsigned int find_any(const std::list<DNA3Seq> &m_seq,
		std::list<const DNA3Seq*> &m_match);
	//iterator find_any(const std::list<DNA2Seq> &m_seq);
	
	// It is crucial that this function be inlined!
	bool operator==(const DNA3Seq &m_seq) const
	{
			
		// "==" is currently defined as strict equality
		if(len == 0){
			return (m_seq.len == 0);
		}
		else{	
			return ((len == m_seq.len) ? (memcmp(seq, m_seq.seq, len*DNA3UnitSizeInByte) == 0): false);
		}
		
		// Same as above, but in a "compact" (i.e. hard to read) form
		//return (len == 0) ? (m_seq.len == 0) : 
		//	((len == m_seq.len) ? (memcmp(seq, m_seq.seq, len*DNA3UnitSizeInByte) == 0): false);

	};
	
	bool operator!=(const DNA3Seq &m_seq) const
	{
		return !(operator==(m_seq));
	};
	
	// To allow use of the DNA3Seq in STL containers, we must define the 
	// comparison operator "<".
	bool operator < (const DNA3Seq &m_seq) const
	{
		if(len == 0){
			return !(m_seq.len == 0); // If both are length 0, "<" is false
								 // because the seq are equal.
		}
		else{
			if(len < m_seq.len){
				return (memcmp(seq, m_seq.seq, len*DNA3UnitSizeInByte) <= 0);
			}
			
			return (memcmp(seq, m_seq.seq, m_seq.len*DNA3UnitSizeInByte) < 0);
		}
	};
	
	bool operator > (const DNA3Seq &m_seq) const
	{
		if(len == 0){
			return false; // A zero length sequence can not be less than another
					    // sequence of any length!
		}
		else{
			if(len < m_seq.len){
				return (memcmp(seq, m_seq.seq, len*DNA3UnitSizeInByte) > 0);
			}
			else{
				if(len > m_seq.len){
					return (memcmp(seq, m_seq.seq, m_seq.len*DNA3UnitSizeInByte) >= 0);
				}
				
				// len ==  m_seq.len
				return (memcmp(seq, m_seq.seq, len*DNA3UnitSizeInByte) > 0);
			}
		}
	};
	
	// Convert the sequence to human readable form
	std::string str() const;
	
	// Return the length of the sequence (in bases)
	// inlined for efficiency
	inline unsigned int size() const
	{
		return num_base;
	};
	
	inline bool empty() const
	{
		return (num_base == 0);
	};
	
	// Are all bases in this sequence either A, T, G, or C?
	bool is_ATGC() const;
	
	// Return the fractional gc content
	float gc_content() const;

	#ifdef USE_MPI
	// Compute the size of this structure in bytes
	unsigned int mpi_size() const
	{
		// How much information do we need to transmite this
		// structure?
		return (sizeof(num_base) + sizeof(len) + len*DNA3UnitSizeInByte);
	};
	
	// Copy this structure into a buffer for MPI transport
	// Note that buffer MUST contain at least mpi_size bytes!
	// Return the current buffer position
	unsigned char* mpi_pack(unsigned char* m_buffer) const
	{
		memcpy(m_buffer, &num_base, sizeof(num_base));
		m_buffer += sizeof(num_base);
		
		memcpy(m_buffer, &len, sizeof(len));
		m_buffer += sizeof(len);
		
		memcpy(m_buffer, seq, len*DNA3UnitSizeInByte);
		m_buffer += len*DNA3UnitSizeInByte;
		
		return m_buffer;
	};
	
	// Unpack the structure contained in the buffer
	unsigned char* mpi_unpack(unsigned char* m_buffer)
	{
		// Free previous memory (if any)
		if(seq){
			delete [] seq;
		}
		
		memcpy(&num_base, m_buffer, sizeof(num_base));
		m_buffer += sizeof(num_base);
		
		memcpy(&len, m_buffer, sizeof(len));
		m_buffer += sizeof(len);
		
		seq = new unsigned char[len*DNA3UnitSizeInByte];
				
		memcpy(seq, m_buffer, len*DNA3UnitSizeInByte);
		m_buffer += len*DNA3UnitSizeInByte;
		
		return m_buffer;
	};
	
	#endif // USE_MPI
};

// Allow direct output sequence  
std::ostream& operator << (std::ostream &s, const DNA3Seq &m_seq);

//////////////////////////// DNA 2 bit packing ////////////////////////////////////
#define	DNA2UnitSizeInBase	4
#define	DNA2UnitSizeInByte	1

// The 2 bit, binary DNA base dictionary
// A = 00
// C = 01
// G = 10
// T = 11

// Implementation notes
// 1) To save space, bases are store in groups of 4
// 2) The number of bases (num_base) is stored to make the size()
//	 function efficient (at the cost of additional memory overhead).
// Note that in the 2 bit packing scheme, there is no allowance for
// gaps or unknown characters. Attempting to use any base other than
// A, T, G or C will result in a thrown error.
//
// Also note that there are differences in the layout of DNA2Seq vs
// DNA3Seq due to the simplification DNA2UnitSizeInByte == 1.

class DNA2Seq{
	friend class DNA3Seq;
private:
	unsigned char *seq;
	unsigned int len;		// Number of DNAUnits (length of array).
	unsigned int num_base;	// Number of DNA bases
	unsigned int seq_remainder; // The number of bases in the last unit
	
	// Provide efficient access to DNA2Seq data
	class __iterator{
		public:
			unsigned char *seq;
			unsigned int num_base;
			int base_index;
			unsigned int byte_index;
			unsigned char offset;

			__iterator()
			{
				seq = NULL;
				num_base = 0;
				base_index = -1;
				byte_index = 0;
				offset = 0;
			};
			
			virtual ~__iterator()
			{
				// Do nothing!
			};
			
			inline bool operator==(const __iterator &m_iter) const
			{
				return ((seq == m_iter.seq) && 
					   (base_index == m_iter.base_index));
			};
			
			inline bool operator!=(const __iterator &m_iter) const
			{
				return ((seq != m_iter.seq) || 
					   (base_index != m_iter.base_index));
			};
			
			inline __iterator& operator=(const __iterator &m_iter)
			{
				seq = m_iter.seq;
				num_base = m_iter.num_base;
				base_index = m_iter.base_index;
				byte_index = m_iter.byte_index;
				offset = m_iter.offset;
				
				return (*this);
			};
			
			inline int index() const
			{
				return base_index;
			};
			
			inline bool is_valid()
			{
				return ((base_index >= 0) && (base_index < (int)num_base));
			};
			
			// postfix
			//virtual __iterator& operator++(int) = 0;

			//virtual __iterator operator+(const unsigned int &m_i) = 0;
			//virtual __iterator operator-(const unsigned int &m_i) = 0;
			virtual void rewind() = 0;
			virtual unsigned char base() = 0;
	};
	
public:
	class iterator: public __iterator{
		public:
		
		// Provide a shortcut for iter = DNA2Seq::begin()
		inline void rewind()
		{
			offset = 0;
			base_index = 0;
			byte_index = 0;
		};
		
		// Return the iterator + m_i bases
		inline iterator operator+(const unsigned int &m_i)
		{
			iterator tmp_iter;

			// Did we go off the end of the sequence?
			if( (tmp_iter.base_index = base_index + m_i) > (int)num_base){
				throw __FILE__ ":iterator: Iterator overflow in (+)";
			}

			tmp_iter.seq = seq;
			tmp_iter.num_base = num_base;

			tmp_iter.offset = tmp_iter.base_index % DNA2UnitSizeInBase;
                
            tmp_iter.byte_index = tmp_iter.base_index/DNA2UnitSizeInBase;
					    			
			return tmp_iter;
		};
		
		inline iterator operator-(const unsigned int &m_i)
		{
			iterator tmp_iter;
			
			// Did we go off the end of the sequence?
			if( (tmp_iter.base_index = base_index - m_i) < 0){
				throw __FILE__ ":iterator: Iterator under flow in (-)";

			}
			
			tmp_iter.seq = seq;
			tmp_iter.num_base = num_base;
			
			tmp_iter.offset = tmp_iter.base_index % DNA2UnitSizeInBase;
			
			tmp_iter.byte_index = tmp_iter.base_index/DNA2UnitSizeInBase;
							  
			return tmp_iter;
		};
		
		inline int operator-(const iterator &m_iter) const
		{
			return (base_index - m_iter.base_index);
		};
		
		inline iterator& operator++(int) // postfix
		{
			base_index++;
			
			#ifdef SEQ_DEBUG
			if(base_index > (int)num_base){
				throw __FILE__ ":iterator ++: iterator overflow";
			}
			#endif // SEQ_DEBUG

			if( (++offset) % DNA2UnitSizeInBase == 0){
				byte_index += DNA2UnitSizeInByte;
				offset = 0;
			}
			
			return (*this);
		};
		
		inline unsigned char base()
		{
			#ifdef SEQ_DEBUG
			if(seq == NULL){
				throw __FILE__ ":DNA2Seq::iterator:base: NULL seq ptr";
			}

			if((base_index < 0) || (base_index >= (int)num_base)){
				std::cerr << "base_index = " << base_index << std::endl;
				std::cerr << "num_base = " << num_base << std::endl;
				
				throw __FILE__ ":DNA2Seq::iterator:base: Invalid iterator";
			}
			#endif // SEQ_DEBUG

			// There does not seem to be any time savings by replacing
			// the switch statement with a simgle computed return
			// [i.e. the bits to shift right is (21 - 3*offset)%8 for
			// offset = {0,1,3,4,6,7}]
			
			return (seq[byte_index] >> (6 - 2*offset)) & BITMASK2;
			
			//switch(offset){
			//	case 0:
			//		return seq[byte_index] >> 6;
			//	case 1:
			//		return (seq[byte_index] >> 4) & BITMASK2;
			//	case 2:
			//		return (seq[byte_index] >> 2) & BITMASK2;
				//case 3:
				//	return (seq[byte_index]) & BITMASK2;
			//};

			//return (seq[byte_index]) & BITMASK2;
		};
		
		// Here is a place where the iterator differs from the 
		// complement_iterator (i.e. we can set bases as well as
		// read them).
		inline void base(const unsigned char &c)
		{
			#ifdef SEQ_DEBUG
			if(seq == NULL){
				throw __FILE__ ":DNA2Seq::iterator:base: NULL seq ptr";
			}

			if((base_index < 0) || (base_index >= (int)num_base)){
				throw __FILE__ ":DNA2Seq::iterator:base: Invalid iterator";
			}

			#endif // SEQ_DEBUG
			
			unsigned int shift = 6 - 2*offset;
			
			seq[byte_index] = (seq[byte_index] & ~(BITMASK2 << shift)) + (c << shift);
			
			//unsigned char *ptr = seq + byte_index;

			//switch(offset){
			//	case 0:
			//		*ptr = (*ptr & ~(BITMASK2 << 6)) + (c << 6);
			//		break;
			//	case 1:
			//		*ptr = (*ptr & ~(BITMASK2 << 4)) + (c << 4);
			//		break;
			//	case 2:
			//		*ptr = (*ptr & ~(BITMASK2 << 2)) + (c << 2);
			//		break;
			//	case 3:
			//		*ptr = (*ptr & ~BITMASK2) + c;
			//		break;
			//};
		};
		
		inline char c_base()
		{
			switch(base()){
				case DNA2Seq::A:
					return 'A';
				case DNA2Seq::T:
					return 'T';
				case  DNA2Seq::G:
					return 'G';
				case DNA2Seq::C:
					return 'C';
			};

			// We should never get here!
			throw __FILE__ ":DNA2Seq::c_base: Illegal base";
			
			return 0;
		};

		inline void c_base(const char &c)
		{
			switch(c){
				case 'A':
				case 'a':
					base(DNA2Seq::A);
					break;
				case 'T':
				case 't':
					base(DNA2Seq::T);
					break;
				case 'G':
				case 'g':
					base(DNA2Seq::G);
					break;
				case 'C':
				case 'c':
					base(DNA2Seq::C);
					break;
				default:
				throw __FILE__ ":DNA2Seq::c_base: Illegal base";				
			};
		};
	};
	
	class complement_iterator: public __iterator{
		public:
		
		// Note that rewinding a complement_iterator is not
		// going to be efficient thanks to the "len" calculation.
		inline void rewind()
		{
			offset = num_base % DNA2UnitSizeInBase;
			base_index = num_base - 1;
			
			byte_index = base_index/DNA2UnitSizeInBase;
		};
		
		inline complement_iterator operator+(const unsigned int &m_i)
		{
			complement_iterator tmp_iter;
			
			///////////////////////////////////////////////////////
			// Since this is the complement strand, adding m_i
			// is really subtracting m_i from the current position.
			///////////////////////////////////////////////////////
			// Did we go off the end of the sequence?
			if( (tmp_iter.base_index = base_index - m_i) < -1){
				throw __FILE__ ":complement_iterator: Iterator overflow in (+)";
			}
			
			tmp_iter.seq = seq;
			tmp_iter.num_base = num_base;
			tmp_iter.offset = tmp_iter.base_index % DNA2UnitSizeInBase;
			
			tmp_iter.byte_index = tmp_iter.base_index/DNA2UnitSizeInBase;
					    			
			return tmp_iter;
		};
		
		inline complement_iterator operator-(const int &m_i)
		{
			complement_iterator tmp_iter;
			
			///////////////////////////////////////////////////////
			// Since this is the complement strand, subtracting m_i
			// is really adding m_i to the current position.
			///////////////////////////////////////////////////////
			// Did we go off the end of the sequence?
			if( (tmp_iter.base_index = base_index + m_i) >= (int)num_base){
				throw __FILE__ ":complement_iterator: Iterator underflow in (-)";
			}
			
			tmp_iter.seq = seq;
			tmp_iter.num_base = num_base;
			
			tmp_iter.offset = tmp_iter.base_index % DNA2UnitSizeInBase;
			
			tmp_iter.byte_index = tmp_iter.base_index/DNA2UnitSizeInBase;
							  
			return tmp_iter;
		};
		
		inline complement_iterator& operator++(int) // postfix
		{	
			base_index--;
			
			#ifdef SEQ_DEBUG
			if(base_index < -1){
				throw __FILE__ ":complement_iterator ++: iterator overflow";
			}
			#endif // SEQ_DEBUG

			if(offset == 0){
				byte_index -= DNA2UnitSizeInByte;
				offset = DNA2UnitSizeInBase - 1;
			}
			else{
				offset--;
			}
			
			return (*this);
		};
		
		inline unsigned char base()
		{
			#ifdef SEQ_DEBUG
			if(seq == NULL){
				throw __FILE__ ":DNA2Seq::complement_iterator:base: NULL seq ptr";
			}

			if((base_index < 0) || (base_index >= (int)num_base)){
				throw __FILE__ ":DNA2Seq::complement_iterator:base: Invalid iterator";
			}

			#endif // SEQ_DEBUG
			
			switch((seq[byte_index] >> (6 - 2*offset)) & BITMASK2){
				case DNA2Seq::A:
					return DNA2Seq::T;
				case DNA2Seq::T:
					return DNA2Seq::A;
				case DNA2Seq::G:
					return DNA2Seq::C;
				case DNA2Seq::C:
					return DNA2Seq::G;
			};
			
			throw __FILE__ ":DNA2Seq::complement_iterator:base: Unknown base!";
		};
		
		inline char c_base()
		{
			switch(base()){
				case DNA2Seq::A:
					return 'A';
				case DNA2Seq::T:
					return 'T';
				case  DNA2Seq::G:
					return 'G';
				case DNA2Seq::C:
					return 'C';
			};
			
			throw __FILE__ ":DNA2Seq::complement_iterator:c_base: Unknown base!";
			return 0;
		};
	};
	
	inline iterator begin() const
	{
		iterator dummy;
		
		dummy.seq = seq;
		dummy.num_base = num_base;
		dummy.base_index = 0;
		
		return dummy;
	};
	
	inline iterator end() const
	{
		iterator dummy;
		
		dummy.seq = seq;
		dummy.num_base = num_base;

		// This is the end()
		dummy.base_index = num_base;

		return dummy;
	};
	
	inline complement_iterator complement_begin() const
	{
		complement_iterator dummy;
		
		dummy.seq = seq;
		dummy.num_base = num_base;
		dummy.base_index = num_base - 1; // [0 ... num_base - 1]
		dummy.byte_index = (len - 1); // [0 ... (len-1)*DNA2UnitSizeInByte]
		dummy.offset = (dummy.base_index) % DNA2UnitSizeInBase; // [0 ... DNA2UnitSizeInBase - 1]
		
		return dummy;
	};
	
	inline complement_iterator complement_end() const
	{
		complement_iterator dummy;
		
		dummy.seq = seq;
		dummy.num_base = num_base;

		// This is the end
		dummy.base_index = -1;

		return dummy;
	};
	
	// Note that A, C, G and T must have the same values for both
	// DNA2Seq and DNA3Seq.
	enum {A = 0, C = 1, G = 2, T = 3};
	
	// Since there is no NONE symbol, initialize any remainder bases
	// with A. DEFAULT_BYTE is = AAAA;
	enum {DEFAULT_BYTE = 0};
	
	friend std::ostream& operator << (std::ostream &s, const DNA2Seq &m_seq);
	
	DNA2Seq();
	DNA2Seq(const std::string &m_seq);
	DNA2Seq(const std::string &m_seq, const int &m_from, const int &m_len,
		const char &m_strand);
	DNA2Seq(const DNA2Seq &m_seq, const int &m_from, const int &m_len,
		const char &m_strand);
	DNA2Seq(const char *m_seq);
	DNA2Seq(const DNA2Seq &m_seq);
	DNA2Seq(const DNA3Seq &m_seq);
	DNA2Seq(const int &m_size);
	~DNA2Seq();
	
	// Make a copy of the given sequence
	DNA2Seq& operator=(const std::string &m_seq);
	DNA2Seq& operator=(const char *m_seq);
	DNA2Seq& operator=(const DNA2Seq &m_seq);
	DNA2Seq& operator=(const DNA3Seq &m_seq);
	
	// Concatinate sequences
	DNA2Seq operator+(const DNA2Seq &m_seq);
	
	// Read only access (for efficient sequential access, please use the iterator)
	inline const unsigned char operator[](const unsigned int &m_i) const
	{
		iterator iter = begin() + m_i;
		
		// Return base, not c_base
		return iter.base();
	};
	
	DNA2Seq subseq(const int &m_start, const unsigned int &m_len, 
		const bool &m_circular = false) const;
	
	// Set the current sequence equal the complement of the given sequence
	void complement(const char *m_seq);
	void complement(const std::string &m_seq);
	void complement(const DNA2Seq &m_seq);
	
	// Return the complement of the given sequence
	// Not efficient! Please don't use if speed is critical !!!
	DNA2Seq complement() const;

	// Find the first occurance of the given sequence as a 
	// substring in the current sequence.
#ifndef SEQ_DEBUG	
private:
#endif // SEQ_DEBUG
	// Private functions using brute force base comparison
	iterator __find(const DNA2Seq &m_seq) const;
	iterator __find(const DNA2Seq &m_seq, 
		const iterator &m_begin) const;
	iterator __find(const DNA2Seq &m_seq,
		const iterator &m_begin,
		const iterator &m_end) const;
	
	iterator __find(const DNA3Seq &m_seq) const;
	complement_iterator __find_complement(const DNA2Seq &m_seq) const;
	complement_iterator __find_complement(const DNA3Seq &m_seq) const;
public:
	// Public functions using an implementation of the Rabin-Karp string matching
	// algorithm.
	iterator find(const DNA2Seq &m_seq) const;
	iterator find(const DNA2Seq &m_seq,
		const iterator &m_begin) const;
	iterator find(const DNA2Seq &m_seq,
		const iterator &m_begin,
		const iterator &m_end) const;
	
	iterator find(const DNA3Seq &m_seq) const;
	complement_iterator find_complement(const DNA2Seq &m_seq) const;
	complement_iterator find_complement(const DNA3Seq &m_seq) const;
	
	// Find the first occurance of either m_seq_a or m_seq_b (i.e. 
	// search for a sequence and its complement). Note that the
	// list-based form of this function is slightly less efficient than
	// the vector-based form.
	iterator find_any(const DNA2Seq &m_seq_a, const DNA2Seq &m_seq_b);
	iterator find_any(const DNA3Seq &m_seq_a, const DNA3Seq &m_seq_b);
	unsigned int find_any(const std::vector<DNA2Seq> &m_seq,
		std::list<unsigned int> &m_match);
	//iterator find_any(const std::vector<DNA3Seq> &m_seq);
	unsigned int find_any(const std::list<DNA2Seq> &m_seq,
		std::list<const DNA2Seq*> &m_match);
	//iterator find_any(const std::list<DNA3Seq> &m_seq);
	
	// It is crucial that this function be inlined!
	bool operator==(const DNA2Seq &m_seq) const
	{
		// "==" is currently defined as strict equality
		if(len == 0){
			return (m_seq.len == 0);
		}
		else{
			//return ((len == m_seq.len) ? (memcmp(seq, m_seq.seq, len*DNA2UnitSizeInByte) == 0): false);
			return ((len == m_seq.len) ? (memcmp(seq, m_seq.seq, len) == 0): false);
		}
		
		// Same as above, but in a "compact" (i.e. hard to read) form
		//return (len == 0) ? (m_seq.len == 0) : 
		//	((len == m_seq.len) ? (memcmp(seq, m_seq.seq, len) == 0): false);

	};
	
	bool operator!=(const DNA2Seq &m_seq) const
	{
		return !(operator==(m_seq));
	};
	
	// To allow use of the DNA2Seq in STL containers, we must define the 
	// comparison operator "<".
	bool operator < (const DNA2Seq &m_seq) const
	{
		if(len == 0){
			return !(m_seq.len == 0); // If both are length 0, "<" is false
								 // because the seq are equal.
		}
		else{
			if(len < m_seq.len){
				return (memcmp(seq, m_seq.seq, len) <= 0);
			}
			
			return (memcmp(seq, m_seq.seq, m_seq.len) < 0);
		}
	};
	
	bool operator > (const DNA2Seq &m_seq) const
	{
		if(len == 0){
			return false; // A zero length sequence can not be less than another
					    // sequence of any length!
		}
		else{
			if(len < m_seq.len){
				//return (memcmp(seq, m_seq.seq, len*DNA2UnitSizeInByte) > 0);
				return (memcmp(seq, m_seq.seq, len) > 0);
			}
			else{
				if(len > m_seq.len){
					//return (memcmp(seq, m_seq.seq, m_seq.len*DNA2UnitSizeInByte) >= 0);
					return (memcmp(seq, m_seq.seq, m_seq.len) >= 0);
				}
				
				// len ==  m_seq.len
				//return (memcmp(seq, m_seq.seq, len*DNA2UnitSizeInByte) > 0);
				return (memcmp(seq, m_seq.seq, len) > 0);
			}
		}
	};
	
	// Convert the sequence to human readable form
	std::string str() const;
	
	// Return the length of the sequence (in bases)
	// inlined for efficiency
	inline unsigned int size() const
	{
		return num_base;
	};
	
	inline bool empty() const
	{
		return (num_base == 0);
	};
	
	// Are all bases in this sequence either A, T, G, or C?
	bool is_ATGC() const
	{
		// This must always return true for DNA2Seq
		return true;
	}
	
	#ifdef USE_MPI
	// Compute the size of this structure in bytes
	unsigned int mpi_size() const
	{	
		// How much information do we need to transmite this
		// structure?
		return (sizeof(num_base) + sizeof(len) + sizeof(seq_remainder) +
			len);
	};
	
	// Copy this structure into a buffer for MPI transport
	// Note that buffer MUST contain at least mpi_size bytes!
	// Return the current buffer position
	unsigned char* mpi_pack(unsigned char* m_buffer) const
	{
		memcpy(m_buffer, &num_base, sizeof(num_base));
		m_buffer += sizeof(num_base);
		
		memcpy(m_buffer, &len, sizeof(len));
		m_buffer += sizeof(len);
		
		memcpy(m_buffer, &seq_remainder, sizeof(seq_remainder));
		m_buffer += sizeof(seq_remainder);
		
		memcpy(m_buffer, seq, len);
		m_buffer += len;
		
		return m_buffer;
	};
	
	// Unpack the structure contained in the buffer
	unsigned char* mpi_unpack(unsigned char* m_buffer)
	{
		// Free previous memory (if any)
		if(seq){
			delete [] seq;
		}
		
		memcpy(&num_base, m_buffer, sizeof(num_base));
		m_buffer += sizeof(num_base);
		
		memcpy(&len, m_buffer, sizeof(len));
		m_buffer += sizeof(len);
				
		memcpy(&seq_remainder, m_buffer, sizeof(seq_remainder));
		m_buffer += sizeof(seq_remainder);
		
		seq = new unsigned char[len];
		
		memcpy(seq, m_buffer, len);
		m_buffer += len;
		
		return m_buffer;
	};
	
	#endif // USE_MPI
};

// Allow direct output sequence  
std::ostream& operator << (std::ostream &s, const DNA2Seq &m_seq);
	
#endif // __BIO_SEQUENCE
