// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"
#include "sequence.h"
#include <iostream>

using namespace std;

/////////////////////////////////////////////////////////////////////
// Code for 3 bit DNA encoding
/////////////////////////////////////////////////////////////////////

// Find the first occurance of the given sequence as a 
// substring in the current sequence
DNA3Seq::iterator DNA3Seq::__find(const DNA3Seq &m_seq,
	const iterator &m_begin, const iterator &m_end,
	const bool &m_periodic /*= false*/) const
{
	int size = m_end - m_begin;

	if(((int)m_seq.size() > size) || !size || !m_seq.size()){
		return end();
	}

	iterator query_iter = m_seq.begin();
	iterator target_iter, start_iter;
	bool found_match = false;
	
	// Note that the "is_valid()" member function is a short cut
	// for "iter != end()" that reduces the number of required 
	// comparisons.
	//for(target_iter = m_begin;target_iter != m_end;target_iter++){

	target_iter = m_begin;

	while(target_iter != m_end){
		if(target_iter.base() == query_iter.base()){
			if(found_match == false){
				start_iter = target_iter;
				
				found_match = true;
			}
			
			query_iter ++;
			
			if(!query_iter.is_valid()){
				return start_iter;
			}

			target_iter++;

			// Allow queries to wrap around periodic targets
			if(m_periodic && (target_iter == m_end)){
				// Rewind to the begining of the sequence, not the
				// begining of the search!
				//target_iter = m_begin;
				target_iter = begin();
			}
		}
		else{
			if(found_match){
				found_match = false;
				
				// Rewind both iterators
				//query_iter = m_seq.begin();
				query_iter.rewind();
				target_iter = start_iter;
			}

			target_iter++;
		}
	}
	
	return end();
}

DNA3Seq::iterator DNA3Seq::__find(const DNA3Seq &m_seq,
								  const bool &m_periodic /*= false*/) const
{
	return __find(m_seq, begin(), end(), m_periodic);
}

DNA3Seq::iterator DNA3Seq::__find(const DNA3Seq &m_seq, 
	const iterator &m_begin, const bool &m_periodic /*= false*/) const
{
	return __find(m_seq, m_begin, end(), m_periodic);
}

DNA3Seq::iterator DNA3Seq::__find(const DNA2Seq &m_seq) const
{
	if((m_seq.size() > size()) || !size() || !m_seq.size()){
		return end();
	}
	
	DNA2Seq::iterator query_iter = m_seq.begin();
	iterator target_iter, start_iter;
	bool found_match = false;
	
	// Note that the "is_valid()" member function is a short cut
	// for "iter != end()" that reduces the number of required 
	// comparisons.
	for(target_iter = begin();target_iter.is_valid();target_iter++){
		if(target_iter.base() == query_iter.base()){
			if(found_match == false){
				start_iter = target_iter;
				
				found_match = true;
			}
			
			query_iter ++;
			
			if(!query_iter.is_valid()){
				return start_iter;
			}
		}
		else{
			if(found_match){
				found_match = false;
				
				// Rewind both iterators
				//query_iter = m_seq.begin();
				query_iter.rewind();
				target_iter = start_iter;
			}
		}
	}
	
	return end();
}

// Find the first occurance of the given sequence as a 
// substring in the complement of the current sequence
DNA3Seq::complement_iterator DNA3Seq::__find_complement(const DNA3Seq &m_seq) const
{
	if((m_seq.size() > size()) || !size() || !m_seq.size()){
		return complement_end();
	}
	
	iterator query_iter = m_seq.begin();
	complement_iterator target_iter, start_iter;
	bool found_match = false;
	
	// Note that the "is_valid()" member function is a short cut
	// for "iter != end()" that reduces the number of required 
	// comparisons.
	for(target_iter = complement_begin();target_iter.is_valid();target_iter++){
		if(target_iter.base() == query_iter.base()){
			if(found_match == false){
				start_iter = target_iter;
				
				found_match = true;
			}
			
			query_iter ++;
			
			if(!query_iter.is_valid()){
				return start_iter;
			}
		}
		else{
			if(found_match){
				found_match = false;
				
				// Rewind both iterators
				query_iter.rewind();
				target_iter = start_iter;
			}
		}
	}
	
	return complement_end();
}

DNA3Seq::complement_iterator DNA3Seq::__find_complement(const DNA2Seq &m_seq) const
{
	if((m_seq.size() > size()) || !size() || !m_seq.size()){
		return complement_end();
	}
	
	DNA2Seq::iterator query_iter = m_seq.begin();
	complement_iterator target_iter, start_iter;
	bool found_match = false;
	
	// Note that the "is_valid()" member function is a short cut
	// for "iter != end()" that reduces the number of required 
	// comparisons.
	for(target_iter = complement_begin();target_iter.is_valid();target_iter++){
		if(target_iter.base() == query_iter.base()){
			if(found_match == false){
				start_iter = target_iter;
				
				found_match = true;
			}
			
			query_iter ++;
			
			if(!query_iter.is_valid()){
				return start_iter;
			}
		}
		else{
			if(found_match){
				found_match = false;
				
				// Rewind both iterators
				query_iter.rewind();
				target_iter = start_iter;
			}
		}
	}
	
	return complement_end();
}

DNA3Seq::iterator DNA3Seq::find(const DNA3Seq &m_seq, 
	const iterator &m_begin, const iterator &m_end,
	const bool &m_periodic /*= false*/) const
{
	int size = m_end - m_begin;	
	int comp_len = size;
	
	// JDG 8/7/06 
	// Commented out the code below (should not subtract m_seq.size(), but
	// rather [m_seq.size() - DNA3UnitSizeInBase] to accound for bases already
	// in the query_buffer. Correct was not needed anyway (the optimization loss
	// is minor and the code is now clearer).
	//if(m_periodic == false){
	//	comp_len -= m_seq.size();
	//}

	
	if( (comp_len < 0) || !size || !m_seq.size() ){
		return end();
	}
	
	if(m_seq.size() < DNA3UnitSizeInBase){
		return __find(m_seq, m_begin, m_end, m_periodic);
	}
	
	iterator target_iter = m_begin;
	
	// This implementation assumes that sizeof(int) >= 3 which should be
	// true for evey modern OS.
	unsigned int comp_buffer, query_buffer;
	
	// Load the comp_buffer with data from the target seq
	comp_buffer = target_iter.base();
	target_iter ++;
	
	for(unsigned int j = 1;j < DNA3UnitSizeInBase;j++){
		comp_buffer = (comp_buffer << 3) + target_iter.base();
		target_iter ++;
	}
	
	query_buffer = (m_seq.seq[0] << 16) + (m_seq.seq[1] << 8) + m_seq.seq[2];
	
	// BUG fixed 1/28/06 J. D. Gans
	// Changed -> for(int i = 0;i < comp_len;i++){
	// to the line you see below:
	for(int i = target_iter - m_begin;i < comp_len;i++){
	
		// By packing the sequence data into an unsigned int, we can perform
		// fast comparisons.
		if(comp_buffer == query_buffer){
			
			// Explore possible match in detail
			iterator query_iter = m_seq.begin() + DNA3UnitSizeInBase;
			iterator tmp_iter = target_iter;
			
			while(true){
			
				if( !query_iter.is_valid() ){
				
					// We found a match!
					return target_iter - DNA3UnitSizeInBase;
				}
				
				if( (tmp_iter == m_end) || 
				   ( tmp_iter.base() != query_iter.base() ) ){
				   
					break;
				}
				
				tmp_iter++;
				
				// Wrap around periodic boundary conditions
				if( m_periodic && (tmp_iter == m_end) ){
				
					// Wrap around to the begining of the sequence, not
					// the start of the search!
					//tmp_iter = m_begin;
					tmp_iter = begin();
				}

				query_iter++;
			}
		}
		
		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = ( (comp_buffer << 3) + target_iter.base() ) & 16777215;
		
		target_iter++;

		if( m_periodic && (target_iter == m_end) ){
			
			// Wrap around to the begining of the sequence, not
			// the start of the search!
			//target_iter = m_begin;
			target_iter = begin();
		}
	}
	
	// By packing the sequence data into an unsigned int, we can perform
	// fast comparisons.
	if(comp_buffer == query_buffer){
			
		// Explore possible match in detail
		iterator query_iter = m_seq.begin() + DNA3UnitSizeInBase;
		iterator tmp_iter = target_iter;
			
		while(true){
		
			if( !query_iter.is_valid() ){
			
				// We found a match!
				return target_iter - DNA3UnitSizeInBase;
			}
				
			if( (tmp_iter == m_end) || 
			   ( tmp_iter.base() != query_iter.base() ) ){
			   
				break;
			}
				
			tmp_iter++;

			// Wrap around periodic boundary conditions
			if( m_periodic && (tmp_iter == m_end) ){

				// Wrap around to the begining of the sequence, not
				// the start of the search!
				//tmp_iter = m_begin;
				tmp_iter = begin();
			}

			query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
		
	return end();
}

DNA3Seq::iterator DNA3Seq::find(const DNA3Seq &m_seq,
								const bool &m_periodic /*= false*/) const
{
	return find(m_seq, begin(), end(), m_periodic);
}


DNA3Seq::iterator DNA3Seq::find(const DNA3Seq &m_seq,
	const iterator &m_begin, const bool &m_periodic /*= false*/) const
{
	return find(m_seq, m_begin, end(), m_periodic);
}

DNA3Seq::iterator DNA3Seq::find(const DNA2Seq &m_seq) const
{
	int comp_len = size() - m_seq.size();
	
	if((comp_len < 0) || !size() || !m_seq.size()){
		return end();
	}
	
	if(m_seq.size() < DNA3UnitSizeInBase){
		return __find(m_seq);
	}
	
	iterator target_iter = begin() + DNA3UnitSizeInBase;
	DNA2Seq::iterator query_iter = m_seq.begin();
	
	// This implementation assumes that sizeof(int) >= 3 which should be
	// true for evey modern OS.
	unsigned int comp_buffer, query_buffer;
	
	// Load the comp_buffer with data from the target seq
	comp_buffer = (seq[0] << 16) + (seq[1] << 8) + seq[2];
	
	//query_buffer = (m_seq.seq[0] << 16) + (m_seq.seq[1] << 8) + m_seq.seq[2];
	query_buffer = query_iter.base();
	query_iter++;
	
	for(int j = 1;j < DNA3UnitSizeInBase;j++){
		query_buffer = (query_buffer << 3) + query_iter.base();
		query_iter++;
	}
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned int, we can perform
		// fast comparisons.
		if(comp_buffer == query_buffer){
			
			// Explore possible match in detail
			DNA2Seq::iterator tmp_query_iter = query_iter;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!tmp_query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA3UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != tmp_query_iter.base())){
					break;
				}
				
				tmp_iter++;
				tmp_query_iter++;
			}
		}
						
		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = ((comp_buffer << 3) + target_iter.base()) & 16777215;
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned int, we can perform
	// fast comparisons.
	if(comp_buffer == query_buffer){
			
		// Explore possible match in detail
		DNA2Seq::iterator tmp_query_iter = query_iter;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!tmp_query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA3UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != tmp_query_iter.base())){
				break;
			}
				
			tmp_iter++;
			tmp_query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	return end();
}

DNA3Seq::iterator DNA3Seq::find_any(const DNA3Seq &m_seq_a, const DNA3Seq &m_seq_b)
{
	unsigned int min_query_size = min(m_seq_a.size(), m_seq_b.size());

	int comp_len = size() - min_query_size;
	
	if((comp_len < 0) || !size() || !min_query_size){
		return end();
	}
	
	if(min_query_size < DNA3UnitSizeInBase){
		iterator tmp_iter = __find(m_seq_a);

		if(tmp_iter != end()){
			return tmp_iter;
		}

		return __find(m_seq_b);
	}
	
	iterator target_iter = begin() + DNA3UnitSizeInBase;
	
	// This implementation assumes that sizeof(int) >= 3 which should be
	// true for evey modern OS.
	unsigned int comp_buffer, 
				 query_buffer_a,
				 query_buffer_b;
	
	// Load the comp_buffer with data from the target seq
	comp_buffer = (seq[0] << 16) + (seq[1] << 8) + seq[2];
	query_buffer_a = (m_seq_a.seq[0] << 16) + (m_seq_a.seq[1] << 8) + m_seq_a.seq[2];
	query_buffer_b = (m_seq_b.seq[0] << 16) + (m_seq_b.seq[1] << 8) + m_seq_b.seq[2];
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned int, we can perform
		// fast comparisons.
		if(comp_buffer == query_buffer_a){
			
			// Explore possible match in detail
			iterator query_iter = m_seq_a.begin() + DNA3UnitSizeInBase;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA3UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}
				
				tmp_iter++;
				query_iter++;
			}
		}
		
		if(comp_buffer == query_buffer_b){
			
			// Explore possible match in detail
			iterator query_iter = m_seq_b.begin() + DNA3UnitSizeInBase;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA3UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}
				
				tmp_iter++;
				query_iter++;
			}
		}

		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = ((comp_buffer << 3) + target_iter.base()) & 16777215;
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned int, we can perform
	// fast comparisons.
	if(comp_buffer == query_buffer_a){
			
		// Explore possible match in detail
		iterator query_iter = m_seq_a.begin() + DNA3UnitSizeInBase;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA3UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != query_iter.base())){
				break;
			}
				
			tmp_iter++;
			query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	if(comp_buffer == query_buffer_b){
			
		// Explore possible match in detail
		iterator query_iter = m_seq_b.begin() + DNA3UnitSizeInBase;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA3UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != query_iter.base())){
				break;
			}
				
			tmp_iter++;
			query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}

	return end();
}

DNA3Seq::iterator DNA3Seq::find_any(const DNA2Seq &m_seq_a, const DNA2Seq &m_seq_b)
{
	unsigned int min_query_size = min(m_seq_a.size(), m_seq_b.size());

	int comp_len = size() - min_query_size;
	
	if((comp_len < 0) || !size() || !min_query_size){
		return end();
	}
	
	if(min_query_size < DNA3UnitSizeInBase){
		iterator tmp_iter = __find(m_seq_a);

		if(tmp_iter != end()){
			return tmp_iter;
		}

		return __find(m_seq_b);
	}
	
	iterator target_iter = begin() + DNA3UnitSizeInBase;
	
	DNA2Seq::iterator query_iter_a = m_seq_a.begin();
	DNA2Seq::iterator query_iter_b = m_seq_b.begin();
	
	// This implementation assumes that sizeof(int) >= 3 which should be
	// true for evey modern OS.
	unsigned int comp_buffer, query_buffer_a, query_buffer_b;
	
	// Load the comp_buffer with data from the target seq
	comp_buffer = (seq[0] << 16) + (seq[1] << 8) + seq[2];
	//query_buffer_a = (m_seq_a.seq[0] << 16) + (m_seq_a.seq[1] << 8) + m_seq_a.seq[2];
	//query_buffer_b = (m_seq_b.seq[0] << 16) + (m_seq_b.seq[1] << 8) + m_seq_b.seq[2];
	
	query_buffer_a = query_iter_a.base();
	query_iter_a++;
	query_buffer_b = query_iter_b.base();
	query_iter_b++;
	
	for(int j = 1;j < DNA3UnitSizeInBase;j++){
		query_buffer_a = (query_buffer_a << 3) + query_iter_a.base();
		query_iter_a++;
		query_buffer_b = (query_buffer_b << 3) + query_iter_b.base();
		query_iter_b++;
	}
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned int, we can perform
		// fast comparisons.
		if(comp_buffer == query_buffer_a){
			
			// Explore possible match in detail
			DNA2Seq::iterator tmp_query_iter = query_iter_a;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!tmp_query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA3UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != tmp_query_iter.base())){
					break;
				}
				
				tmp_iter++;
				tmp_query_iter++;
			}
		}
		
		if(comp_buffer == query_buffer_b){
			
			// Explore possible match in detail
			DNA2Seq::iterator tmp_query_iter = query_iter_b;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!tmp_query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA3UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != tmp_query_iter.base())){
					break;
				}
				
				tmp_iter++;
				tmp_query_iter++;
			}
		}

		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = ((comp_buffer << 3) + target_iter.base()) & 16777215;
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned int, we can perform
	// fast comparisons.
	if(comp_buffer == query_buffer_a){
			
		// Explore possible match in detail
		DNA2Seq::iterator tmp_query_iter = query_iter_a;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!tmp_query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA3UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != query_iter_a.base())){
				break;
			}
				
			tmp_iter++;
			tmp_query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	if(comp_buffer == query_buffer_b){
			
		// Explore possible match in detail
		DNA2Seq::iterator tmp_query_iter = query_iter_b;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!tmp_query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA3UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != tmp_query_iter.base())){
				break;
			}
				
			tmp_iter++;
			tmp_query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}

	return end();
}

// Return the number of matches between this sequence and the entries of
// m_seq. Count multiple matches between the same sequences.
unsigned int DNA3Seq::find_any(const vector<DNA3Seq> &m_seq,
	std::list<unsigned int> &m_match)
{
	unsigned int i, j;
	unsigned int num_query = m_seq.size();
	unsigned int match_count = 0;
	
	if(num_query == 0){
		return match_count;
	}
	
	unsigned int min_query_size = m_seq[0].size();
	
	for(i = 1;i < num_query;i++){
		min_query_size = min(min_query_size, m_seq[i].size());
	}
	
	int comp_len = size() - min_query_size;
	
	if((comp_len < 0) || !size() || !min_query_size){
		return match_count;
	}
	
	if(min_query_size < DNA3UnitSizeInBase){
	
		for(i = 0;i < num_query;i++){
			iterator tmp_iter = __find(m_seq[i]);

			if(tmp_iter != end()){
				match_count ++;
				m_match.push_back(i);
			}
		}
		return match_count;
	}
	
	iterator target_iter = begin() + DNA3UnitSizeInBase;
	
	// This implementation assumes that sizeof(int) >= 3 which should be
	// true for evey modern OS.
	unsigned int comp_buffer;
	vector<unsigned int> query_buffer(num_query);
	
	// Load the comp_buffer with data from the target seq
	comp_buffer = (seq[0] << 16) + (seq[1] << 8) + seq[2];
	
	for(i = 0;i < num_query;i++){
		const unsigned char *ptr = m_seq[i].seq;
		
		query_buffer[i] = (ptr[0] << 16) + (ptr[1] << 8) + ptr[2];
	}
	
	for(i = 0;i < (unsigned int)comp_len;i++){
		// By packing the sequence data into an unsigned int, we can perform
		// fast comparisons.
		for(j = 0;j < num_query;j++){
			if(comp_buffer == query_buffer[j]){

				// Explore possible match in detail
				iterator query_iter = m_seq[j].begin() + DNA3UnitSizeInBase;
				iterator tmp_iter = target_iter;

				while(true){
					if(!query_iter.is_valid()){
						// We found a match!
						match_count ++;
						m_match.push_back(j);
						break;
					}

					if(!tmp_iter.is_valid() || 
					   (tmp_iter.base() != query_iter.base())){
						break;
					}

					tmp_iter++;
					query_iter++;
				}
			}
		}

		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = ((comp_buffer << 3) + target_iter.base()) & 16777215;
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned int, we can perform
	// fast comparisons.
	for(j = 0;j < num_query;j++){
		if(comp_buffer == query_buffer[j]){

			// Explore possible match in detail
			iterator query_iter = m_seq[j].begin() + DNA3UnitSizeInBase;
			iterator tmp_iter = target_iter;

			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					match_count ++;
					m_match.push_back(j);
					break;
				}

				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}

				tmp_iter++;
				query_iter++;
			}

			// If we get here without a match, then there is no
			// match!
		}
	}

	return match_count;
}

unsigned int DNA3Seq::find_any(const list<DNA3Seq> &m_seq, std::list<const DNA3Seq*> &m_match)
{
	unsigned int i, j, k;
	unsigned int num_query = m_seq.size();
	unsigned int match_count = 0;
	
	if(num_query == 0){
		return match_count;
	}
	
	list<DNA3Seq>::const_iterator list_iter = m_seq.begin();
	
	unsigned int min_query_size = list_iter->size();
	
	for(;list_iter != m_seq.end();list_iter++){
		min_query_size = min(min_query_size, list_iter->size());
	}
	
	int comp_len = size() - min_query_size;
	
	if((comp_len < 0) || !size() || !min_query_size){
		return match_count;
	}
	
	if(min_query_size < DNA3UnitSizeInBase){
	
		for(list_iter = m_seq.begin();list_iter != m_seq.end();list_iter++){
			iterator tmp_iter = __find(*list_iter);

			if(tmp_iter != end()){
				match_count++;
				m_match.push_back(&(*list_iter));
			}
		}
		return match_count;
	}
	
	iterator target_iter = begin() + DNA3UnitSizeInBase;
	
	// This implementation assumes that sizeof(int) >= 3 which should be
	// true for evey modern OS.
	unsigned int comp_buffer;
	vector<unsigned int> query_buffer(num_query);
	
	// Load the comp_buffer with data from the target seq
	comp_buffer = (seq[0] << 16) + (seq[1] << 8) + seq[2];
	
	for(i = 0, list_iter = m_seq.begin();i < num_query;i++, list_iter++){		
		query_buffer[i] = (list_iter->seq[0] << 16) + (list_iter->seq[1] << 8) + 
			list_iter->seq[2];
	}
	
	for(i = 0;i < (unsigned int)comp_len;i++){
		// By packing the sequence data into an unsigned int, we can perform
		// fast comparisons.
		for(j = 0;j < num_query;j++){
			if(comp_buffer == query_buffer[j]){
				
				// Position the list iterator on the j^th element
				for(k = 0, list_iter = m_seq.begin();k < j;k++, list_iter++);
				
				// Explore possible match in detail
				iterator query_iter = list_iter->begin() + DNA3UnitSizeInBase;
				iterator tmp_iter = target_iter;

				while(true){
					if(!query_iter.is_valid()){
						// We found a match!
						match_count++;
						m_match.push_back(&(*list_iter));
						break;
					}

					if(!tmp_iter.is_valid() || 
					   (tmp_iter.base() != query_iter.base())){
						break;
					}

					tmp_iter++;
					query_iter++;
				}
			}
		}

		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = ((comp_buffer << 3) + target_iter.base()) & 16777215;
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned int, we can perform
	// fast comparisons.
	for(j = 0;j < num_query;j++){
		if(comp_buffer == query_buffer[j]){
			
			// Position the list iterator on the j^th element
			for(k = 0, list_iter = m_seq.begin();k < j;k++, list_iter++);
			
			// Explore possible match in detail
			iterator query_iter = list_iter->begin() + DNA3UnitSizeInBase;
			iterator tmp_iter = target_iter;

			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					match_count++;
					m_match.push_back(&(*list_iter));
					break;
				}

				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}

				tmp_iter++;
				query_iter++;
			}

			// If we get here without a match, then there is no
			// match!
		}
	}

	return match_count;
}

DNA3Seq::complement_iterator DNA3Seq::find_complement(const DNA3Seq &m_seq) const
{
	int comp_len = size() - m_seq.size();
	
	if((comp_len < 0) || !size() || !m_seq.size()){
		return complement_end();
	}
	
	if(m_seq.size() < DNA3UnitSizeInBase){
		return __find_complement(m_seq);
	}
	
	complement_iterator target_iter = complement_begin();
	
	// This implementation assumes that sizeof(int) >= 3 which should be
	// true for evey modern OS.
	unsigned int comp_buffer, query_buffer;
	
	// Load the comp_buffer with data from the target seq
	// This loop (which adds DNA3UnitSizeInBase based to comp_buffer)
	// has been unrolled.
	comp_buffer = target_iter.base();
	target_iter++;
	comp_buffer = ((comp_buffer << 3) + target_iter.base());
	target_iter++;
	comp_buffer = ((comp_buffer << 3) + target_iter.base());
	target_iter++;
	comp_buffer = ((comp_buffer << 3) + target_iter.base());
	target_iter++;
	comp_buffer = ((comp_buffer << 3) + target_iter.base());
	target_iter++;
	comp_buffer = ((comp_buffer << 3) + target_iter.base());
	target_iter++;
	comp_buffer = ((comp_buffer << 3) + target_iter.base());
	target_iter++;
	comp_buffer = ((comp_buffer << 3) + target_iter.base());
	target_iter++;

	query_buffer = (m_seq.seq[0] << 16) + (m_seq.seq[1] << 8) + m_seq.seq[2];
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned int, we can perform
		// fast comparisons.
		if(comp_buffer == query_buffer){
			
			// Explore possible match in detail
			iterator query_iter = m_seq.begin() + DNA3UnitSizeInBase;
			complement_iterator tmp_iter = target_iter;
			
			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA3UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}
				
				tmp_iter++;
				query_iter++;
			}
		}
						
		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = ((comp_buffer << 3) + target_iter.base()) & 16777215;
		
		target_iter++;
	}
	
	
	// By packing the sequence data into an unsigned int, we can perform
	// fast comparisons.
	if(comp_buffer == query_buffer){
			
		// Explore possible match in detail
		iterator query_iter = m_seq.begin() + DNA3UnitSizeInBase;
		complement_iterator tmp_iter = target_iter;
			
		while(true){
			if(!query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA3UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != query_iter.base())){
				break;
			}
				
			tmp_iter++;
			query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	return complement_end();
}

DNA3Seq::complement_iterator DNA3Seq::find_complement(const DNA2Seq &m_seq) const
{
	int comp_len = size() - m_seq.size();
	
	if((comp_len < 0) || !size() || !m_seq.size()){
		return complement_end();
	}
	
	if(m_seq.size() < DNA3UnitSizeInBase){
		return __find_complement(m_seq);
	}
	
	complement_iterator target_iter = complement_begin();
	DNA2Seq::iterator query_iter = m_seq.begin();
	
	// This implementation assumes that sizeof(int) >= 3 which should be
	// true for evey modern OS.
	unsigned int comp_buffer, query_buffer;
	
	//query_buffer = (m_seq.seq[0] << 16) + (m_seq.seq[1] << 8) + m_seq.seq[2];
	
	// Load the comp_buffer with data from the target seq
	// This loop (which adds DNA3UnitSizeInBase based to comp_buffer)
	// has been unrolled.
	comp_buffer = target_iter.base();
	target_iter++;
	
	query_buffer = query_iter.base();
	query_iter++;
	
	for(int j = 1;j < DNA3UnitSizeInBase;j++){
		comp_buffer = (comp_buffer << 3) + target_iter.base();
		target_iter++;
		
		query_buffer = (query_buffer << 3) + query_iter.base();
		query_iter++;
	}
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned int, we can perform
		// fast comparisons.
		if(comp_buffer == query_buffer){
			
			// Explore possible match in detail
			DNA2Seq::iterator tmp_query_iter = query_iter;
			complement_iterator tmp_iter = target_iter;
			
			while(true){
				if(!tmp_query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA3UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != tmp_query_iter.base())){
					break;
				}
				
				tmp_iter++;
				tmp_query_iter++;
			}
		}
						
		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = ((comp_buffer << 3) + target_iter.base()) & 16777215;
		
		target_iter++;
	}
	
	
	// By packing the sequence data into an unsigned int, we can perform
	// fast comparisons.
	if(comp_buffer == query_buffer){
			
		// Explore possible match in detail
		DNA2Seq::iterator tmp_query_iter = query_iter;
		complement_iterator tmp_iter = target_iter;
			
		while(true){
			if(!tmp_query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA3UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != tmp_query_iter.base())){
				break;
			}
				
			tmp_iter++;
			tmp_query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	return complement_end();
}

/////////////////////////////////////////////////////////////////////
// Code for 2 bit DNA encoding
/////////////////////////////////////////////////////////////////////

// Find the first occurance of the given sequence as a 
// substring in the current sequence
DNA2Seq::iterator DNA2Seq::__find(const DNA2Seq &m_seq,
	const iterator &m_begin, const iterator &m_end) const
{
	int size = m_end - m_begin;
	
	if(((int)m_seq.size() > size) || !size || !m_seq.size()){
		return end();
	}
	
	iterator query_iter = m_seq.begin();
	iterator target_iter, start_iter;
	bool found_match = false;
	
	// Note that the "is_valid()" member function is a short cut
	// for "iter != end()" that reduces the number of required 
	// comparisons.
	for(target_iter =m_begin;target_iter != m_end;target_iter++){
		if(target_iter.base() == query_iter.base()){
			if(found_match == false){
				start_iter = target_iter;
				
				found_match = true;
			}
			
			query_iter ++;
			
			if(!query_iter.is_valid()){
				return start_iter;
			}
		}
		else{
			if(found_match){
				found_match = false;
				
				// Rewind both iterators
				//query_iter = m_seq.begin();
				query_iter.rewind();
				target_iter = start_iter;
			}
		}
	}
	
	return end();
}

DNA2Seq::iterator DNA2Seq::__find(const DNA2Seq &m_seq,
	const iterator &m_begin) const
{
	return __find(m_seq, m_begin, end());
}

DNA2Seq::iterator DNA2Seq::__find(const DNA2Seq &m_seq) const
{
	return __find(m_seq, begin(), end());
}

DNA2Seq::iterator DNA2Seq::__find(const DNA3Seq &m_seq) const
{
	if((m_seq.size() > size()) || !size() || !m_seq.size()){
		return end();
	}
	
	DNA3Seq::iterator query_iter = m_seq.begin();
	iterator target_iter, start_iter;
	bool found_match = false;
	
	// Note that the "is_valid()" member function is a short cut
	// for "iter != end()" that reduces the number of required 
	// comparisons.
	for(target_iter = begin();target_iter.is_valid();target_iter++){
		if(target_iter.base() == query_iter.base()){
			if(found_match == false){
				start_iter = target_iter;
				
				found_match = true;
			}
			
			query_iter ++;
			
			if(!query_iter.is_valid()){
				return start_iter;
			}
		}
		else{
			if(found_match){
				found_match = false;
				
				// Rewind both iterators
				//query_iter = m_seq.begin();
				query_iter.rewind();
				target_iter = start_iter;
			}
		}
	}
	
	return end();
}

// Find the first occurance of the given sequence as a 
// substring in the complement of the current sequence
DNA2Seq::complement_iterator DNA2Seq::__find_complement(const DNA2Seq &m_seq) const
{
	if((m_seq.size() > size()) || !size() || !m_seq.size()){
		return complement_end();
	}
	
	iterator query_iter = m_seq.begin();
	complement_iterator target_iter, start_iter;
	bool found_match = false;
	
	// Note that the "is_valid()" member function is a short cut
	// for "iter != end()" that reduces the number of required 
	// comparisons.
	for(target_iter = complement_begin();target_iter.is_valid();target_iter++){
		if(target_iter.base() == query_iter.base()){
			if(found_match == false){
				start_iter = target_iter;
				
				found_match = true;
			}
			
			query_iter ++;
			
			if(!query_iter.is_valid()){
				return start_iter;
			}
		}
		else{
			if(found_match){
				found_match = false;
				
				// Rewind both iterators
				query_iter.rewind();
				target_iter = start_iter;
			}
		}
	}
	
	return complement_end();
}

DNA2Seq::complement_iterator DNA2Seq::__find_complement(const DNA3Seq &m_seq) const
{
	if((m_seq.size() > size()) || !size() || !m_seq.size()){
		return complement_end();
	}
	
	DNA3Seq::iterator query_iter = m_seq.begin();
	complement_iterator target_iter, start_iter;
	bool found_match = false;
	
	// Note that the "is_valid()" member function is a short cut
	// for "iter != end()" that reduces the number of required 
	// comparisons.
	for(target_iter = complement_begin();target_iter.is_valid();target_iter++){
		if(target_iter.base() == query_iter.base()){
			if(found_match == false){
				start_iter = target_iter;
				
				found_match = true;
			}
			
			query_iter ++;
			
			if(!query_iter.is_valid()){
				return start_iter;
			}
		}
		else{
			if(found_match){
				found_match = false;
				
				// Rewind both iterators
				query_iter.rewind();
				target_iter = start_iter;
			}
		}
	}
	
	return complement_end();
}

DNA2Seq::iterator DNA2Seq::find(const DNA2Seq &m_seq,
	const iterator &m_begin, const iterator &m_end) const
{
	int size = m_end - m_begin;
	
	int comp_len = size - m_seq.size();
	
	if((comp_len < 0) || !size || !m_seq.size()){
		return end();
	}
	
	if(m_seq.size() < DNA2UnitSizeInBase){
		return __find(m_seq, m_begin, m_end);
	}
	
	iterator target_iter = m_begin;
	
	// Load the comp_buffer with data from the target seq
	unsigned char comp_buffer = target_iter.base();
	target_iter++;
	
	comp_buffer = (comp_buffer << 2) + target_iter.base();
	target_iter++;
	
	comp_buffer = (comp_buffer << 2) + target_iter.base();
	target_iter++;
	
	comp_buffer = (comp_buffer << 2) + target_iter.base();
	target_iter++;
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned char, we can perform
		// fast comparisons.
		if(comp_buffer == m_seq.seq[0]){
			
			// Explore possible match in detail
			iterator query_iter = m_seq.begin() + DNA2UnitSizeInBase;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA2UnitSizeInBase;
				}
				
				if((tmp_iter == m_end) || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}
				
				tmp_iter++;
				query_iter++;
			}
		}
						
		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = (comp_buffer << 2) + target_iter.base();
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned char, we can perform
	// fast comparisons.
	if(comp_buffer == m_seq.seq[0]){
			
		// Explore possible match in detail
		iterator query_iter = m_seq.begin() + DNA2UnitSizeInBase;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA2UnitSizeInBase;
			}
				
			if((tmp_iter == m_end) || 
			   (tmp_iter.base() != query_iter.base())){
				break;
			}
				
			tmp_iter++;
			query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	return end();
}

DNA2Seq::iterator DNA2Seq::find(const DNA2Seq &m_seq) const
{
	return find(m_seq, begin(), end());
}


DNA2Seq::iterator DNA2Seq::find(const DNA2Seq &m_seq,
	const iterator &m_begin) const
{
	return find(m_seq, m_begin, end());
}

DNA2Seq::iterator DNA2Seq::find(const DNA3Seq &m_seq) const
{
	int comp_len = size() - m_seq.size();
	
	if((comp_len < 0) || !size() || !m_seq.size()){
		return end();
	}
	
	if(m_seq.size() < DNA2UnitSizeInBase){
		return __find(m_seq);
	}
	
	DNA3Seq::iterator query_iter = m_seq.begin();
	iterator target_iter = begin() + DNA2UnitSizeInBase;
	
	// Load the comp_buffer with data from the target seq
	unsigned char comp_buffer = seq[0];
	unsigned char query_buffer;
	
	query_buffer = query_iter.base();
	query_iter++;
	
	for(int j = 1;j < DNA2UnitSizeInBase;j++){
		query_buffer = (query_buffer << 2) + query_iter.base();
		query_iter++;
	}
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned char, we can perform
		// fast comparisons.
		if(comp_buffer == query_buffer){
			
			// Explore possible match in detail
			DNA3Seq::iterator tmp_query_iter = query_iter;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!tmp_query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA2UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != tmp_query_iter.base())){
					break;
				}
				
				tmp_iter++;
				tmp_query_iter++;
			}
		}
						
		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = (comp_buffer << 2) + target_iter.base();
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned char, we can perform
	// fast comparisons.
	if(comp_buffer == query_buffer){
			
		// Explore possible match in detail
		DNA3Seq::iterator tmp_query_iter = query_iter;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!tmp_query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA2UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != tmp_query_iter.base())){
				break;
			}
				
			tmp_iter++;
			tmp_query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	return end();
}

DNA2Seq::iterator DNA2Seq::find_any(const DNA2Seq &m_seq_a, const DNA2Seq &m_seq_b)
{
	unsigned int min_query_size = min(m_seq_a.size(), m_seq_b.size());

	int comp_len = size() - min_query_size;
	
	if((comp_len < 0) || !size() || !min_query_size){
		return end();
	}
	
	if(min_query_size < DNA2UnitSizeInBase){
		iterator tmp_iter = __find(m_seq_a);

		if(tmp_iter != end()){
			return tmp_iter;
		}

		return __find(m_seq_b);
	}
	
	iterator target_iter = begin() + DNA2UnitSizeInBase;
	
	// Load the comp_buffer with data from the target seq
	unsigned char comp_buffer = seq[0];
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned char, we can perform
		// fast comparisons.
		if(comp_buffer == m_seq_a.seq[0]){
			
			// Explore possible match in detail
			iterator query_iter = m_seq_a.begin() + DNA2UnitSizeInBase;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA2UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}
				
				tmp_iter++;
				query_iter++;
			}
		}
		
		if(comp_buffer == m_seq_b.seq[0]){
			
			// Explore possible match in detail
			iterator query_iter = m_seq_b.begin() + DNA2UnitSizeInBase;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA2UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}
				
				tmp_iter++;
				query_iter++;
			}
		}

		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = (comp_buffer << 2) + target_iter.base();
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned char, we can perform
	// fast comparisons.
	if(comp_buffer == m_seq_a.seq[0]){
			
		// Explore possible match in detail
		iterator query_iter = m_seq_a.begin() + DNA2UnitSizeInBase;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA2UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != query_iter.base())){
				break;
			}
				
			tmp_iter++;
			query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	if(comp_buffer == m_seq_b.seq[0]){
			
		// Explore possible match in detail
		iterator query_iter = m_seq_b.begin() + DNA2UnitSizeInBase;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA2UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != query_iter.base())){
				break;
			}
				
			tmp_iter++;
			query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}

	return end();
}

DNA2Seq::iterator DNA2Seq::find_any(const DNA3Seq &m_seq_a, const DNA3Seq &m_seq_b)
{
	unsigned int min_query_size = min(m_seq_a.size(), m_seq_b.size());

	int comp_len = size() - min_query_size;
	
	if((comp_len < 0) || !size() || !min_query_size){
		return end();
	}
	
	if(min_query_size < DNA2UnitSizeInBase){
		iterator tmp_iter = __find(m_seq_a);

		if(tmp_iter != end()){
			return tmp_iter;
		}

		return __find(m_seq_b);
	}
	
	iterator target_iter = begin() + DNA2UnitSizeInBase;
	DNA3Seq::iterator query_iter_a = m_seq_a.begin();
	DNA3Seq::iterator query_iter_b = m_seq_b.begin();
	
	// Load the comp_buffer with data from the target seq
	unsigned char comp_buffer = seq[0];
	unsigned char query_buffer_a, query_buffer_b;
	
	query_buffer_a = query_iter_a.base();
	query_iter_a++;
	
	query_buffer_b = query_iter_b.base();
	query_iter_b++;
	
	for(int j = 1;j < DNA2UnitSizeInBase;j++){
		query_buffer_a = (query_buffer_a << 2) + query_iter_a.base();
		query_iter_a++;
	
		query_buffer_b = (query_buffer_b << 2) + query_iter_b.base();
		query_iter_b++;
	}
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned char, we can perform
		// fast comparisons.
		if(comp_buffer == query_buffer_a){
			
			// Explore possible match in detail
			DNA3Seq::iterator tmp_query_iter = query_iter_a;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!tmp_query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA2UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != tmp_query_iter.base())){
					break;
				}
				
				tmp_iter++;
				tmp_query_iter++;
			}
		}
		
		if(comp_buffer == query_buffer_b){
			
			// Explore possible match in detail
			DNA3Seq::iterator tmp_query_iter = query_iter_b;
			iterator tmp_iter = target_iter;
			
			while(true){
				if(!tmp_query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA2UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != tmp_query_iter.base())){
					break;
				}
				
				tmp_iter++;
				tmp_query_iter++;
			}
		}

		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = (comp_buffer << 2) + target_iter.base();
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned char, we can perform
	// fast comparisons.
	if(comp_buffer == query_buffer_a){
			
		// Explore possible match in detail
		DNA3Seq::iterator tmp_query_iter = query_iter_a;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!tmp_query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA2UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != tmp_query_iter.base())){
				break;
			}
				
			tmp_iter++;
			tmp_query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	if(comp_buffer == query_buffer_b){
			
		// Explore possible match in detail
		DNA3Seq::iterator tmp_query_iter = query_iter_b;
		iterator tmp_iter = target_iter;
			
		while(true){
			if(!tmp_query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA2UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != tmp_query_iter.base())){
				break;
			}
				
			tmp_iter++;
			tmp_query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}

	return end();
}

unsigned int DNA2Seq::find_any(const vector<DNA2Seq> &m_seq, std::list<unsigned int> &m_match)
{
	unsigned int i, j;
	unsigned int num_query = m_seq.size();
	unsigned int match_count = 0;
	
	if(num_query == 0){
		return match_count;
	}
	
	unsigned int min_query_size = m_seq[0].size();
	
	for(i = 1;i < num_query;i++){
		min_query_size = min(min_query_size, m_seq[i].size());
	}
	
	int comp_len = size() - min_query_size;
	
	if((comp_len < 0) || !size() || !min_query_size){
		return match_count;
	}
	
	if(min_query_size < DNA2UnitSizeInBase){
	
		for(i = 0;i < num_query;i++){
			iterator tmp_iter = __find(m_seq[i]);

			if(tmp_iter != end()){
				match_count++;
				m_match.push_back(i);
			}
		}
		return match_count;
	}
	
	iterator target_iter = begin() + DNA2UnitSizeInBase;
	
	// Load the comp_buffer with data from the target seq
	unsigned char comp_buffer = seq[0];
	
	vector<unsigned char> query_buffer(num_query);
	
	for(i = 0;i < num_query;i++){
		query_buffer[i] = m_seq[i].seq[0];
	}
	
	for(i = 0;i < (unsigned int)comp_len;i++){
		// By packing the sequence data into an unsigned char, we can perform
		// fast comparisons.
		for(j = 0;j < num_query;j++){
			if(comp_buffer == query_buffer[j]){

				// Explore possible match in detail
				iterator query_iter = m_seq[j].begin() + DNA2UnitSizeInBase;
				iterator tmp_iter = target_iter;

				while(true){
					if(!query_iter.is_valid()){
						// We found a match!
						match_count++;
						m_match.push_back(j);
						break;
					}

					if(!tmp_iter.is_valid() || 
					  (tmp_iter.base() != query_iter.base())){
						break;
					}

					tmp_iter++;
					query_iter++;
				}
			}
		}
		
		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = (comp_buffer << 2) + target_iter.base();
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned char, we can perform
	// fast comparisons.
	for(j = 0;j < num_query;j++){
		if(comp_buffer == query_buffer[j]){

			// Explore possible match in detail
			iterator query_iter = m_seq[j].begin() + DNA2UnitSizeInBase;
			iterator tmp_iter = target_iter;

			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					match_count++;
					m_match.push_back(j);
					break;
				}

				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}

				tmp_iter++;
				query_iter++;
			}

			// If we get here without a match, then there is no
			// match!
		}
	}

	return match_count;
}

unsigned int DNA2Seq::find_any(const list<DNA2Seq> &m_seq, std::list<const DNA2Seq*> &m_match)
{
	unsigned int i, j;
	unsigned int num_query = m_seq.size();
	unsigned int match_count = 0;
	
	if(num_query == 0){
		return match_count;
	}
	
	list<DNA2Seq>::const_iterator list_iter = m_seq.begin();
	
	unsigned int min_query_size = list_iter->size();
	
	for(;list_iter != m_seq.end();list_iter++){
		min_query_size = min(min_query_size, list_iter->size());
	}
	
	int comp_len = size() - min_query_size;
	
	if((comp_len < 0) || !size() || !min_query_size){
		return match_count;
	}
	
	if(min_query_size < DNA2UnitSizeInBase){
	
		for(list_iter = m_seq.begin();list_iter != m_seq.end();list_iter++){
			iterator tmp_iter = __find(*list_iter);

			if(tmp_iter != end()){
				match_count++;
				m_match.push_back(&(*list_iter));
			}
		}
		return match_count;
	}
	
	iterator target_iter = begin() + DNA2UnitSizeInBase;
	
	unsigned char comp_buffer;
	vector<unsigned char> query_buffer(num_query);
	
	// Load the comp_buffer with data from the target seq
	comp_buffer = seq[0];
	
	for(i = 0, list_iter = m_seq.begin();i < num_query;i++, list_iter++){		
		query_buffer[i] = list_iter->seq[0];
	}
	
	for(i = 0;i < (unsigned int)comp_len;i++){
		// By packing the sequence data into an unsigned char, we can perform
		// fast comparisons.
		for(j = 0, list_iter = m_seq.begin();j < num_query;j++, list_iter++){
			if(comp_buffer == query_buffer[j]){
								
				// Explore possible match in detail
				iterator query_iter = list_iter->begin() + DNA2UnitSizeInBase;
				iterator tmp_iter = target_iter;

				while(true){
					if(!query_iter.is_valid()){
						// We found a match!
						match_count++;
						m_match.push_back(&(*list_iter));
						break;
					}

					if(!tmp_iter.is_valid() || 
					  (tmp_iter.base() != query_iter.base())){
						break;
					}

					tmp_iter++;
					query_iter++;
				}
			}
		}

		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = (comp_buffer << 2) + target_iter.base();
		
		target_iter++;
	}
	
	// By packing the sequence data into an unsigned char, we can perform
	// fast comparisons.
	for(j = 0, list_iter = m_seq.begin();j < num_query;j++, list_iter++){
		if(comp_buffer == query_buffer[j]){
			
			// Explore possible match in detail
			iterator query_iter = list_iter->begin() + DNA2UnitSizeInBase;
			iterator tmp_iter = target_iter;

			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					match_count++;
					m_match.push_back(&(*list_iter));
					break;
				}

				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}

				tmp_iter++;
				query_iter++;
			}

			// If we get here without a match, then there is no
			// match!
		}
	}

	return match_count;
}

DNA2Seq::complement_iterator DNA2Seq::find_complement(const DNA2Seq &m_seq) const
{
	int comp_len = size() - m_seq.size();
	
	if((comp_len < 0) || !size() || !m_seq.size()){
		return complement_end();
	}
	
	if(m_seq.size() < DNA2UnitSizeInBase){
		return __find_complement(m_seq);
	}
	
	complement_iterator target_iter = complement_begin();
	
	// This implementation assumes that sizeof(int) >= 3 which should be
	// true for evey modern OS.
	unsigned char comp_buffer, query_buffer;
	
	// Load the comp_buffer with data from the target seq
	// This loop (which adds DNA2UnitSizeInBase based to comp_buffer)
	// has been unrolled.
	comp_buffer = target_iter.base();
	target_iter++;
	comp_buffer = (comp_buffer << 2) + target_iter.base();
	target_iter++;
	comp_buffer = (comp_buffer << 2) + target_iter.base();
	target_iter++;
	comp_buffer = (comp_buffer << 2) + target_iter.base();
	target_iter++;
	
	query_buffer = m_seq.seq[0];
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned int, we can perform
		// fast comparisons.
		if(comp_buffer == query_buffer){
			
			// Explore possible match in detail
			iterator query_iter = m_seq.begin() + DNA2UnitSizeInBase;
			complement_iterator tmp_iter = target_iter;
			
			while(true){
				if(!query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA2UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != query_iter.base())){
					break;
				}
				
				tmp_iter++;
				query_iter++;
			}
		}
						
		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = (comp_buffer << 2) + target_iter.base();
		
		target_iter++;
	}
	
	
	// By packing the sequence data into an unsigned int, we can perform
	// fast comparisons.
	if(comp_buffer == query_buffer){
			
		// Explore possible match in detail
		iterator query_iter = m_seq.begin() + DNA2UnitSizeInBase;
		complement_iterator tmp_iter = target_iter;
			
		while(true){
			if(!query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA2UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != query_iter.base())){
				break;
			}
				
			tmp_iter++;
			query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	return complement_end();
}

DNA2Seq::complement_iterator DNA2Seq::find_complement(const DNA3Seq &m_seq) const
{
	int comp_len = size() - m_seq.size();
	
	if((comp_len < 0) || !size() || !m_seq.size()){
		return complement_end();
	}
	
	if(m_seq.size() < DNA2UnitSizeInBase){
		return __find_complement(m_seq);
	}
	
	complement_iterator target_iter = complement_begin();
	
	DNA3Seq::iterator query_iter = m_seq.begin();
	
	// This implementation assumes that sizeof(int) >= 3 which should be
	// true for evey modern OS.
	unsigned char comp_buffer, query_buffer;
	
	// Load the comp_buffer with data from the target seq
	// This loop (which adds DNA2UnitSizeInBase based to comp_buffer)
	// has been unrolled.
	comp_buffer = target_iter.base();
	target_iter++;
	
	query_buffer = query_iter.base();
	query_iter++;
	
	for(int j = 1;j < DNA2UnitSizeInBase;j++){
		comp_buffer = (comp_buffer << 2) + target_iter.base();
		target_iter++;
		
		query_buffer = (query_buffer << 2) + query_iter.base();
		query_iter++;
	}
	
	for(int i = 0;i < comp_len;i++){
		// By packing the sequence data into an unsigned int, we can perform
		// fast comparisons.
		if(comp_buffer == query_buffer){
			
			// Explore possible match in detail
			DNA3Seq::iterator tmp_query_iter = query_iter;
			complement_iterator tmp_iter = target_iter;
			
			while(true){
				if(!tmp_query_iter.is_valid()){
					// We found a match!
					return target_iter - DNA2UnitSizeInBase;
				}
				
				if(!tmp_iter.is_valid() || 
				   (tmp_iter.base() != tmp_query_iter.base())){
					break;
				}
				
				tmp_iter++;
				tmp_query_iter++;
			}
		}
						
		// If we get here, not match has been found yet. Note that the 
		// unsigned int representation of the comparison buffer allow for
		// efficient updates.
		comp_buffer = (comp_buffer << 2) + target_iter.base();
		
		target_iter++;
	}
	
	
	// By packing the sequence data into an unsigned int, we can perform
	// fast comparisons.
	if(comp_buffer == query_buffer){
			
		// Explore possible match in detail
		DNA3Seq::iterator tmp_query_iter = query_iter;
		complement_iterator tmp_iter = target_iter;
			
		while(true){
			if(!tmp_query_iter.is_valid()){
				// We found a match!
				return target_iter - DNA2UnitSizeInBase;
			}
				
			if(!tmp_iter.is_valid() || 
			   (tmp_iter.base() != tmp_query_iter.base())){
				break;
			}
				
			tmp_iter++;
			tmp_query_iter++;
		}
			
		// If we get here without a match, then there is no
		// match!
	}
	
	return complement_end();
}

#if defined(WIN32) && !defined(_WINDOWS)
template <class A>
inline A min(const A &m_a, const A &m_b)
{
	return (m_a < m_b) ? m_a : m_b;
}
#endif // WIN32
