// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#ifndef __GENOME_VIEW
#define __GENOME_VIEW

// Genomerama
// J. D. Gans
// Bioscience Division, Los Alamos National Lab,
// Los Alamos, NM
//
// This header file describes a portable, OpenGL, ASN.1 based
// viewer for DNA molecules. 
//
// From the NCBI sdk.doc file:
// "Locations of regions on Bioseqs are always given as 
//  integer offsets, also described in the Biological 	
//  Sequences chapter. So the first residue is always 0 and 
//  the last residue is always (length - 1)." 

// Include the OpenGL headers
#ifdef _MAC_OSX
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/gl.h>
#include <GL/glu.h>
#endif

#include "annotation.h"
#include "text_font_gl.h"
#include "genome_color.h"
#include <list>

/////////////////////////////////////////////////////////////////////////////////////
// OS specific defines and includes
#ifdef WIN32

// Disable the annoying STL warnings for visual c++
#pragma warning(disable:4786)

#define	WM_FORCE_UPDATE		(WM_USER + 1)

#endif // WIN32
/////////////////////////////////////////////////////////////////////////////////////

// Version control
#define	GV_VERSION_MAJOR		"1"
#define	GV_VERSION_MINOR		"5"

#define	AVAILIABILITY		"http://snp.lanl.gov/genomorama"
#define	CONTACT_ADDRESS		"jgans@lanl.gov"

#define	ZOOM_LEVEL_TOP		2	// Display top level annotation
#define	ZOOM_LEVEL_MED		1	// Display mid-level annotation
#define	ZOOM_LEVEL_SEQ		0	// Display the genome sequence

// How many bases across (the base window/2) does each zoom level 
// correspond to?
#define	BASE_TOP_THRESHOLD		15000	// Show all annotations from 70 - 1000
#define	BASE_MID_THRESHOLD		70		// Show Sequence from 0 - 70
#define	MIN_BASE_WINDOW		10		// Minimum number of bases to show

#define	DEFAULT_BASE_WINDOW		30000	// Starting base window

// How many points should we use to plot G+C content along a sequence?
#define	GC_PLOT_POINTS		150

// How many points should we use to plot Tm along a sequence?
#define	TM_PLOT_POINTS		150

// How many points should we use to plot multiple sequence scores?
#define	MULTI_PLOT_POINTS	150

// What range of Tm should we plot?
#define TM_MIN		0.0f
#define TM_MAX		100.0f
#define TM_DELTA	(TM_MAX - TM_MIN)

// OpenGL settings
#define	GV_PERSPECTIVE	60.0f
#define	GV_NEAR_CLIP_Z	0.5f
#define	GV_FAR_CLIP_Z	7.0f

#define	GP_MOUSE_CLEAR				0
#define	GP_MOUSE_L_BUTTON_DOWN		1
#define	GP_MOUSE_M_BUTTON_DOWN		2
#define	GP_MOUSE_R_BUTTON_DOWN		4

// Define INVERT_MOUSE_Y to flip the mouse y-axis
// (so that [0,0] is the lower left corner as opposed to the
// upper left corner
#define INVERT_MOUSE_Y

#define		GP_POLYGON_FONT_SCALE		0.03f
#define		GP_POLYGON_FONT_RESOLUTION	0.3f

// Hybridization and PCR parameters
#define		HYBRID_MIN_PROBE_LEN		10
#define		HYBRID_MAX_PROBE_LEN		100
#define		HYBRID_MIN_PRIMER_LEN		10
#define		HYBRID_MAX_PRIMER_LEN		100
#define		HYBRID_MAX_EXACT_MATCH		HYBRID_MAX_PRIMER_LEN

// This is now a user adjustable parameter
//#define		HYBRID_MAX_AMPLICON_LEN		2000

#define		PCR_PRIMER_MIN_LEN			10
#define		PCR_PRIMER_MAX_LEN			50
#define		PCR_PRIMER_MIN_SALT			1.0e-3f
#define		PCR_PRIMER_MAX_SALT			10.0f
#define		PCR_PRIMER_MIN_STRAND		1.0e-12f
#define		PCR_PRIMER_MAX_STRAND		1.0e-3f

// Only look at matches with better than 50% identity (for now)
#define		TM_IDENTITY_CUTOFF			0.5f

// Store base composition counts
struct BaseEntropy
{
	BaseEntropy()
	{
		A = T = G = C = N = 0.0f;
	};

	float A;
	float T;
	float G;
	float C;
	float N;
};

// A class for storing a genome
class Genome{
private:
	unsigned int base_center;
	unsigned int base_window;
	unsigned int num_base;
	unsigned char zoom_level;
	unsigned int max_gene_depth;
	unsigned int max_annot_depth;
	
	float text_font_scale_factor;

	// How many bases should be averaged over to 
	// compute G+C ,A+T (and other) content?
	unsigned int content_window;

	// How many bases should be used to 
	// compute Tm?
	unsigned int tm_window;
	
	float salt;
	float strand;

	// A floating point buffer to allow smooth side scrolling
	float shift_buffer;

	// An floating point index for reordering
	// genomes via a sort. This is a temporary
	// variable and must be set immediately before
	// every use.
	float sort_rank;

	bool is_selected;
	bool is_shown;
	bool is_annotation_shown;

	bool display_gc_graph;
	bool display_at_graph;

	bool display_tm_graph;
	
	bool display_gc_skew_graph;
	bool display_at_skew_graph;

	bool display_keto_skew_graph;
	bool display_purine_skew_graph;

	bool display_genome_plot;
	bool is_blending;

	std::list<GeneAnnotation>::iterator auto_pick_iter;
	
	DNAMol mol;
	
	// Use supplied genome plot
	std::vector< std::pair<unsigned int, float> > genome_plot;

	// The maximum and minimum Y value of genome_plot
	std::pair<float, float> genome_plot_bounds_y;

	// Save users search results
	std::list< std::list<GeneAnnotation>::iterator > search_results;
	std::list< std::list<GeneAnnotation>::iterator >::iterator cur_search_iter;

	// The matching sequence regions
	std::list<SeqHit> hit_list;
	std::list<SeqHit>::iterator cur_search_hit_iter;

	GLfloat select_color[4];
	GLfloat background_color[4];
	GLfloat text_color[4];
	GLfloat gene_color[4];
	GLfloat pseudo_gene_color[4];
	GLfloat cds_color[4];
	GLfloat annot_color[4];
	GLfloat rna_color[4];
	GLfloat primer_color[4];
	GLfloat user_color[4];
	GLfloat intergenic_color[4];
	GLfloat select_neg_color[4];
	GLfloat select_plus_color[4];

	// Toggle global blending (alpha channel) of all rendered objects
	inline void enable_blend(const GLfloat &m_alpha);
	inline void disable_blend();

	inline void enable_blend_text(const GLfloat &m_alpha);
	inline void disable_blend_text();

	void render_top(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y,
		const FontMetricGL &m_font_metric,
		const bool &m_select, const bool &m_clip,
		const unsigned int &m_clip_edge);
	void render_mid(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y,
		const FontMetricGL &m_font_metric,
		const bool &m_select, const bool &m_clip,
		const unsigned int &m_clip_edge);
	void render_seq(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y,
		const bool &m_select);

	void render_GC_graph(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y);

	void render_AT_graph(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y);

	void render_GC_skew_graph(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y);

	void render_AT_skew_graph(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y);

	void render_keto_skew_graph(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y);

	void render_purine_skew_graph(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y);

	void render_Tm_graph(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y);

	void render_genome_plot(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y);

	void render_annot_info(const GLfloat &m_lower_y, const GLfloat &m_upper_y,
		const GeneAnnotation &m_annot);
	void render_info(const GLfloat &m_lower_y, const GLfloat &m_upper_y);
	void render_select(const GLfloat &m_lower_y, const GLfloat &m_upper_y);
	void render_auto_pick(const std::list<GeneAnnotation>::iterator &m_iter, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y,
		const FontMetricGL &m_font_metric);
	
	// Render arrows that represent annotations. The direction of the
	// arrow represents 5'->3' (right arrow) or 3'->5' (left arrow).
	void render_left_arrow(const GLfloat &x1, const GLfloat &y1,
					   const GLfloat &x2, const GLfloat &y2,
					   const std::list<GeneAnnotation>::iterator &m_iter,
					   const FontMetricGL &m_font_metric,
					   const GLfloat &m_text_alpha);
	void render_right_arrow(const GLfloat &x1, const GLfloat &y1,
					   const GLfloat &x2, const GLfloat &y2,
					   const std::list<GeneAnnotation>::iterator &m_iter,
					   const FontMetricGL &m_font_metric,
					   const GLfloat &m_text_alpha);

	// Render arrows that represent segmented annotations. The direction of the
	// arrow represents 5'->3' (right arrow) or 3'->5' (left arrow).
	void render_left_segmented_arrow(const GLfloat &x1, const GLfloat &y1,
					   const GLfloat &x2, const GLfloat &y2,
					   const std::list<GeneAnnotation>::iterator &m_iter,
					   const FontMetricGL &m_font_metric,
					   const GLfloat &m_text_alpha);
	void render_right_segmented_arrow(const GLfloat &x1, const GLfloat &y1,
					   const GLfloat &x2, const GLfloat &y2,
					   const std::list<GeneAnnotation>::iterator &m_iter,
					   const FontMetricGL &m_font_metric,
					   const GLfloat &m_text_alpha);

	// Intergenic space is represented by a box
	void render_box(const GLfloat &x1, const GLfloat &y1,
					const GLfloat &x2, const GLfloat &y2);

	void set_color(const std::list<GeneAnnotation>::iterator &m_iter);
	void set_shadow_color(const std::list<GeneAnnotation>::iterator &m_iter);
	void set_inv_color(const std::list<GeneAnnotation>::iterator &m_iter);
	
	bool process_hits(unsigned int &m_annot_id, const GLint &hits, 
					  const int &bufSize, const GLuint *buffer);

	std::string base_composition_info(GeneAnnotation &m_annot);
	std::string genome_plot_info(GeneAnnotation &m_annot);
public:

	enum {SELECT_COLOR = 0, BACKGROUND_COLOR, TEXT_COLOR, GENE_COLOR, PSEUDO_GENE_COLOR, CDS_COLOR, 
		ANNOT_COLOR, PRIMER_COLOR, RNA_COLOR, INTERGENIC_COLOR, SELECT_NEG_COLOR, SELECT_PLUS_COLOR,
		USER_COLOR, CUSTOM_COLOR};

	Genome(const unsigned int &m_mode);

	// Render this genome in the supplied view port
	void render(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y,
		const unsigned int &m_clip_edge, 
		const FontMetricGL &m_font_metric,
		const bool &m_is_printing);
	
	// Render a translation guide to help the user select sequence
	void render_translation_guide(const unsigned int &m_x1, const GLfloat &m_y1,
								  const unsigned int &m_x2, const GLfloat &m_y2);

	// Pick a gene, annotation or base from this genome
	bool pick(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y,
		const unsigned int m_base, const GLfloat &m_y,
		const FontMetricGL &m_font_metric);

	bool auto_pick(const unsigned int &m_width, 
		const GLfloat &m_lower_y, const GLfloat &m_upper_y,
		const unsigned int m_base, const GLfloat &m_y,
		const FontMetricGL &m_font_metric);
	
	// If an annotation is currently "auto-picked", clear it
	inline void clear_auto_pick()
	{
		auto_pick_iter = mol.end();
	};
		
	void complement();

	inline bool show() const
	{
		return is_shown;
	};

	inline bool empty() const
	{
		return mol.empty();
	};

	inline void show(const bool &m_show)
	{
		is_shown = m_show;
	};

	inline void show_annotation(const bool &m_show)
	{
		is_annotation_shown = m_show;
	};

	inline bool show_annotation() const
	{
		return is_annotation_shown;
	};

	void mode(const unsigned int &m_mode);

	inline void toggle_gc_display()
	{
		display_gc_graph = !display_gc_graph;
	};
	
	inline void toggle_at_display()
	{
		display_at_graph = !display_at_graph;
	};

	inline void toggle_gc_skew_display()
	{
		display_gc_skew_graph = !display_gc_skew_graph;
	};

	inline void toggle_at_skew_display()
	{
		display_at_skew_graph = !display_at_skew_graph;
	};

	inline void toggle_purine_skew_display()
	{
		display_purine_skew_graph = !display_purine_skew_graph;
	};

	inline void toggle_keto_skew_display()
	{
		display_keto_skew_graph = !display_keto_skew_graph;
	};

	inline void content_window_size(const int &m_size)
	{
		content_window = m_size;
	};

	inline void toggle_tm_display()
	{
		display_tm_graph = !display_tm_graph;
	};

	inline void tm_window_size(const int &m_size)
	{
		tm_window = m_size;
	};

	inline void tm_salt(const float &m_salt)
	{
		salt = m_salt;
	};

	inline void tm_strand(const float &m_strand)
	{
		strand = m_strand;
	};

	inline void font_scale_factor(const float &m_scale)
	{
		text_font_scale_factor = m_scale;
	};

	inline void toggle_genome_plot_display()
	{
		display_genome_plot = genome_plot.empty() ? false : !display_genome_plot;
	};

	void set_genome_plot(const std::list< std::pair<unsigned int, float> > &m_plot);
	
	bool color(const unsigned char &m_r, const unsigned char &m_g, const unsigned char &m_b,
		const unsigned int &m_feature);

	bool color(const GLfloat m_color[], const unsigned int &m_feature);

	inline void base_range(unsigned int &m_base_start, unsigned int &m_base_stop) const
	{
		m_base_start = (base_center >= base_window) ? 
			(base_center - base_window) % num_base :
			num_base - (base_window - base_center);
		m_base_stop = (base_center + base_window) % num_base;

		if(m_base_start == m_base_stop){
			m_base_start = wrap(m_base_stop + 1, num_base);
		}
	};

	inline unsigned int seq_size() const
	{
		return num_base;
	};
	
	inline unsigned int cur_base_loc(const GLfloat &m_x) const
	{
		unsigned int tmp_start = (base_center >= base_window) ? 
			(base_center - base_window) % num_base :
			num_base - (base_window - base_center);

		return (unsigned int)(tmp_start + m_x) % num_base;
	};

	inline bool cur_show_tm()
	{
		return display_tm_graph;
	};

	inline bool cur_show_gc()
	{
		return display_gc_graph;
	};

	inline bool cur_show_at()
	{
		return display_at_graph;
	};

	inline bool cur_show_gc_skew()
	{
		return display_gc_skew_graph;
	};

	inline bool cur_show_at_skew()
	{
		return display_at_skew_graph;
	};

	inline bool cur_show_keto_skew()
	{
		return display_keto_skew_graph;
	};

	inline bool cur_show_purine_skew()
	{
		return display_purine_skew_graph;
	};

	inline bool cur_show_genome_plot()
	{
		return display_genome_plot;
	};

	float cur_tm(const GLfloat &m_x);
	float cur_gc(const GLfloat &m_x);
	float cur_at(const GLfloat &m_x);
	float cur_genome_plot(const GLfloat &m_x);
	float cur_gc_skew(const GLfloat &m_x);
	float cur_at_skew(const GLfloat &m_x);
	float cur_keto_skew(const GLfloat &m_x);
	float cur_purine_skew(const GLfloat &m_x);

	inline void jump_to(const long &m_goto)
	{
		if(m_goto < 0){
			return;
		}

		// Move the display to center of the current annotation
		base_center = (m_goto > (long)num_base) ? num_base : m_goto;
	};

	inline void shift(const GLfloat &dx, const bool &m_add_to_shift_buffer = false)
	{
		int new_base_center = base_center;

		// The shift_buffer is used to accumulate translations that are less than
		// one base left or right. These fractional bases build up (for example, after
		// successive move movements) into whole base shifts.
		if(m_add_to_shift_buffer){
			shift_buffer += dx*base_window;
			
			double ip;

			modf(shift_buffer, &ip);

			if(fabs(ip) > 0){
				new_base_center -= (int)ip;
				shift_buffer -= ip;
			}
			else{
				return; // No shifting to perform (only a fraction of a base in 
						// the shift buffer.
			}
		}
		else{
			new_base_center -= (int)(dx*base_window);
		}

		base_center = (new_base_center < 0) ? 
			num_base - (-new_base_center)%num_base :
			new_base_center%num_base;	
	};

	inline void zoom(const unsigned char &m_level, const unsigned int &m_base_window)
	{
		zoom_level = m_level;

		#ifdef _WINDOWS
		base_window = min(m_base_window, num_base/2);
		#else
		base_window = std::min(m_base_window, num_base/2);
		#endif // _WINDOWS
	};
	
	inline bool select() const
	{
		return is_selected;
	};

	inline void select(const bool &m_select)
	{
		is_selected = m_select;

		if(!is_selected){
			mol.clear_select();
		}
	};

	// Return the number of selected annotations
	inline unsigned int num_select() const
	{
		return mol.num_select();
	};
	
	inline std::list<GeneAnnotation>::const_iterator selected_annotation() const
	{
		return mol.selected_annotation();
	};
	
	void update_annotation(const GeneAnnotation &m_annot)
	{
		mol.update_annotation( m_annot, mol.end() );
		
		// Reprocess the annotation list
		mol.processGeneList();
		
		// Every time we call processGeneList(), the list of annotations is sorted and
		// we need to reset iterators that refer to elements in the annotation list.
		clear_auto_pick();
		
		// Compute the maximum depths
		std::list<GeneAnnotation>::iterator iter;

		max_gene_depth = max_annot_depth = 0;

		// Recompute the maximum depth
		for(iter = mol.begin();iter != mol.end();iter++){
		
			if( max_gene_depth < iter->depth() ){
				max_gene_depth = iter->depth();
			}

			if( max_annot_depth < iter->annot_depth() ){
				max_annot_depth = iter->annot_depth();
			}
		}
	};
	
	void update_annotation(const GeneAnnotation &m_annot, 
		const std::list<GeneAnnotation>::const_iterator &m_iter)
	{
		mol.update_annotation(m_annot, m_iter);
		
		// Reprocess the annotation list
		mol.processGeneList();
		
		// Every time we call processGeneList(), the list of annotations is sorted and
		// we need to reset iterators that refer to elements in the annotation list.
		clear_auto_pick();
		
		// Compute the maximum depths
		std::list<GeneAnnotation>::iterator iter;

		max_gene_depth = max_annot_depth = 0;

		// Recompute the maximum depth
		for(iter = mol.begin();iter != mol.end();iter++){
		
			if( max_gene_depth < iter->depth() ){
				max_gene_depth = iter->depth();
			}

			if( max_annot_depth < iter->annot_depth() ){
				max_annot_depth = iter->annot_depth();
			}
		}
	};
	
	// Return infomation about the first selected annotation
	std::string select_info();

	inline std::string taxa_info() const
	{
		return mol.info(DNAMol::SOURCE) + " " +  mol.info(DNAMol::TAXA_NAME) + " " + 
			mol.info(DNAMol::LINEAGE) + " " + mol.info(DNAMol::GENUS) + " " + 
			mol.info(DNAMol::SPECIES) + " " + mol.info(DNAMol::SUBSPECIES);
	};

	inline std::string taxa_name() const
	{
		return mol.info(DNAMol::TAXA_NAME) + "(" + mol.seq_id_str() + ")";
	};
	
	bool load(const std::string &m_filename, const unsigned int m_type, 
		std::streampos &m_pos);

	void load(const int &m_gi);
	void load_and_save(const int &m_gi, const std::string &m_filename);

	bool import(const Genome &m_genome, const unsigned int &m_type);
		
	inline void rank(const float &m_rank)
	{
		sort_rank = m_rank;
	};

	// Sort in order of smallest to largest rank
	inline bool operator<(const Genome &m_copy) const
	{
		return (sort_rank < m_copy.sort_rank);
	};

	inline bool operator>(const Genome &m_copy) const
	{
		return (sort_rank > m_copy.sort_rank);
	};

	inline bool operator==(const SeqIdPtr &m_sip)
	{
		return (mol == m_sip);
	}

	std::list<GeneAnnotation>::iterator find_annot(const SeqIdPtr &m_sip);

	// Search functions:
	std::string search_name(const std::string &m_name);
	std::string search_gi(const unsigned int &m_gi);
	std::string search_base(const unsigned int &m_base);
	std::string search_text(const std::list<std::string> &m_text);
	std::string search_seq(const std::list<DNA3Seq> &m_seq);
	std::string search_primers(const DNA3Seq &m_primer_1, const DNA3Seq &m_primer_2,
		const float &m_salt, const float &m_strand, const float &m_tm,
		const unsigned int &m_exact_match, const unsigned int &m_min_amplicon_length,
		const unsigned int &m_max_amplicon_length, const bool &m_clear_results);
	std::string search_probe(const DNA3Seq &m_probe, const float &m_salt, const float &m_strand,
		const float &m_tm, const bool &m_clear_results);
	std::string search_MOL_primers(const DNA3Seq &m_primer_1, const DNA3Seq &m_primer_2,
		const float &m_salt, const float &m_strand, const float &m_tm,
		const unsigned int &m_exact_match);

	void search_next();
	std::string search_info();
	std::string probe_info();
	std::string primer_info();
	std::string MOL_primer_info();

	inline unsigned int num_search_results()
	{
		return search_results.size();
	};

	// Output any selected sequences in FASTA format
	std::string selected_seq_as_protein();
	std::string selected_seq_as_dna();

	// Translate the nucleotide sequence [m_start, m_stop]
	std::string translate_sequence(const long &m_start, const long &m_stop);

	// Show the raw sequence
	std::string sequence(const long &m_start, const long &m_stop);

	// Compute all primers that match the specified options
	void compute_primers(const unsigned int &m_options, const unsigned int &m_len,
						 const unsigned int &m_run,
						 const float &m_salt, const float &m_strand,
						 const float &m_min_gc, const float &m_max_gc, 
						 const float &m_tm, const float &m_mis_prime_tm);

	// Clear all primers
	void clear_primers();

	// Turn all currently selected text into annotation
	bool annotate_selection();

	// Delete all currently selected annotations
	bool delete_annotation();

	void conservation_score(std::vector<BaseEntropy> &m_s, const unsigned int &m_dx);

	std::list<GeneAnnotation>::iterator end()
	{
		return mol.end();
	};

	std::list<GeneAnnotation>::iterator begin()
	{
		return mol.begin();
	};
	
	void export_fasta_genome(std::ofstream &m_fout);
	void export_fasta_dna(std::ofstream &m_fout, 
		const std::list<unsigned int> &m_annot);
	void export_fasta_protein(std::ofstream &m_fout, 
		const std::list<unsigned int> &m_annot);
	void export_gbk(std::ofstream &m_fout, 
		const std::list<unsigned int> &m_annot);
	void export_ptt(std::ofstream &m_fout, 
		const std::list<unsigned int> &m_annot);
		
	inline bool has_sequence_data() const
	{
		return mol.has_sequence_data();
	};
};

// A class for storing selected annotations (with associated genome and list info as well)
class selected_annotation_info : public GeneAnnotation {
	private:
		std::list<Genome>::const_iterator genome_iter;
		std::list<GeneAnnotation>::const_iterator annot_iter;
		bool genome_valid;
		bool annot_valid;
		
	public:
		selected_annotation_info(const std::list<Genome>::const_iterator &m_genome_iter,
			const std::list<GeneAnnotation>::const_iterator &m_annot_iter) :
			genome_iter(m_genome_iter), annot_iter(m_annot_iter)
		{
			genome_valid = true;
			annot_valid = true;
		};
		
		selected_annotation_info(const std::list<Genome>::const_iterator &m_genome_iter) :
			genome_iter(m_genome_iter)
		{
			genome_valid = true;
			annot_valid = false;
		};
		
		selected_annotation_info()
		{
			genome_valid = false;
			annot_valid = false;
		};
				
		selected_annotation_info& operator=(const GeneAnnotation &m_annot)
		{
			*((GeneAnnotation*)this) = m_annot;
			
			return *this;
		};
		
		inline bool is_genome_valid() const
		{
			return genome_valid;
		};
		
		inline bool is_annot_valid() const
		{
			return annot_valid;
		};
		
		std::list<Genome>::const_iterator genome() const
		{
			return genome_iter;
		};
		
		std::list<GeneAnnotation>::const_iterator annotation() const
		{
			return annot_iter;
		};
};

// A class for storing and rendering a set of genomes
class GenomePage{
private:
	std::list<Genome> genome_list;
	unsigned int text_font_size;
	float text_font_scale_factor;

	unsigned int content_window_size;
	unsigned int tm_window_size;

	float salt;
	float strand;

	unsigned char zoom_level;

	// Are we in gene centric or size centric mode?
	unsigned int scale_mode;

	// The maximum number of genomes on the
	// screen (with out scrolling)
	unsigned int max_genome_display;

	// NCBI ASN.1 management
	ErrHookProc old_error_hook; // Save this variable in case we ever
						   // need to restore the original error
						   // handler.

	// A jump_to_buffer buffer for jumping to a particular base number
	// A value < 0 means that buffer has not been initialized
	long jump_to_buffer;

	std::list<long> translation_buffer_base;
	std::list< std::list<Genome>::iterator > translation_buffer_genome;

	// OpenGL settings
	GLfloat mat_specular[4];
	GLfloat mat_diffuse[4];
	GLfloat mat_shininess[1];
	GLfloat background[4];
	GLfloat light_color[4];
	GLfloat light_position[3];
	GLfloat scale_factor;
	GLfloat perspective;

	// View dimensions
	unsigned int base_window;	// X C [0 ... 2*base_window]
	GLfloat y_bottom;			// 0 <= y_bottom <= y_top
	GLfloat y_top;				// y_bottom <= y_top <= Num_genomes

	// Font metric information
	FontMetricGL font_metric;

	// Screen dimensions
	unsigned int width, height;

	// Keep track of the mouse state and location
	unsigned int mouse_status;

	// Locations of button down events (since coordinate
	// systems are relative to a given window, these values
	// can be negative!).
	int mouse_l_start[2];
	int mouse_m_start[2];
	int mouse_r_start[2];

	int mouse_loc[2];
	
	int scroll_loc;
	int zoom_loc;
	
	enum {X = 0, Y};

	// Genome animation -- ala a roulette wheel
	bool momentum;
	
	// Are we actually spinning?
	bool spinning;

	// Are we automatically displaying information for the gene currently under
	// the pointer?
	bool pointer_auto_pick;
	
	// How fast are we spinning?
	float spin_dx;

	long elapsed_time;
	long last_elapsed_time;

	void spin(const float &m_dx);

	// Allow delayed file opening for command
	// line processing
	std::list<std::string> files_to_open;

	// An internal buffer of annotations to export
	std::list<unsigned int> annotations_to_export;
	unsigned int genome_export_format;
	
	// Are we displaying a multiple sequence conservation plot?
	bool conservation_plot;

	// Render a multiple sequence conservation plot
	void render_multiple_sequence_conservation();
	void render_splash_screen(const FontMetricGL &m_font_metric);
public:

	// Export formats for genome data
	enum {EXPORT_FASTA_GENOME, EXPORT_FASTA_DNA, EXPORT_FASTA_PROTEIN,
		EXPORT_GBK, EXPORT_PTT, EXPORT_INVALID};
	
	GenomePage();
	
	// Command line argument processing
	inline void add_files_to_open(const std::string &m_filename)
	{
		files_to_open.push_back(m_filename);
	};

	inline void add_files_to_open(const std::list<std::string> &m_filename_list)
	{
		std::list<std::string>::const_iterator iter;

		for(iter = m_filename_list.begin();iter != m_filename_list.end();iter++){
			files_to_open.push_back(*iter);
		}

	};

	inline void clear_files_to_open()
	{
		files_to_open.clear();
	};
	
	void process_files_to_open();

	inline void add_annotation_to_export(const unsigned int &m_annot)
	{
		annotations_to_export.push_back(m_annot);
	};
	
	inline void clear_annotations_to_export()
	{
		annotations_to_export.clear();
	};
	
	inline unsigned int num_annotations_to_export()
	{
		return annotations_to_export.size();
	};
	
	inline bool set_export_format(const unsigned int &m_format)
	{
		if(m_format >= EXPORT_INVALID){
			throw __FILE__ ":set_export_format: Invalid format";
		}
		
		genome_export_format = m_format;
		
		// Is the export format valid (given the number of genomes)
		if(genome_export_format == EXPORT_PTT){
		
			// PTT only supports a single genome
			switch( num_select() ){
				case 0:
					if(num_display() > 1){
						genome_export_format = EXPORT_INVALID;
						return false;
					}
					break;
				case 1:
					return true;
				default:
					genome_export_format = EXPORT_INVALID;
					return false;
			};
		}
		
		// This format is valid
		return true;
	};
	
	void export_genomes(const std::string &m_filename);
	
	void export_fasta_genome(const std::string &m_filename);
	void export_fasta_dna(const std::string &m_filename);
	void export_fasta_protein(const std::string &m_filename);
	void export_gbk(const std::string &m_filename);
	void export_ptt(const std::string &m_filename);
	
	// Render to the screen
	bool render(const bool &m_is_printing = false);

	// Render to a PostScript file
	void render_ps(const std::string &m_name);
	
	// Render to a gif file
	void render_gif(const std::string &m_name);

	// Render a guide that highlights the current mouse position
	void render_mouse_guide();
	
	enum {PICKED_NONE, PICKED_GENOME, PICKED_ANNOT};

	int pick(const unsigned int &m_x, const unsigned int &m_y,
		std::list<Genome>::iterator &m_pick_iter);

	bool auto_pick(const unsigned int &m_x, const unsigned int &m_y);
		
	inline unsigned int font_size() const
	{
		return text_font_size;
	};

	inline FontMetricGL& get_font_metric()
	{
		return font_metric;
	};

	inline unsigned int max_mol()
	{
		return max_genome_display;
	};

	inline void max_mol(const unsigned int &m_max)
	{
		max_genome_display = m_max;

		// Count the number of displayed genomes
		std::list<Genome>::iterator iter;
		unsigned int render_count = 0;

		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if(iter->show()){
				render_count++;
			}
		}

		y_bottom = 0.0f;
		y_top = (render_count < max_genome_display) ? render_count : max_genome_display;

		// Protect the gluOrtho2D transformation from
		// a collapsed y dimension
		y_top = (y_top > 1.0f) ? y_top : 1.0f; // Where possible, avoid min and max (not portable)

		glLoadIdentity();
		gluOrtho2D(0.0, 2*base_window, y_bottom, y_top);
	};

	inline unsigned int window() const
	{
		return base_window;
	};

	inline void screen_size(const unsigned int &m_w, const unsigned int &m_h)
	{
		width = m_w;
		height = m_h;
	};

	inline void view_size()
	{
		// Recompute the view boundaries
		y_bottom = 0.0f;

		#ifdef _WINDOWS

			const unsigned int render_count = min(num_display(), max_genome_display);
			y_top = max(1.0f, (float)render_count);

		#else

			const unsigned int render_count = std::min(num_display(), max_genome_display);
			y_top = std::max(1.0f, (float)render_count);

		#endif // _WINDOWS

		glLoadIdentity();
		gluOrtho2D(0.0, 2*base_window, y_bottom, y_top);
	};

	inline void set_char_dx(const char &m_char, const GLfloat &m_dx)
	{
		font_metric[m_char] = m_dx;
	};

	void background_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);	
	void gene_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void pseudo_gene_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void cds_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void rna_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void annot_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void primer_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void user_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void text_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void intergenic_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void seq_plus_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void seq_minus_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);
	void select_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);

	void custom_color(const unsigned char &m_r, const unsigned char &m_g,
		const unsigned char &m_b);

	// The location of the mouse X coordinate in the genome that
	// currently has the mouse focus
	unsigned int cur_base_loc();
	
	// Will the  value of m_x for the mouse_loc[X] cause the base to change?
	bool base_loc_change(const int &m_x);

	// Is the genome that currently has the mouse focus displaying Tm?
	bool cur_show_tm();
	
	float cur_tm();

	// Is the genome that currently has the mouse focus displaying %GC?
	bool cur_show_gc();
	float cur_gc();

	bool cur_show_at();
	float cur_at();

	bool cur_show_gc_skew();
	float cur_gc_skew();

	bool cur_show_at_skew();
	float cur_at_skew();

	bool cur_show_purine_skew();
	float cur_purine_skew();

	bool cur_show_keto_skew();
	float cur_keto_skew();

	// Is the genome that currently has the mouse focus displaying a genome plot?
	bool cur_show_genome_plot();

	float cur_genome_plot();

	// Are we currently showing a conservation plot?
	bool show_conservation_plot()
	{
		return conservation_plot;
	};

	// The number of selected genomes
	inline unsigned int num_select() const
	{
		std::list<Genome>::const_iterator iter;
		unsigned int select_count = 0;

		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->select() ){
				select_count++;
			}
		}

		return select_count;
	};

	inline unsigned int num_display() const
	{
		std::list<Genome>::const_iterator iter;
		unsigned int render_count = 0;

		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			if( iter->show() ){
				render_count++;
			}
		}

		return render_count;
	};

	// Return selected annotation information
	selected_annotation_info selected_annotation();
	
	void update_annotation(const selected_annotation_info &m_annot);
	
	void zoom(const int &m_base_window);
	void fit_to_window();
	void scroll(const GLfloat &m_dy);

	inline unsigned int mode() const
	{
		return scale_mode;
	};

	void mode(const unsigned int &m_mode);

	// Jump to the given base for the selected genomes
	// If no base is provided (i.e. m_base_num < 0) use the internal buffer
	void jump_to(const long &m_base_num = -1);
	void jump_buffer_push(const unsigned int &m_i);
	inline void jump_buffer_clear()
	{
		jump_to_buffer = -1;
	};


	std::list<unsigned int> gi_list_from_search_key(const std::string &m_search_key, 
		const int &m_type);

	bool load(const std::string &m_filename, const int &m_type = DNAMol::ASN_1);
	bool load(const unsigned int &m_gi);
	
	bool load_and_save(const unsigned int &m_gi, const std::string &m_filename);
	
	// Import annotations to an existing genome
	bool import_annotation(const std::string &m_name, const int &m_type = -1);
	
	// Read a data file of genome plot values
	void read_genome_plot(const std::string &m_filename);

	// OpenGL management
	bool initialize_OGL();
	bool setup_view_OGL(const unsigned int &m_width, const unsigned int &m_height);

	#ifdef _DEBUG
	void get_error_OGL(char *m_file, int m_line);
	#endif // _DEBUG

	// Mouse/pointer interface: 
	// Functions return true if the view needs to be redraw, false otherwise
	bool l_button_down(const int &m_x, const int &m_y);
	bool l_button_ctrl_down(const int &m_x, const int &m_y);
	bool l_button_shift_down(const int &m_x, const int &m_y, std::string &m_seq);
	bool l_button_dbl_down(const int &m_x, const int &m_y, std::string &m_info);
	bool l_button_up(const int &m_x, const int &m_y);
	bool m_button_down(const int &m_x, const int &m_y);
	bool m_button_up(const int &m_x, const int &m_y);
	bool r_button_down(const int &m_x, const int &m_y);
	bool r_button_up(const int &m_x, const int &m_y);
	
	bool mouse_wheel(const int &m_x, const int &m_y, const int &m_delta);

	bool mouse_move(const int &m_x, const int &m_y);
	
	bool scroll_bar(const int &m_loc, const int &m_width);
	inline void scroll_bar(const int &m_loc)
	{
		scroll_loc = m_loc;
	};

	bool zoom_bar(const int &m_loc, const int &m_width);
	inline void zoom_bar(const int &m_loc)
	{
		zoom_loc = m_loc;
	};
	
	// Keyboard interface (the choice of (most) keys are platform dependent):
	// Functions return true if the view needs to be redraw, false otherwise
	
	// Deselect any selected genomes. If m_clear_translation is true, then
	// clear the translation buffer as well
	bool clear_selection(const bool m_clear_translation = true); 
	
	bool select_all();				// Select all displayed genomes
	bool delete_selection();			// Delete any selected genomes
	bool delete_annotation();		// Delete any selected annotations
	bool shift_selection_up();		// Shift selected genomes up one place
	bool shift_selection_down();		// Shift selected genomes down one place
	bool shift_selection_to_bottom();	// Shift selected genomes to the bottom of the view
	bool complement_selection();		// Take the complement of selected genomes
	
	// Toggle the display of graphs for all selected genomes
	bool toggle_gc_selection();			
	bool toggle_at_selection();
	bool toggle_gc_skew_selection();
	bool toggle_at_skew_selection();
	bool toggle_keto_skew_selection();
	bool toggle_purine_skew_selection();

	bool toggle_tm_selection();			// Toggle the display of the Tm graph of
										// any selected genomes
	bool toggle_genome_plot_selection();// Toggle the display of the genome plot of
										// any selected genomes

	bool toggle_conservation_plot();	// Toggle the display of the multiple alignment conservation
										// plot for all currently selected genomes

	bool hide_selection();				// Hide any selected genomes
	bool hide_selection_annotation();	// Hide any selected genomes (annotation only)
	bool show_all();					// Show all genomes
	bool show_annotation();				// Show all selected genome annotation
	bool increase_genome_display();		// Increase the number of genomes displayed by 1
	bool decrease_genome_display();		// Decrease the number of genomes displayed by 1

	bool key_left_arrow();
	bool key_right_arrow();
	bool key_up_arrow();
	bool key_down_arrow();
	bool key_page_down();
	bool key_page_up();
	bool key_plus();
	bool key_minus();

	bool annotate_selection();			// Annotate the currently selected sequence

	// Allow users to scale the text font size
	// by choosing an integer (i.e. slider)
	inline float font_scale() const
	{
		return text_font_scale_factor;
	};

	inline int min_font_scale_index() const
	{
		return 1;
	};

	inline int max_font_scale_index() const
	{
		return 100;	
	};
	
	inline float min_font_scale() const
	{
		return 0.5f;
	};

	inline float max_font_scale() const
	{
		return 5.0f;
	};

	inline int font_scale_index() const
	{
		// Uniform scaling
		const float scale = (text_font_scale_factor - min_font_scale())/
			(max_font_scale() - min_font_scale());

		return (int)(min_font_scale_index() + scale*(max_font_scale_index() - min_font_scale_index()));
	};

	void font_scale_index(const int &m_index);

	inline unsigned int content_window() const
	{
		return content_window_size;
	};

	void content_window(const unsigned int &m_index);

	inline unsigned int min_content_window() const
	{
		return 3u;
	};

	inline unsigned int max_content_window() const
	{
		return 1000u;
	};

	inline unsigned int tm_window() const
	{
		return tm_window_size;
	};

	void tm_window(const unsigned int &m_index);

	inline unsigned int min_tm_window() const
	{
		return 10u;
	};

	inline unsigned int max_tm_window() const
	{
		return 30u;
	};

	inline float tm_salt() const
	{
		return salt;
	};

	inline int min_tm_salt_index() const
	{
		return 1;
	};

	inline int max_tm_salt_index() const
	{
		return 100;
	};

	inline float min_tm_ln_salt() const
	{
		return log(1.0e-3f); // [Na+] >= Milimolar
	};

	inline float max_tm_ln_salt() const
	{
		return log(10.0f); // [Na+] <= 10 M
	};

	inline int tm_salt_index() const
	{
		// Uniform scaling in the log
		return (int)(min_tm_salt_index() + (max_tm_salt_index() - min_tm_salt_index())*
			(log(tm_salt()) - min_tm_ln_salt())/(max_tm_ln_salt() - min_tm_ln_salt()));
	};

	void tm_salt_index(const int &m_index);

	inline float tm_strand() const
	{
		return strand;
	};

	inline int min_tm_strand_index() const
	{
		return 1;
	};

	inline int max_tm_strand_index() const
	{
		return 100;
	};

	inline float min_tm_ln_strand() const
	{
		return log(1.0e-12f); // [Ct]
	};

	inline float max_tm_ln_strand() const
	{
		return log(1.0e-3f); // [Ct]
	};

	inline int tm_strand_index() const
	{
		// Uniform scaling in the log
		return (int)(min_tm_strand_index() + (max_tm_strand_index() - min_tm_strand_index())*
			(log(tm_strand()) - min_tm_ln_strand())/(max_tm_ln_strand() - min_tm_ln_strand()));
	};

	void tm_strand_index(const int &m_index);

	// Output actual sequence
	std::string selected_seq_as_protein();
	std::string selected_seq_as_dna();

	// Search functions:
	enum {	SEARCH_KEY_NAME = 0, 
			SEARCH_KEY_GI, 
		  	SEARCH_KEY_BASE,
		  	SEARCH_KEY_TEXT,
		  	SEARCH_KEY_DNA,
			SEARCH_KEY_GENOME_NAME};

	bool validate_search_key(const unsigned int m_key, const std::string &m_value);
	std::string search(const unsigned int m_key, const std::string &m_value);
	
	std::string search_primers(const std::string &m_primer_1, const std::string &m_primer_2,
		const float &m_salt, const float &m_strand, const float &m_tm,
		const unsigned int &m_exact_match, const int &m_min_amplicon_length,
		const int &m_max_amplicon_length);
	std::string search_probe(const std::string &m_probe, const float &m_salt, const float &m_strand,
		const float &m_tm);
	
	// Search using MOL pcr primers
	std::string search_MOL_primers(const std::string &m_primer_1, const std::string &m_primer_2,
		const float &m_salt, const float &m_strand, const float &m_tm,
		const unsigned int &m_exact_match);

	// Return the next hits from the last search performed
	void search_next();

	// Compute valid primers for the selected genomes
	bool compute_primers(const unsigned int &m_options, const unsigned int &m_len,
						 const unsigned int &m_run,
						 const float &m_salt, const float &m_strand,
						 const float &m_min_gc, const float &m_max_gc, 
						 const float &m_tm, const float &m_mis_prime_tm);

	// Clear any previously computed primers for the selected genomes
	bool clear_primers();

	inline bool spin_momentum() const
	{
		return momentum;
	};

	inline void spin_momentum(const bool &m_use_momentum)
	{
		momentum = m_use_momentum;

		// Start off stationary
		spinning = false;
	};
	
	inline bool magnify() const
	{
		return pointer_auto_pick;
	};

	inline void magnify(const bool &m_use_magnify)
	{
		pointer_auto_pick = m_use_magnify;
		
		if(pointer_auto_pick == false){
		
			std::list<Genome>::iterator iter;

			for(iter = genome_list.begin();iter != genome_list.end();iter++){
				iter->clear_auto_pick();
			}
		}
	};		
};


// Misc functions:
std::string primer_test(const DNA3Seq &m_primer_1, 
		const DNA3Seq &m_primer_2, const float &m_salt, const float &m_strand, 
		const float &m_tm, const bool &m_display_seq = false);

#endif // __GENOME_VIEW
