// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "genome.h"

#include <fstream>
#include <sstream>

extern "C"{
	#include "sqnutils.h"
}

using namespace std;

enum {BLANK_LINE, SEQID_LINE, DATA_LINE, COMMENT_LINE};

// Local functions
int line_type(const string &m_line, list< pair<unsigned int, float> > &m_plot_list,
			  unsigned int &m_base_index);


// Win32 STL implementation requires std::greater for std::list.sort
// I suspect that other implementations use std::less
struct plot_comp : public std::greater< pair<unsigned int, float> >
{
	bool operator()(const pair<unsigned int, float> &m_a, 
		const pair<unsigned int, float> &m_b)
	{
		return (m_a.first > m_b.first); 
	};
};

// Import genome plot data from a file
void GenomePage::read_genome_plot(const std::string &m_filename)
{
	if(genome_list.empty()){
		throw "Please load a genome before loading a genome plot";
	}

	ifstream fin( m_filename.c_str() );

	if(!fin){
		throw "Unable to open genome plot file";
	}

	// The format of a genome plot file record is:
	// > SeqID string
	// int0	float0
	// int1 float1
	// ...
	//
	// There can be multiple plot records per file.

	string line;
	
	unsigned int pos;
	unsigned int line_len;
	unsigned int stop;

	unsigned int base_index = 0;

	// Remember the number of selected genomes
	const unsigned int num_selected_genomes = num_select();

	// The user can specify a plot by genes
	list< pair<unsigned int, float> > gene_plot_list;

	// The user can specify a plot by genome
	list< pair<unsigned int, float> > plot_list;

	SeqIdPtr sip = NULL;

	while(getline(fin, line)){
		switch(line_type(line, plot_list, base_index)){
			case BLANK_LINE:
			case COMMENT_LINE:
				// Skip blank lines and comments
				break;
			case SEQID_LINE:

				// Reset the base index
				base_index = 0;

				if(plot_list.empty() == false){
					// Trim duplicate points
					typedef list< pair<unsigned int, float> >::iterator I;

					list<I> buffer;

					I list_iter;

					for(list_iter = plot_list.begin();list_iter != plot_list.end();list_iter++){
						buffer.push_back(list_iter);

						if(buffer.size() == 3){
							// Do we need to delete an entry?
							list<I>::iterator iter0, iter1, iter2;
							
							iter0 = iter2 = buffer.begin();

							iter2++;
							
							iter1 = iter2;

							iter2++;

							if(((*iter0)->second == (*iter1)->second) && ((*iter1)->second == (*iter2)->second)){
								plot_list.erase(*iter1);
								buffer.erase(iter1);
							}
							else{
								buffer.pop_front();
							}
						}
					}

					// Save the current data set and flush the list
					// Iterate through the Genome list and find the DNA
					// molecule that matches the current SeqId
					list<Genome>::iterator iter;
					
					// Sort the genome plot data
					//plot_list.sort( plot_comp());
					
					if(sip){
						// We may need to load the same genome plot into multiple
						// duplicates of the same genome

						bool found_genome_sip = false;

						for(iter = genome_list.begin();iter != genome_list.end();iter++){
							if(*iter == sip){
								found_genome_sip = true;
								iter->set_genome_plot(plot_list);
							}
						}
						
						// Did we find a genome to add the plot to?
						if(found_genome_sip == false){
							// We *didn't* find the genome. Perhaps the user has labled the
							// plot data by gene sip

							list<GeneAnnotation>::const_iterator target_iter;
							bool found_annot_sip = false;

							if(genome_list.size() == 1){
								target_iter = genome_list.back().find_annot(sip);

								if( target_iter != genome_list.back().end() ){
									found_annot_sip = true;
								}
							}
							else{
						
								list<Genome>::iterator iter;

								for(iter = genome_list.begin();iter != genome_list.end();iter++){
									if(iter->select() == false){
										continue;
									}

									target_iter = iter->find_annot(sip);

									if( target_iter != iter->end() ){
										found_annot_sip = true;
									}
								}
							}

							sip = SeqIdSetFree(sip);

							if(found_annot_sip == false){
								throw "Could not find fasta header in plot file -- tried all selected genomes";
							}
							
							// Make sure that the user has 0 or 1 selected genomes (otherwise, we would have to allow
							// for a single file to contain annotation plots for different genomes).
							if(num_selected_genomes > 1){
								throw "To read annotation based plot data, please select no more than one genome";
							}

							// We have found the seqid in the genome. Find the position of the starting base in 
							// the annotations (we'll use this to adjust the plot indexing
							const unsigned int gene_start = target_iter->start();

							list< pair<unsigned int, float> >::iterator plot_iter = plot_list.begin();
							
							// We have already checked to make sure that the plot list is not empty
							const unsigned int base_offset = plot_iter->first;

							while( plot_iter != plot_list.end() ){
								
								plot_iter->first = (plot_iter->first - base_offset) + gene_start;

								plot_iter++;
							}

							
							// Move all of the data stored in plot_list to gene_plot_list
							gene_plot_list.splice(gene_plot_list.end(), plot_list);

							// The gene_plot_list will be added to this genome at the end of this function
						}
						
						if(sip){
							sip = SeqIdSetFree(sip);
						}
					}
					else{
						// The user did not add a header. Try all selected genomes.
						bool ret = false;
						
						list<Genome>::iterator iter;

						for(iter = genome_list.begin();iter != genome_list.end();iter++){
							if(iter->select() == false){
								continue;
							}

							try{
								iter->set_genome_plot(plot_list);
								ret = true;
							}
							catch(...){
							}
						}
						
						if(ret == false){
							throw "Could not find fasta header in plot file -- tried all selected genomes";
						}
					}

					// Clean up last plot list
					plot_list.clear();
				}

				// Read the SeqId from the current line
				pos = line.find('>');

				if(string::npos == pos){
					throw "Mangled fasta header";
				}
				
				pos ++;
				
				while(isspace(line[pos])){
					pos++;
				}

				line_len = line.size();

				stop = pos;
				
				while( (stop < line_len) && !isspace(line[stop]) ){
					stop ++;
				}

				sip = MakeSeqID( (char*)( line.substr(pos, stop - pos).c_str() ) );
				
				if(sip == NULL){
					throw "Unable to parse SeqId";
				}
				
				if( (sip->choice == SEQID_LOCAL) && (sip->next == NULL) ){
					// Promote this seqid to a genbank accession (to allow more
					// robust comparisons)
					string seq_id_str = line.substr(pos, stop - pos);

					seq_id_str = "gb|" + seq_id_str;
					sip->next = MakeSeqID( (char*)( seq_id_str.c_str() ) );
				}

				break;
			case DATA_LINE:
				// Nothing to do here!
				break;
		};
	}

	// Trim duplicate points
	typedef list< pair<unsigned int, float> >::iterator I;

	list<I> buffer;

	I list_iter;

	for(list_iter = plot_list.begin();list_iter != plot_list.end();list_iter++){
	
		buffer.push_back(list_iter);

		if(buffer.size() == 3){
			// Do we need to delete an entry?
			list<I>::iterator iter0, iter1, iter2;
			
			iter0 = iter2 = buffer.begin();

			iter2++;
			
			iter1 = iter2;

			iter2++;

			if(((*iter0)->second == (*iter1)->second) && ((*iter1)->second == (*iter2)->second)){
				plot_list.erase(*iter1);
				buffer.erase(iter1);
			}
			else{
				buffer.pop_front();
			}
		}
	}

	// Save the last record
	if(sip){
	
		// Save the current data set and flush the list
		// Iterate through the Genome list and find the DNA
		// molecule that matches the current SeqId
		list<Genome>::iterator iter;
		
		bool found_genome_sip = false;

		// Sort the genome plot data
		//plot_list.sort(plot_comp());

		// We may need to load the same genome plot into multiple
		// duplicates of the same genome
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
			if(*iter == sip){
			
				found_genome_sip = true;
				iter->set_genome_plot(plot_list);
			}
		}
		
		// Did we find a genome to add the plot to?
		if(found_genome_sip == false){
		
			// We *didn't* find the genome. Perhaps the user has labled the
			// plot data by gene sip?
			
			list<GeneAnnotation>::const_iterator target_iter;
			bool found_annot_sip = false;
			
			if(genome_list.size() == 1){
			
				target_iter = genome_list.back().find_annot(sip);
				
				if( target_iter != genome_list.back().end() ){
					found_annot_sip = true;
				}
			}
			else{
				
				list<Genome>::iterator iter;
				
				for(iter = genome_list.begin();iter != genome_list.end();iter++){
				
					if(iter->select() == false){
						continue;
					}
					
					target_iter = iter->find_annot(sip);
					
					if( target_iter != iter->end() ){
						found_annot_sip = true;
					}
				}
			}
			
			sip = SeqIdSetFree(sip);
			
			if(found_annot_sip == false){
				throw "Could not find fasta header in plot file -- tried all selected genomes";
			}
			
			// Make sure that the user has 0 or 1 selected genomes (otherwise, we would have to allow
			// for a single file to contain annotation plots for different genomes).
			if(num_selected_genomes > 1){
				throw "To read annotation based plot data, please select no more than one genome";
			}
			
			// We have found the seqid in the genome. Find the position of the starting base in 
			// the annotations (we'll use this to adjust the plot indexing
			const unsigned int gene_start = target_iter->start();
			
			list< pair<unsigned int, float> >::iterator plot_iter = plot_list.begin();
			
			// We have already checked to make sure that the plot list is not empty
			const unsigned int base_offset = plot_iter->first;
			
			while( plot_iter != plot_list.end() ){

				plot_iter->first = (plot_iter->first - base_offset) + gene_start;
				
				plot_iter++;
			}
			
			
			// Move all of the data stored in plot_list to gene_plot_list
			gene_plot_list.splice(gene_plot_list.end(), plot_list);
			
			// Sort the gene_plot
			gene_plot_list.sort();

			if(num_selected_genomes == 0){
				genome_list.back().set_genome_plot(gene_plot_list);
			}
			else{
				list<Genome>::iterator iter;

				for(iter = genome_list.begin();iter != genome_list.end();iter++){
					if(iter->select() == false){
						continue;
					}
					
					iter->set_genome_plot(gene_plot_list);
				}
			}

		}

		// Clean up last plot list
		if(sip){
			sip = SeqIdSetFree(sip);			
		}

		plot_list.clear();
	}
	else{
		// The user did not add a header. Try all selected genomes or, if only a single genome is
		// loaded, the lone genome
		bool ret = false;
		
		list<Genome>::iterator iter;

		for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
			if(iter->select() == false){
				continue;
			}

			try{
				iter->set_genome_plot(plot_list);
				ret = true;
			}
			catch(...){
			}
		}
		
		// Try loading into a lone genome
		if(!ret && genome_list.size() == 1){
		
			try{
				genome_list.back().set_genome_plot(plot_list);
				ret = true;
			}
			catch(...){
			}
		}
		
		if(ret == false){
			throw "Could not find fasta header in plot file -- tried all selected genomes";
		}
	}

	fin.close();
}

int line_type(const string &m_line, list< pair<unsigned int, float> > &m_plot_list, 
			  unsigned int &m_base_index)
{
	
	string::const_iterator iter;
	unsigned int start = 0;

	for(iter = m_line.begin();iter != m_line.end();iter++, start++){
		if(isspace(*iter)){
			continue;
		}

		if(*iter == '>'){
			return SEQID_LINE;
		}
		
		if(*iter == '#'){
			return COMMENT_LINE;
		}
		
		// Compute the base index and plot value
		const unsigned int len = m_line.size();
		unsigned int stop = start;

		list< pair<unsigned int, unsigned int> > input_list;
		
		while(stop < len){
			// Find the length of the value
			while( (stop < len) && !isspace(m_line[stop]) ){
				stop ++;
			}
			
			input_list.push_back( make_pair(start, stop - start) );
			
			// Skip any spaces
			while( (stop < len) && isspace(m_line[stop]) ){
				stop ++;
			}
			
			start = stop;
		}
		
		switch( input_list.size() ){
			case 0:
				throw "No plot data found!";
			case 1:
				// Fall through instead of throwing an error
				// throw "Only one plot value found!";
			case 2:
				if(m_base_index == 0){

					const unsigned int base = atoi( m_line.substr(input_list.front().first, 
						input_list.front().second).c_str() );

					const float value = atof( m_line.substr(input_list.back().first, 
						input_list.back().second).c_str() );
					
					m_plot_list.push_back( make_pair(base, value) );
					
					break;
				}
				
				// If base_index != 0 we fall through to the default, multi-value case
			default:
				{
					list< pair<unsigned int, unsigned int> >::const_iterator iter;
					
					for(iter = input_list.begin();iter != input_list.end();iter++){
						const float value = atof( m_line.substr(iter->first, iter->second).c_str() );

						// Increment the base index as we go
						m_plot_list.push_back( make_pair(m_base_index, value) );

						m_base_index++;
					}
				}
				break;
			
		};
		
		return DATA_LINE;
	}

	return BLANK_LINE;
}

void GenomePage::process_files_to_open()
{

	// Do we have any files to process?
	if(files_to_open.empty()){
		return;
	}

	list<string>::const_iterator iter;
	bool load_error = false;

	for(iter = files_to_open.begin();iter != files_to_open.end();iter++){
		try{
			load(*iter, file_type(*iter));
		}
		catch(...){
			load_error = true;
		}
	}

	// Remove all files from the list!
	files_to_open.clear();

	if(load_error){
		throw "Error opening one or more files!";
	}
}
