// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "genome.h"

using namespace std;


bool Genome::pick(const unsigned int &m_width, 
				  const GLfloat &m_lower_y, const GLfloat &m_upper_y,
				  const unsigned int m_base, const GLfloat &m_y,
				  const FontMetricGL &m_font_metric)
{
	unsigned int obj_count = 0;

	// Count the number of objects that will be rendered into the
	// selection buffer.
	switch(zoom_level){
		case ZOOM_LEVEL_TOP:
			obj_count =  mol.top_level_count();
			break;
		case ZOOM_LEVEL_MED:
		case ZOOM_LEVEL_SEQ:
			obj_count =  mol.mid_level_count();
			break;		
	};

	// Allocate memory for the selection buffer
	// (buffer size is four times the total number of
	// selectable objects).
	int bufSize = 4*obj_count;
	GLuint *selectionBuffer = new GLuint [bufSize];

	if(selectionBuffer == NULL){
		throw "Unable to allocate memory for selection";
	}

  	int hits;

  	glSelectBuffer(bufSize, selectionBuffer);
  	glRenderMode(GL_SELECT);

	glInitNames();
	
	// Stick an element on the stack so we can call glLoadName()
	// later on.
	glPushName(0);
	
	switch(zoom_level){
		case ZOOM_LEVEL_TOP:
			render_top(m_width, m_lower_y, m_upper_y, m_font_metric, true,
				false /* don't clip */, 0 /* dummy clip edge */);
			break;
		case ZOOM_LEVEL_MED:
		case ZOOM_LEVEL_SEQ:
			render_mid(m_width, m_lower_y, m_upper_y, m_font_metric, true,
				false /* don't clip */, 0 /* dummy clip edge */);
			break;
	};
  	
  	hits = glRenderMode(GL_RENDER);
	
	unsigned int annot_id = 0;

  	bool ret = process_hits(annot_id, hits, bufSize, selectionBuffer);

	if(selectionBuffer){
		delete [] selectionBuffer;
	}

	if(ret == false){
		return false;
	}
	
	list<GeneAnnotation>::iterator iter = mol.begin();

	while(annot_id > 0){
		annot_id --;
		iter ++;

		if(iter == mol.end()){
			throw "Woops -- ran off list";
		}
	}

	if(iter == mol.end()){
		throw "Woops -- ran off list";
	}

	// Select (or deselect) this object
	iter->toggle_select();

	return true;
}

bool Genome::auto_pick(const unsigned int &m_width, 
				  const GLfloat &m_lower_y, const GLfloat &m_upper_y,
				  const unsigned int m_base, const GLfloat &m_y,
				  const FontMetricGL &m_font_metric)
{
	unsigned int obj_count = 0;

	// Count the number of objects that will be rendered into the
	// selection buffer.
	switch(zoom_level){
		case ZOOM_LEVEL_TOP:
			obj_count =  mol.top_level_count();
			break;
		case ZOOM_LEVEL_MED:
			obj_count =  mol.mid_level_count();
			break;
		case ZOOM_LEVEL_SEQ:
		
			// Autopicking does not apply to ZOOM_LEVEL_SEQ 
			return false;		
	};

	// Allocate memory for the selection buffer
	// (buffer size is four times the total number of
	// selectable objects).
	int bufSize = 4*obj_count;
	GLuint *selectionBuffer = new GLuint [bufSize];

	if(selectionBuffer == NULL){
		throw "Unable to allocate memory for selection";
	}

  	int hits;

  	glSelectBuffer(bufSize, selectionBuffer);
  	glRenderMode(GL_SELECT);

	glInitNames();
	
	// Stick an element on the stack so we can call glLoadName()
	// later on.
	glPushName(0);
	
	switch(zoom_level){
		case ZOOM_LEVEL_TOP:
			render_top(m_width, m_lower_y, m_upper_y, m_font_metric, true,
				false /* don't clip */, 0 /* dummy clip edge */);
			break;
		case ZOOM_LEVEL_MED:
		case ZOOM_LEVEL_SEQ:
			render_mid(m_width, m_lower_y, m_upper_y, m_font_metric, true,
				false /* don't clip */, 0 /* dummy clip edge */);
			break;
	};
  	
  	hits = glRenderMode(GL_RENDER);
	
	unsigned int annot_id = 0;

  	bool ret = process_hits(annot_id, hits, bufSize, selectionBuffer);

	if(selectionBuffer){
		delete [] selectionBuffer;
	}

	if(ret == false){
		return false;
	}
	
	auto_pick_iter = mol.begin();

	while(annot_id > 0){
	
		annot_id --;
		auto_pick_iter ++;

		if(auto_pick_iter == mol.end()){
			throw "Woops -- ran off list";
		}
	}
	
	return true;
}

// Return {PICKED_NONE, PICKED_GENOME, PICKED_ANNOT} depending on the 
// users selection. If PICKED_ANNOT, or PICKED_GENOME, then return a Pointer 
// to the selected genome
int GenomePage::pick(const unsigned int &m_x, const unsigned int &m_y,
	list<Genome>::iterator &m_pick_iter)
{
	// What did the user pick?
	int ret_value = PICKED_NONE;

	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	const GLfloat dy = 1.0f;
	
	//GLfloat y = ((GLfloat)(m_y))/height;
	GLfloat y = m_y*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	glPushMatrix();
	glLoadIdentity();

	GLint viewport[4];

	viewport[0] = 0;
	viewport[1] = 0;
	viewport[2] = width;
	viewport[3] = height;

	// Use a 1X1 pixel box for picking
	// We can get away with such a small box because we are selecting
	// non-overlapping 2D polygons!
	gluPickMatrix(m_x, m_y, 1.0f, 1.0f, viewport);
	gluOrtho2D(0.0, 2*base_window, y_bottom, y_top);

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!
			ret_value = PICKED_GENOME;

			// Did the user select any genes?
			if( iter->pick(width, yl, yh, m_x, y, font_metric) ){
				ret_value = PICKED_ANNOT;
			}

			glPopMatrix();

			// Return the selected genome
			m_pick_iter = iter;
			return ret_value;
		}

		yl = yh;
		yh += dy;
	}


	glPopMatrix();

	// If we get here, the user did not pick an
	// object
	return ret_value;
}

// Return true is the user picked an annotation
bool GenomePage::auto_pick(const unsigned int &m_x, const unsigned int &m_y)
{
	// What did the user pick?
	int ret_value = PICKED_NONE;

	// Pick the genomes by order in the list
	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		iter->clear_auto_pick();
	}
	
	if(pointer_auto_pick == false){
		
		return ret_value;
	}
	
	const GLfloat dy = 1.0f;
	
	//GLfloat y = ((GLfloat)(m_y))/height;
	GLfloat y = m_y*(y_top - y_bottom)/height + y_bottom;

	GLfloat yl, yh;

	yl = 0.0f;
	yh = dy;

	glPushMatrix();
	glLoadIdentity();

	GLint viewport[4];

	viewport[0] = 0;
	viewport[1] = 0;
	viewport[2] = width;
	viewport[3] = height;

	// Use a 1X1 pixel box for picking
	// We can get away with such a small box because we are selecting
	// non-overlapping 2D polygons!
	gluPickMatrix(m_x, m_y, 1.0f, 1.0f, viewport);
	gluOrtho2D(0.0, 2*base_window, y_bottom, y_top);

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(!iter->show()){
			continue;
		}
		
		if( (y >= yl) && (y <= yh) ){

			// We found a hit!
			ret_value = PICKED_GENOME;

			// Did the user select any genes?
			if( iter->auto_pick(width, yl, yh, m_x, y, font_metric) ){
				ret_value = PICKED_ANNOT;
			}

			glPopMatrix();

			return ret_value;
		}

		yl = yh;
		yh += dy;
	}

	glPopMatrix();

	// If we get here, the user did not pick an
	// object
	return (ret_value == PICKED_ANNOT);
}

// Return the base number of the selected object
bool Genome::process_hits(unsigned int &m_annot_id, const GLint &hits, 
						  const int &bufSize, const GLuint *buffer)
{
	int i;
	GLint names;
	int match = -1;
	GLuint *ptr;
	GLuint minz, z;

	if(hits < 0 ){
		throw "WARNING: select buffer overflow!";
	}

	if(hits == 0){
		return false;
	}

	minz = 0xffffffff;
	ptr = (GLuint *) buffer;
	
	for(i = 0; i < hits; i++){  /* for each hit  */
		names = *ptr;
		ptr++;
		z = *ptr;
		ptr++;
		ptr++;
		
		if(z <= minz){
			minz = z;
			match = *ptr;
		}
		ptr += names;
	}
	
	
	if(match == -1){
		throw "Selection buffer underflow -> match not found!";
	}

	//if(match >= bufSize){
	//	throw "Selection buffer overflow!";
	//}

	m_annot_id = match;

	return true;
}
