// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "genome.h"
#include <math.h>
#include <vector>

using namespace std;

void GenomePage::render_multiple_sequence_conservation()
{
	const unsigned int displayed_genome_count = num_display();
	
	if(displayed_genome_count <= 1){
		// Nothing to do!
		return;
	}
	
	// If 0 or 1 genomes are selected, use all displayed genomes
	bool show_all = (num_select() <= 1);
	
	const unsigned int num_plot_points = (MULTI_PLOT_POINTS > base_window) ?
		2*base_window : 2*MULTI_PLOT_POINTS;

	// We need to render MULTI_PLOT_POINTS
	const unsigned int dx = max( 1, (int)ceil((2.0f*base_window)/num_plot_points) );
	
	// Initialize the entropy to 0 (we will accumulate a sum)
	vector<BaseEntropy> s(num_plot_points);

	// Sum the base composition probability vectors across all selected genomes
	list<Genome>::iterator iter;
	
	unsigned int num_selected_genomes = 0;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
	
		if( show_all || iter->select() ){
		
			num_selected_genomes ++;

			iter->conservation_score(s, dx);
		}
	}
		
	const float dy = 0.98f*(y_top - y_bottom);

	// Draw the graph in the inverse background color
	glColor4f(1.0f - background[0],
			1.0f - background[1],
			1.0f - background[2],
			1.0f);

	glBegin(GL_LINE_STRIP);

	const unsigned int dim = 5; // Dimensionality is 5 -> {A, T, G, C, N}

	// All genomes have the same base a given position
	const float max_score = num_selected_genomes*log( (double)(num_selected_genomes) );

	// All vectors along a unit axis in the dim dimensional space
	// of base composition
	const float min_score = (num_selected_genomes <= dim) ? 0.0f :
		(num_selected_genomes%dim)*( (num_selected_genomes/dim + 1.0f)*log(num_selected_genomes/dim + 1.0f) ) +
		(dim - num_selected_genomes%dim)*( (num_selected_genomes/dim)*log( (double)(num_selected_genomes/dim) ) );
		
	const float score_scale = 1.0f/(max_score - min_score);
	
	for(unsigned int i = 0;i < num_plot_points;i++){

		// The best score is 1, the worst score is 0
		float score = 0.0;
		
		float tmp = s[i].A;
		
		score += (tmp > 0.0f) ? tmp*log(tmp) : 0.0f;
		
		tmp = s[i].T;
		score += (tmp > 0.0f) ? tmp*log(tmp) : 0.0f;
		
		tmp = s[i].G;
		score += (tmp > 0.0f) ? tmp*log(tmp) : 0.0f;
		
		tmp = s[i].C;
		score += (tmp > 0.0f) ? tmp*log(tmp) : 0.0f;
		
		tmp = s[i].N;
		score += (tmp > 0.0f) ? tmp*log(tmp) : 0.0f;

		score = score_scale*(score - min_score);

		glVertex2f(i*dx, y_bottom + dy*score);
	}

	glEnd();
}

void Genome::conservation_score(vector<BaseEntropy> &m_count, const unsigned int &m_dx)
{
	const unsigned int num_points = m_count.size();
	
	// Get the first displayed base (on the left hand side of the screen)
	unsigned int base_start;
	unsigned int base_stop;

	base_range(base_start, base_stop);
	
	DNA3Seq::iterator iter = mol.sequence().begin() + base_start;
	
	unsigned int index = 0;
	//unsigned int i = base_start;
	unsigned int i = 0;
	
	while(index < num_points){
		
		if(i%m_dx == 0){
		
			switch( iter.base() ){
				case DNA3Seq::A:
					m_count[index].A ++;
					break;
				case DNA3Seq::T:
					m_count[index].T ++;
					break;
				case DNA3Seq::G:
					m_count[index].G ++;
					break;
				case DNA3Seq::C:
					m_count[index].C ++;
					break;
				case DNA3Seq::UNKNOWN:
					m_count[index].N ++;
					break;
			};
			
			index++;
		}
		
		iter++;
		i++;
		
		if( iter == mol.sequence().end() ){
			iter = mol.sequence().begin();
		}
	}
	
	// DEBUG 
	if( index != m_count.size() ){
		throw "Genome::conservation_score: index != m_count.size()";
	}
}

#ifdef OLD_VERSION
void Genome::conservation_score(vector<pVector> &m_p, const unsigned int &m_dx)
{

	const unsigned int num_points = m_p.size();

	// Get the first displayed base (on the left hand side of the screen)
	unsigned int base_start;
	unsigned int base_stop;

	base_range(base_start, base_stop);

	for(unsigned int i = 0;i < num_points;i++){

		DNA3Seq::iterator start_iter = mol.sequence().begin() + 
			(base_start + i*m_dx)%num_base;

		DNA3Seq::iterator iter = start_iter;

		pVector tmp;
		unsigned int j;

		for(j = 0;j <= content_window;j++){
			
			switch( iter.base() ){
				case DNA3Seq::A:
					tmp.A ++;
					break;
				case DNA3Seq::T:
					tmp.T ++;
					break;
				case DNA3Seq::G:
					tmp.G ++;
					break;
				case DNA3Seq::C:
					tmp.C ++;
					break;
				case DNA3Seq::UNKNOWN:
					tmp.N ++;
					break;
			};
			
			iter ++;

			if( iter == mol.sequence().end() ){
				iter = mol.sequence().begin();
			}
		}

		if( start_iter == mol.sequence().begin() ){
			start_iter = mol.sequence().end() - 1;
		}
		else{
			start_iter = start_iter - 1;
		}

		iter = start_iter;

		for(j = 1;j <= content_window;j++){
			
			switch( iter.base() ){
				case DNA3Seq::A:
					tmp.A ++;
					break;
				case DNA3Seq::T:
					tmp.T ++;
					break;
				case DNA3Seq::G:
					tmp.G ++;
					break;
				case DNA3Seq::C:
					tmp.C ++;
					break;
				case DNA3Seq::UNKNOWN:
					tmp.N ++;
					break;
			};
			
			if( iter == mol.sequence().begin() ){
				iter = mol.sequence().end() - 1;
			}
			else{
				iter = iter - 1;
			}
		}

		tmp.remap(2*content_window + 1);

		m_p[i] += tmp;
	}
}

#endif // OLD_VERSION
