// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "genome.h"
#include "univout.h"
#include "degenerate_na.h"

#include <sstream>

using namespace std;

// Local functions
string string_to_search_name(const string &m_value);
unsigned int string_to_search_uint(const string &m_value);
list<string> string_to_search_list(const string &m_value);
list<DNA3Seq> string_to_search_seq(const string &m_value);
string remove_space(const string &m_value);

string tolower(string m_str);

list<GeneAnnotation>::iterator Genome::find_annot(const SeqIdPtr &m_sip)
{
	list<GeneAnnotation>::iterator iter;
		
	for(iter = mol.begin();iter != mol.end();iter++){

		if(*iter == m_sip){
			return iter;
		}
	}

	return mol.end();
}

string Genome::search_name(const string &m_name)
{
	// Throw away any previous search results
	search_results.clear();
	cur_search_iter = search_results.end();

	hit_list.clear();
	cur_search_hit_iter = hit_list.end();

	// Keep the user informed
	uout << "Searching " << mol.info(DNAMol::TAXA_NAME) 
		 << "(" << mol.seq_id_str() << ")"
		 << uout_endl;

	search_results = mol.find_name(m_name);

	search_next();

	return search_info();
}

string Genome::search_gi(const unsigned int &m_gi)
{
	// Throw away any previous search results
	search_results.clear();
	cur_search_iter = search_results.end();

	hit_list.clear();
	cur_search_hit_iter = hit_list.end();

	// Keep the user informed
	uout << "Searching " << mol.info(DNAMol::TAXA_NAME) 
		 << "(" << mol.seq_id_str() << ")"
		 << uout_endl;

	search_results = mol.find_gi(m_gi);

	search_next();

	return search_info();
}

string Genome::search_base(const unsigned int &m_base)
{
	// Throw away any previous search results
	search_results.clear();
	cur_search_iter = search_results.end();

	hit_list.clear();
	cur_search_hit_iter = hit_list.end();

	// Keep the user informed
	uout << "Searching " << mol.info(DNAMol::TAXA_NAME) 
		 << "(" << mol.seq_id_str() << ")"
		 << uout_endl;

	search_results = mol.find_base(m_base);

	search_next();

	return search_info();
}

string Genome::search_text(const list<string> &m_text)
{
	// Throw away any previous search results
	search_results.clear();
	cur_search_iter = search_results.end();

	hit_list.clear();
	cur_search_hit_iter = hit_list.end();

	// Keep the user informed
	uout << "Searching " << mol.info(DNAMol::TAXA_NAME) 
		 << "(" << mol.seq_id_str() << ")"
		 << uout_endl;

	search_results = mol.find_text(m_text);

	search_next();

	return search_info();
}

string Genome::search_seq(const list<DNA3Seq> &m_seq)
{
	// Throw away any previous search results
	search_results.clear();
	cur_search_iter = search_results.end();

	hit_list.clear();

	// OSX apears to invalidate the iterator after sorting (bug).
	// So set this iterator after sorting the hit list
	//cur_search_hit_iter = hit_list.end();

	// Keep the user informed
	uout << "Searching " << mol.info(DNAMol::TAXA_NAME) 
		 << "(" << mol.seq_id_str() << ")"
		 << uout_endl;

	search_results = mol.find_seq(m_seq, hit_list);

	hit_list.sort();
	cur_search_hit_iter = hit_list.end();

	search_next();

	return search_info();
}

string Genome::search_primers(const DNA3Seq &m_primer_1, const DNA3Seq &m_primer_2,
		const float &m_salt, const float &m_strand, const float &m_tm, 
		const unsigned int &m_exact_match, const unsigned int &m_min_amplicon_length,
		const unsigned int &m_max_amplicon_length,
		const bool &m_clear_results)
{
	if(m_clear_results){
	
		// Throw away any previous search results
		search_results.clear();
		hit_list.clear();
	}
	
	cur_search_iter = search_results.end();
	
	// OSX apears to invalidate the iterator after sorting (bug).
	// So set this iterator after sorting the hit list
	//cur_search_hit_iter = hit_list.end();

	// Keep the user informed
	uout << "Searching " << mol.info(DNAMol::TAXA_NAME) 
		 << "(" << mol.seq_id_str() << ")"
		 << uout_endl; 

	search_results = mol.find_primers(m_primer_1, m_primer_2,
		m_salt, m_strand, m_tm, m_exact_match, 
		m_min_amplicon_length, m_max_amplicon_length, 
		TM_IDENTITY_CUTOFF, hit_list);

	hit_list.sort();
	cur_search_hit_iter = hit_list.end();

	search_next();

	// Test each of the primers for
	// hair pin, homo-dimer and hetero-dimer formation
	return primer_info();
}

string Genome::search_probe(const DNA3Seq &m_probe,
		const float &m_salt, const float &m_strand, const float &m_tm,
		const bool &m_clear_results)
{
	if(m_clear_results){
	
		// Throw away any previous search results
		search_results.clear();
		hit_list.clear();
	}
	
	cur_search_iter = search_results.end();
	
	// OSX apears to invalidate the iterator after sorting (bug).
	// So set this iterator after sorting the hit list
	//cur_search_hit_iter = hit_list.end();

	// Keep the user informed
	uout << "Searching " << mol.info(DNAMol::TAXA_NAME) 
		 << "(" << mol.seq_id_str() << ")"
		 << uout_endl;

	search_results = mol.find_probe(m_probe, m_salt, m_strand, m_tm, 
		TM_IDENTITY_CUTOFF, hit_list);

	hit_list.sort();
	cur_search_hit_iter = hit_list.end();

	search_next();

	return probe_info();
}

string Genome::search_MOL_primers(const DNA3Seq &m_primer_1, const DNA3Seq &m_primer_2,
		const float &m_salt, const float &m_strand, const float &m_tm, 
		const unsigned int &m_exact_match)
{
	// Throw away any previous search results
	search_results.clear();
	cur_search_iter = search_results.end();

	hit_list.clear();
	
	// OSX apears to invalidate the iterator after sorting (bug).
	// So set this iterator after sorting the hit list
	//cur_search_hit_iter = hit_list.end();

	// Keep the user informed
	uout << "Searching " << mol.info(DNAMol::TAXA_NAME) 
		 << "(" << mol.seq_id_str() << ")"
		 << uout_endl; 

	search_results = mol.find_MOL_primers(m_primer_1, m_primer_2,
		m_salt, m_strand, m_tm, m_exact_match, 
		TM_IDENTITY_CUTOFF, hit_list);

	hit_list.sort();
	cur_search_hit_iter = hit_list.end();

	search_next();

	// Test each of the primers for
	// hair pin, homo-dimer and hetero-dimer formation
	return MOL_primer_info();
}


void Genome::search_next()
{
	// Clear any selected annotations
	mol.clear_select();

	if( ( (search_results.size() <= 1) || (zoom_level == ZOOM_LEVEL_SEQ) ) && !hit_list.empty()){
		// Move the display to center of the current sequence fragment
		
		// Either initialize the search or advance to next element
		if(cur_search_hit_iter == hit_list.end()){
			// Initialize the search
			cur_search_hit_iter = hit_list.begin();
		}
		else{
			// Advance to next element
			cur_search_hit_iter++;

			if(cur_search_hit_iter == hit_list.end()){
				cur_search_hit_iter = hit_list.begin();
			}
		}

		// All annotations that touch the selected fragment
		//
		// Note that this comparison does not take into account perodic
		// boundary conditions -- the comparison should be handled by a 
		// member function! Fix this!!!

		list<GeneAnnotation>::iterator iter;
		
		for(iter = mol.begin();iter != mol.end();iter++){
			if(overlap(iter->start(), iter->stop(),
					   cur_search_hit_iter->start(), cur_search_hit_iter->stop(),
					   num_base) == true){
				// This annotation overlaps the sequence fragment
				// Note that we don't have to de-select fragments that don't overlap
				// thanks to the premptive call to mol.clear_select() at the start of
				// this function.
				iter->select(true);
			}
		}
		
		// Set the base center to be the center of the sequence fragment
		base_center = center(cur_search_hit_iter->start(), cur_search_hit_iter->stop(), num_base);
	}
	else{
		
		if(search_results.empty()){
			return;
		}

		// Either initialize the search or advance to next element
		if(cur_search_iter == search_results.end()){
			// Initialize the search
			cur_search_iter = search_results.begin();
		}
		else{
			// Advance to next element
			cur_search_iter++;

			if(cur_search_iter == search_results.end()){
				cur_search_iter = search_results.begin();
			}
		}

		// Select the current annotations
		(*cur_search_iter)->select(true);

		// Move the display to center of the current annotation
		base_center = center((*cur_search_iter)->start(), (*cur_search_iter)->stop(), num_base);
	}
}

string Genome::search_info()
{
	// Genenome name [start_base, stop_base]
	stringstream info;

	info << mol.info(DNAMol::TAXA_NAME) << "(" << mol.seq_id_str() << ")" << ENDL;

	if(search_results.empty()){
		info << "** No matches found **" << ENDL;
	}
	else{
		list< list<GeneAnnotation>::iterator >::iterator iter;

		for(iter = search_results.begin();iter != search_results.end();iter++){
			info << make_pair(seq_size(), *(*iter) ) << ENDL;
		}
	}

	info << "-----------------------------------------------------------------------" 
		 << "-----------------------------------------------------------------------" 
		 << ENDL;

	return info.str();
}

string Genome::MOL_primer_info()
{
	// Genenome name [start_base, stop_base]
	stringstream info;

	info << mol.info(DNAMol::TAXA_NAME) << "(" << mol.seq_id_str() << ")" << ENDL;

	if(search_results.empty()){
		info << "** No MOL PCR amplicons found **" << ENDL;
	}
	else{
		list<SeqHit>::iterator iter;

		for(iter = hit_list.begin();iter != hit_list.end();iter++){
			const unsigned int amp_len = wrap((int)iter->stop() - (int)iter->start() + 1,
				seq_size() );

			info << "From " << iter->start() << " to " << iter->stop() 
				 << "; amplicon = " << amp_len << " bases" << ENDL
				 << "Primer " << iter->primer_id_1() 
				 << " Tm = " << iter->primer_tm_1() 
				 << " C, 3' match = " << iter->primer_clamp_1() << " bases;" 
				 << "exact matches = " << iter->primer_match_1() << "/" 
				 << iter->primer_len_1()
				 << ENDL
				 << "Primer " << iter->primer_id_2() 
				 << " Tm = " << iter->primer_tm_2() 
				 << " C, 5' match = " << iter->primer_clamp_2() << " bases;" 
				 << "exact matches = " << iter->primer_match_2() << "/" 
				 << iter->primer_len_2()
				 << ENDL;
		}
	}

	info << "-----------------------------------------------------------------------" 
		 << "-----------------------------------------------------------------------" 
		 << ENDL;

	return info.str();
}

string Genome::primer_info()
{
	// Genenome name [start_base, stop_base]
	stringstream info;

	info << mol.info(DNAMol::TAXA_NAME) << "(" << mol.seq_id_str() << ")" << ENDL;

	if(search_results.empty()){
		info << "** No PCR amplicons found **" << ENDL;
	}
	else{
		list<SeqHit>::iterator iter;

		for(iter = hit_list.begin();iter != hit_list.end();iter++){
			const unsigned int amp_len = wrap((int)iter->stop() - (int)iter->start() + 1,
				seq_size() );

			info << "From " << iter->start() << " to " << iter->stop() 
				 << "; amplicon = " << amp_len << " bases" << ENDL
				 << "Primer " << iter->primer_id_1() 
				 << ": Tm = " << iter->primer_tm_1() 
				 << " C; 3' match = " << iter->primer_clamp_1() << " bases; " 
				 << "exact matches = " << iter->primer_match_1() << "/" 
				 << iter->primer_len_1()
				 << ENDL
				 << "Primer " << iter->primer_id_2() 
				 << ": Tm = " << iter->primer_tm_2() 
				 << " C; 3' match = " << iter->primer_clamp_2() << " bases; " 
				 << "exact matches = " << iter->primer_match_2() << "/" 
				 << iter->primer_len_2()
				 << ENDL;
		}
	}

	info << "-----------------------------------------------------------------------" 
		 << "-----------------------------------------------------------------------" 
		 << ENDL;

	return info.str();
}

string Genome::probe_info()
{
	// Genenome name [start_base, stop_base]
	stringstream info;

	info << mol.info(DNAMol::TAXA_NAME) << "(" << mol.seq_id_str() << ")" << ENDL;

	if(search_results.empty()){
		info << "** No binding partners found **" << ENDL;
	}
	else{
		list<SeqHit>::iterator iter;

		for(iter = hit_list.begin();iter != hit_list.end();iter++){
			info << "From " << iter->start() << " to " << iter->stop() 
				 << "; Tm = " << iter->probe_tm() 
				 << "; exact matches = " << iter->probe_match() << "/" << iter->probe_len()
				 << ENDL;
		}
	}

	info << "-----------------------------------------------------------------------" 
		 << "-----------------------------------------------------------------------" 
		 << ENDL;

	return info.str();
}


bool GenomePage::validate_search_key(const unsigned int m_key, const string &m_value)
{
	// Add validation code here!

	return true;
}

string GenomePage::search(const unsigned int m_key, const string &m_value)
{
	string search_info;

	// Search each genome in order
	list<Genome>::iterator iter;

	// Translate the search string into the appropriate search object
	unsigned int search_uint;
	string search_str;
	list<string> search_str_list;
	list<DNA3Seq> search_seq;
	
	uout << "Starting search ..." << uout_endl;

	switch(m_key){
		case SEARCH_KEY_NAME:
			// value type is string
			search_str = string_to_search_name(m_value);
			break;
		case SEARCH_KEY_GI:
		case SEARCH_KEY_BASE:
			// value type is unsigned int
			search_uint = string_to_search_uint(m_value);
			break;
		case SEARCH_KEY_TEXT:
		case SEARCH_KEY_GENOME_NAME:
			// value type is list<string>
			search_str_list = string_to_search_list(m_value);
			break;
		case SEARCH_KEY_DNA:
			// value type is DNA3Seq
			search_seq = string_to_search_seq(m_value);
			break;
	};

	// Select all of the genomes that have search results
	// (so we must clear the list of previously selected genomes)
	clear_selection();

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		
		if(m_key == SEARCH_KEY_GENOME_NAME){
		
			// Does the name of this genome match the list of search words?
			string genome_name = iter->taxa_info();		
			
			for(string::iterator j = genome_name.begin();j != genome_name.end();j++){
				*j = tolower(*j);
			}

			list<string>::const_iterator key_iter;

			for(key_iter = search_str_list.begin();key_iter != search_str_list.end();key_iter++){
			
				if(genome_name.find(*key_iter) == string::npos){
					break;
				}
			}

			if( key_iter == search_str_list.end() ){
			
				// If we get here we found a match to all of the keywords
				if( iter->show() ){
				
					iter->select(true);	
					search_info += "Found Genome: " + iter->taxa_name() + EOL;
				}
			}
			
			continue;
		}
		
		// Note that we are searching hidden genomes as
		// well as visible ones (this is to keep search results in sync).

		// Concatinate all search results into search_info
		switch(m_key){
			case SEARCH_KEY_NAME:
				// value type is string
				search_info += iter->search_name(search_str);
				break;
			case SEARCH_KEY_GI:
				search_info += iter->search_gi(search_uint);
				break;
			case SEARCH_KEY_BASE:
				// value type is int
				search_info += iter->search_base(search_uint);
				break;
			case SEARCH_KEY_TEXT:
				// value type is list<string>
				search_info += iter->search_text(search_str_list);
				break;
			case SEARCH_KEY_DNA:
				// value type is DNA3Seq
				search_info += iter->search_seq(search_seq);
				break;
		};

		// Only select shown genomes that have search results
		if(iter->show() && iter->num_search_results() > 0){
			iter->select(true);
		}
	}

	uout << "Arranging matches ..." << uout_endl;

	// Move all selected genomes to the bottom of the list
	shift_selection_to_bottom();

	uout << "Finished searching" << uout_endl;

	return search_info;
}

void GenomePage::search_next()
{
	list<Genome>::iterator iter;

	for(iter = genome_list.begin();iter != genome_list.end();iter++){
		if(!iter->show()){
			continue;
		}

		iter->search_next();
	}
}

string GenomePage::search_primers(const string &m_primer_1, const string &m_primer_2,
		const float &m_salt, const float &m_strand, const float &m_tm, 
		const unsigned int &m_exact_match, const int &m_min_amplicon_length,
		const int &m_max_amplicon_length)
{
	string search_info;

	// Search each genome in order
	list<Genome>::iterator iter;

	// Translate the search string into DNA sequence
	list<DNA3Seq> primer1;
	list<DNA3Seq> primer2;

	list<DNA3Seq>::const_iterator iter1;
	list<DNA3Seq>::const_iterator iter2;
	
	// Test all of the input parameters
	if(m_salt <= 0.0f){
		throw "[Na+] must be greater than 0";
	}
	
	if(m_strand <= 0.0f){
		throw "[Strand] must be greater than 0";
	}

	if(m_tm <= 0.0f){
		throw "Tm threshold be greater than 0";
	}

	if(m_exact_match > HYBRID_MAX_EXACT_MATCH){
		throw "Number of exact match bases is too large!";
	}
	
	if(m_max_amplicon_length < 0){
		throw "Max amplicon length < 0";
	}

	if(m_min_amplicon_length > m_max_amplicon_length){
		throw "Min amplicon length > Max amplicon length!";
	}

	// Select all of the genomes that have search results
	// (so we must clear the list of previously selected genomes)
	clear_selection();

	uout << "Starting primer search ..." << uout_endl;
	
	primer1 = expand_nucleic_acid( remove_space(m_primer_1) );
	primer2 = expand_nucleic_acid( remove_space(m_primer_2) );
	
	// Are there degenerate bases?
	bool degenerate_bases = (primer1.size() > 1) || (primer2.size() > 1);

	bool multiple_primers = false;
	
	for(iter1 = primer1.begin();iter1 != primer1.end();iter1++){
		for(iter2 = primer2.begin();iter2 != primer2.end();iter2++){

			if(!iter1->is_ATGC()){
				throw "Non-{A,T,G,C} base encountered in primer 1";
			}

			if(!iter2->is_ATGC()){
				throw "Non-{A,T,G,C} base encountered in primer 2";
			}

			// Enforce length constraints on PCR primers
			if(iter1->size() < HYBRID_MIN_PRIMER_LEN){
				throw "primer 1 length is less than the minimum allowed primer length";
			}
			
			if(iter2->size() < HYBRID_MIN_PRIMER_LEN){
				throw "primer 2 length is less than the minimum allowed primer length";
			}

			if(iter1->size() > HYBRID_MAX_PRIMER_LEN){
				throw "primer 1 length is greater than the maximum allowed primer length";
			}
			
			if(iter2->size() > HYBRID_MAX_PRIMER_LEN){
				throw "primer 2 length is greater than the maximum allowed primer length";
			}

			search_info += primer_test(iter1->str(), iter2->str(),
				m_salt, m_strand, m_tm, degenerate_bases);

			for(iter = genome_list.begin();iter != genome_list.end();iter++){
				
				// Note that we are searching hiddened genomes as
				// well as visible ones (this is to keep search results in sync).
				
				search_info += iter->search_primers(*iter1, *iter2, m_salt, m_strand, 
					m_tm, m_exact_match, m_min_amplicon_length, m_max_amplicon_length,
					!multiple_primers /*Only clear previous results for the first primers */);

				// Only select shown genomes that have search results
				if(iter->show() && iter->num_search_results() > 0){
					iter->select(true);
				}
			}
			
			// If we loop around again, we have multiple primers
			multiple_primers = true;
		}
	}

	uout << "Arranging matches ..." << uout_endl;

	// Move all selected genomes to the bottom of the list
	shift_selection_to_bottom();

	uout << "Finished searching" << uout_endl;

	return search_info;
}

string GenomePage::search_probe(const std::string &m_probe, const float &m_salt, const float &m_strand,
		const float &m_tm)
{
	string search_info;

	// Search each genome in order
	list<Genome>::iterator iter;

	// Translate the search string into DNA sequence
	list<DNA3Seq> probe;
	
	list<DNA3Seq>::const_iterator probe_iter;

	// Test all of the input parameters
	if(m_salt <= 0.0f){
		throw "[Na+] must be greater than 0";
	}
	
	if(m_strand <= 0.0f){
		throw "[Strand] must be greater than 0";
	}

	if(m_tm <= 0.0f){
		throw "Tm threshold be greater than 0";
	}

	// Select all of the genomes that have search results
	// (so we must clear the list of previously selected genomes)
	clear_selection();

	uout << "Starting probe search ..." << uout_endl;

	probe = expand_nucleic_acid( remove_space(m_probe) );

	bool multiple_probes = false;
	
	for(probe_iter = probe.begin();probe_iter != probe.end();probe_iter++){

		// Probe length constraints
		if(probe_iter->size() < HYBRID_MIN_PROBE_LEN){
			throw "The probe length is less than the minimum allowed probe length";
		}

		if(probe_iter->size() > HYBRID_MAX_PROBE_LEN){
			throw "The probe length is greater than the maximum allowed probe length";
		}

		
		for(iter = genome_list.begin();iter != genome_list.end();iter++){
			
			// Note that we are searching hiddened genomes as
			// well as visible ones (this is to keep search results in sync).
			
			search_info += iter->search_probe(*probe_iter, m_salt, m_strand, m_tm, 
				!multiple_probes /*Only clear previous results for the first primers */);

			// Only select shown genomes that have search results
			if(iter->show() && iter->num_search_results() > 0){
				iter->select(true);
			}
		}
		
		// If we loop around again, we have multiple probes
		multiple_probes = true;
	}

	uout << "Arranging matches ..." << uout_endl;

	// Move all selected genomes to the bottom of the list
	shift_selection_to_bottom();

	uout << "Finished searching" << uout_endl;

	return search_info;
}

string GenomePage::search_MOL_primers(const string &m_primer_1, const string &m_primer_2,
		const float &m_salt, const float &m_strand, const float &m_tm, 
		const unsigned int &m_exact_match)
{
	string search_info;

	// Search each genome in order
	list<Genome>::iterator iter;

	// Translate the search string into DNA sequence
	list<DNA3Seq> primer1;
	list<DNA3Seq> primer2;

	list<DNA3Seq>::const_iterator iter1;
	list<DNA3Seq>::const_iterator iter2;
	
	// Test all of the input parameters
	if(m_salt <= 0.0f){
		throw "[Na+] must be greater than 0";
	}
	
	if(m_strand <= 0.0f){
		throw "[Strand] must be greater than 0";
	}

	if(m_tm <= 0.0f){
		throw "Tm threshold be greater than 0";
	}

	if(m_exact_match > HYBRID_MAX_EXACT_MATCH){
		throw "Number of exact match bases is too large!";
	}

	// Select all of the genomes that have search results
	// (so we must clear the list of previously selected genomes)
	clear_selection();

	uout << "Starting MOL primer search ..." << uout_endl;
	
	primer1 = expand_nucleic_acid( remove_space(m_primer_1) );
	primer2 = expand_nucleic_acid( remove_space(m_primer_2) );
	
	// Are there degenerate bases?
	bool degenerate_bases = (primer1.size() > 1) || (primer2.size() > 1);

	for(iter1 = primer1.begin();iter1 != primer1.end();iter1++){
		for(iter2 = primer2.begin();iter2 != primer2.end();iter2++){

			if(!iter1->is_ATGC()){
				throw "Non-{A,T,G,C} base encountered in primer 1";
			}

			if(!iter2->is_ATGC()){
				throw "Non-{A,T,G,C} base encountered in primer 2";
			}

			// Enforce length constraints on PCR primers
			if(iter1->size() < HYBRID_MIN_PRIMER_LEN){
				throw "primer 1 length is less than the minimum allowed primer length";
			}
			
			if(iter2->size() < HYBRID_MIN_PRIMER_LEN){
				throw "primer 2 length is less than the minimum allowed primer length";
			}

			if(iter1->size() > HYBRID_MAX_PRIMER_LEN){
				throw "primer 1 length is greater than the maximum allowed primer length";
			}
			
			if(iter2->size() > HYBRID_MAX_PRIMER_LEN){
				throw "primer 2 length is greater than the maximum allowed primer length";
			}

			search_info += primer_test(iter1->str(), iter2->str(),
				m_salt, m_strand, m_tm, degenerate_bases);

			for(iter = genome_list.begin();iter != genome_list.end();iter++){
				
				// Note that we are searching hiddened genomes as
				// well as visible ones (this is to keep search results in sync).
				
				search_info += iter->search_MOL_primers(*iter1, *iter2, m_salt, m_strand, 
					m_tm, m_exact_match);

				// Only select shown genomes that have search results
				if(iter->show() && iter->num_search_results() > 0){
					iter->select(true);
				}
			}
		}
	}

	uout << "Arranging matches ..." << uout_endl;

	// Move all selected genomes to the bottom of the list
	shift_selection_to_bottom();

	uout << "Finished searching" << uout_endl;

	return search_info;
}

string string_to_search_name(const string &m_value)
{
	stringstream key;

	// Remove all whitespace from the string
	// (white space that occurs in the middle of other text
	// is user error). Make all characters lower case.
	for(unsigned int i = 0;i < m_value.size();i++){
		if(!isspace(m_value[i])){
			key << (char)tolower(m_value[i]);
		}
	}

	return key.str();
}

unsigned int string_to_search_uint(const string &m_value)
{
	int key = atoi(m_value.c_str());

	if(key < 0){
		throw "Negative integers not allowed";
	}

	return (unsigned int)key;

}

// Split a space delimited string into a list of order independant
// search words
list<string> string_to_search_list(const string &m_value)
{
	list<string> key;
	
	int start = 0;
	int stop = 0;

	for(unsigned int i = 0;i < m_value.size();i++){
		if(isspace(m_value[i])){
			if(start != stop){
				// Save in lower case
				key.push_back(tolower(m_value.substr(start, stop - start)));
			}

			start = stop = i + 1;
		}
		else{
			stop++;
		}
	}
	
	if(start != stop){
		key.push_back(tolower(m_value.substr(start, stop - start)));
	}

	return key;
}

list<DNA3Seq> string_to_search_seq(const string &m_value)
{
	// Remove any white space characters from the search key
	stringstream tmp_value;
	list<DNA3Seq> key;
	bool complement = false;
	bool save_seq = false;

	for(unsigned int i = 0;i < m_value.size();i++){
		// Skip all spaces and concatinate sequences
		if( isspace(m_value[i]) ){
			continue;
		}
		
		if((m_value[i] == '+') || (m_value[i] == '-')){
			save_seq = true;
		}

		if(save_seq){
			// Leading wildcards are discarded, along with empty strings
			if(tmp_value.str().size() != 0){
				// There is sequence data to save!
				DNA3Seq tmp_seq;

				tmp_seq = tmp_value.str();
				
				if(!tmp_seq.is_ATGC()){
					throw "Non-{A,T,G,C} base encountered in search sequence";
				}

				if(complement){
					key.push_back(tmp_seq.complement());
				}
				else{
					key.push_back(tmp_seq);
				}

				// Clear the string buffer
				tmp_value.str("");
			}

			save_seq = false;
			
			complement = (m_value[i] == '-');

			continue;
		}

		tmp_value << (char)m_value[i];
	}

	if(tmp_value.str().size() == 0){
		return key;
	}
			
	DNA3Seq tmp_seq;

	tmp_seq = tmp_value.str();
			
	if(!tmp_seq.is_ATGC()){
		throw "Non-{A,T,G,C} base encountered in search sequence";
	}

	if(complement){
		key.push_back(tmp_seq.complement());
	}
	else{
		key.push_back(tmp_seq);
	}

	return key;
}

// Remove any white space from the input string
string remove_space(const string &m_value)
{
	// Remove any white space characters from the search key
	stringstream tmp_value;
	const unsigned int len = m_value.size();

	for(unsigned int i = 0;i < len;i++){
		// Skip all spaces and concatinate sequences
		if( isspace(m_value[i]) ){
			continue;
		}

		tmp_value << (char)m_value[i];
	}

	return tmp_value.str();
}

string tolower(string m_str)
{
	for(unsigned int i = 0;i < m_str.size();i++){
		m_str[i] = tolower(m_str[i]);
	}

	return m_str;
}
