// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include "genome.h"
#include "univout.h"
#include <sstream>

using namespace std;

// Local functions
char amino_acid(unsigned char codon[3]);

// Translate the nucleotide sequence [m_start, m_stop]
string Genome::translate_sequence(const long &m_start, const long &m_stop)
{
	stringstream output;
	unsigned int start, stop;

	#ifdef _DEBUG
	if((m_start < 0) || (m_stop < 0)){
		throw "Genome::translate_sequence: start / stop out of bounds";
	}
	#endif // _DEBUG

	// How should we translate this sequence?
	// 1) The start and stop are contained within one or more genes/CDS
	//		- translate the sequence in each reading frame
	// 2) The start and stop span multiple annotations (without being contained in
	//    at least one gene/CDS) or are contained in intergenic space.
	//		- translate in all six reading frames
	
	list<GeneAnnotation>::iterator annot_iter;
	bool found_match = false;

	if( wrap(m_stop - m_start, num_base) > wrap(m_start - m_stop, num_base)){
		stop = (unsigned int)m_start;
		start = (unsigned int)m_stop;
	}
	else{
		start = (unsigned int)m_start;
		stop = (unsigned int)m_stop;
	}

	for(annot_iter = mol.begin();annot_iter != mol.end();annot_iter++){
		
		// Is this a protein coding annotation?
		if(!annot_iter->is_protein()){
			continue;
		}

		// Is the selected start and stop contained within this annotation?
		if(overlap(start, annot_iter->start(), annot_iter->stop(), num_base) == false){
			// The start site is not in this annotation, so keep going ...
			continue;
		}
		
		if(overlap(stop, annot_iter->start(), annot_iter->stop(), num_base) == false){
			// The stop site is not in this annotation, so keep going ...
			continue;
		}

		found_match = true;

		// Get the sequence
		DNA3Seq sub_seq = annot_iter->subseq(mol.sequence(), start, stop);

		if(annot_iter->is_complement() == true){
			unsigned int offset = (2*( wrap(annot_iter->stop() - stop, num_base)%3))%3;
			unsigned int start_aa = wrap(annot_iter->stop() - stop, num_base)/3;
			unsigned int stop_aa = wrap(annot_iter->stop() - start, num_base)/3;

			output << "> " <<  annot_iter->seq_id_str() 
				   << " dna [" << start << ".." << stop << "] -> aa[" 
				   << start_aa << ".." << stop_aa << "] " 
				   << annot_iter->name() << " "
				   << annot_iter->info(GeneAnnotation::PRODUCT) << ENDL;

			output << dna_to_protein(sub_seq, offset, true /* trim the stop codon */) << ENDL;
		}
		else{
			unsigned int offset = (2*(wrap(start - annot_iter->start(), num_base)%3))%3;
			unsigned int start_aa = wrap(start - annot_iter->start(), num_base)/3;
			unsigned int stop_aa = wrap(stop - annot_iter->start(), num_base)/3;

			output << "> " <<  annot_iter->seq_id_str() 
				   << " dna [" << start << ".." << stop << "] -> aa[" 
				   << start_aa << ".." << stop_aa << "]" << ENDL;

			output << dna_to_protein(sub_seq, offset, true /* trim the stop codon */) << ENDL;
		}
	}

	// Were we able to translate the sequence in an annotated reading frame?
	if(found_match == true){
		return output.str();
	}

	// We were unable to find matching valid reading frames. Scan through
	// all six possible reading frames:

	unsigned int len = 1; // Make the length inclusive of both the start and stop

	// Compute the subsequence length (i.e. the distance between start and stop
	// in periodic coordinates).
	for(long i = (long)start;i != (long)stop;i++){
		len ++;

		// Wrap around
		if(i == (long)num_base - 1){
			i = -1;
		}
	}

	// Make a copy of the sequence. Force the DNA molecule to be circular [even if
	// this is not the case -- i.e. do not use mol.circular()].
	DNA3Seq local_seq = mol.sequence().subseq(start, len, true);

	// Save the complement
	DNA3Seq local_comp = local_seq.complement();

	unsigned int offset;

	for(offset = 0;offset < 3;offset++){
		output << "> forced reading frame "
				   << " dna [" << start << ".." << stop << "] (+) strand, offset = " 
				   << offset << ENDL;

			output << dna_to_protein(local_seq, offset, false /* don't trim the stop codon */) << ENDL;
	}

	for(offset = 0;offset < 3;offset++){
		output << "> forced reading frame "
				   << " dna [" << start << ".." << stop << "] (-) strand, offset = " 
				   << offset << ENDL;

			output << dna_to_protein(local_comp, offset, false /* don't trim the stop codon */) << ENDL;
	}

	return output.str();
}

// If m_trim_stop == true, only include the last amino acid if it does *not* come from
// a stop codon.
string dna_to_protein(DNA3Seq &m_seq, const unsigned int &m_offset, const bool &m_trim_stop)
{
	stringstream output;

	const int num_aa = ((int)m_seq.size() - (int)m_offset)/3;
	int i;

	// We must have at least 3 bases to make an amino acid
	if(num_aa <= 0){
		return "Not enought bases to translate into an amino acid!";
	}

	DNA3Seq::iterator iter = m_seq.begin() + m_offset;

	unsigned char codon[3];
	
	for(i = 0;i < num_aa - 1;i++){
		
		codon[0] = iter.base();
		iter ++;
		
		codon[1] = iter.base();
		iter ++;

		codon[2] = iter.base();
		iter ++;
		
		output << amino_acid(codon);
	}

	// Handle the last base with special care
	codon[0] = iter.base();
	iter ++;

	codon[1] = iter.base();
	iter ++;

	codon[2] = iter.base();
	iter ++;

	const char aa = amino_acid(codon);
	
	if(!m_trim_stop || (aa != '-') ){
		output << aa;
	}
	
	return output.str();

}

char amino_acid(unsigned char codon[3])
{
	const char stop_codon = '-';
	const char unknown_codon = '?';
	
	switch(codon[0]){
		case DNA3Seq::A:
			switch(codon[1]){
				case DNA3Seq::A:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
							return 'N'; // {AAU, AAC} = Asn
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'K'; // {AAA, AAG} = Lys
					};
					break;
				case DNA3Seq::T:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
						case DNA3Seq::A:
							return 'I'; // {AUU, AUC, AUA} = Ile
						case DNA3Seq::G:
							return 'M'; // {AUG} = Met
					};
					break;
				case DNA3Seq::G:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
							return 'S'; // {AGU, AGC} = Ser
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'R'; // {AGA, AGG} = Arg
					};
					break;
				case DNA3Seq::C:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'T'; // {ACU, ACC, ACA, ACG} = Thr
					};
					break;
			};
			break;
		case DNA3Seq::T:
			switch(codon[1]){
				case DNA3Seq::A:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
							return 'Y'; // {UAU, UAC} = Tyr
						case DNA3Seq::A:
						case DNA3Seq::G:
							return stop_codon; // {UAA, UAG} = Stop
					};
					break;
				case DNA3Seq::T:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
							return 'F'; // {UUU, UUC} = Phe
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'L'; // {UUA, UUG} = Leu
					};
					break;
				case DNA3Seq::G:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
							return 'C'; // {UGU, UGC} = Cys
						case DNA3Seq::A:
							return stop_codon;	// UGA = Stop
						case DNA3Seq::G:
							return 'W'; // {UGG} = Trp
					};
					break;
				case DNA3Seq::C:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'S'; // {UCU, UCC, UCA, UCG} = Ser
					};
					break;
			};
			break;
		case DNA3Seq::G:
			switch(codon[1]){
				case DNA3Seq::A:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
							return 'D'; // {GAU, GAC} = Asp
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'E'; // {GAA, GAG} = Glu
					};
					break;
				case DNA3Seq::T:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'V'; // {GUU, GUC, GUA, GUG} = Val
					};
					break;
				case DNA3Seq::G:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'G'; // {GGU, GGC, GGA, GGG} = Gly
					};
					break;
				case DNA3Seq::C:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'A'; // {GCU, GCC, GCA, GCG} = Ala
					};
					break;
			};
			break;
		case DNA3Seq::C:
			switch(codon[1]){
				case DNA3Seq::A:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
							return 'H'; // {CAU, CAC} = His
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'Q'; // {CAA, CAG} = Gln
					};
					break;
				case DNA3Seq::T:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'L'; // {CUU, CUC, CUA, CUG} = Leu
					};
					break;
				case DNA3Seq::G:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'R'; // {CGU, CGC, CGA, CGG} = Arg
					};
					break;
				case DNA3Seq::C:
					switch(codon[2]){
						case DNA3Seq::T:
						case DNA3Seq::C:
						case DNA3Seq::A:
						case DNA3Seq::G:
							return 'P'; // {CCU, CCC, CCA, CCG} = Pro
					};
					break;
			};
			break;
	};
	
	return unknown_codon;
}

DNA3Seq GeneAnnotation::subseq(const DNA3Seq &m_seq, const unsigned int &m_start, const unsigned int &m_stop) const
{
	DNA3Seq sub_seq;
	list<unsigned char> seq_list;

	// Assume that both start and stop are contained within this annotation!
	const unsigned int num_base = m_seq.size();
	const unsigned int last_base = (m_stop + 1)%num_base;

	unsigned int i = m_start;
	DNA3Seq::iterator iter = m_seq.begin() + m_start;

	if(is_segmented()){
		list< pair<unsigned int, unsigned int> >::const_iterator seg_iter;

		for(seg_iter = seg_list.begin();seg_iter != seg_list.end();seg_iter++){

			while((i != last_base) && (i <= seg_iter->second)){

				if(i >= seg_iter->first){
					seq_list.push_back(iter.base());
				}

				i ++;

				// Periodic boundary conditions
				if(i == num_base){
					i = 0;
				}

				iter ++;
				
				// Periodic boundary conditions
				if(iter == m_seq.end()){
					iter = m_seq.begin();
				}
			}
		}

	}
	else{

		while(i != last_base){
			seq_list.push_back(iter.base());

			i ++;

			// Periodic boundary conditions
			if(i == num_base){
				i = 0;
			}

			iter ++;
			
			// Periodic boundary conditions
			if(iter == m_seq.end()){
				iter = m_seq.begin();
			}
		}
	}

	sub_seq = DNA3Seq(seq_list.size());

	for(iter = sub_seq.begin();iter != sub_seq.end();iter++){
		iter.base(seq_list.front());

		seq_list.pop_front();
	}

	if(is_complement()){
		return sub_seq.complement();
	}

	return sub_seq;
}
