// This program was prepared by the Regents of the University of California at 
// Los Alamos National Laboratory (the University) under contract No. W-7405-ENG-36 
// with the U.S. Department of Energy (DOE). All rights in the program are reserved 
// by the DOE and the University.  Permission is granted to the public to copy and 
// use this software without charge, provided that this Notice and any statement of 
// authorship are reproduced on all copies.  Neither the U.S. Government nor the 
// University makes any warranty, express or implied, or assumes any liability or 
// responsibility for the use of this software.

#include "stdafx.h"

#include <sstream>

#include "ncbi_util.h"

#include "ent2api.h"
#include "sequtil.h"

using namespace std;

list<unsigned int> gi_list_from_search_key(const string &m_search_key, const int &m_type)
{
	list<unsigned int> gi_list;
	char *entrez_search = NULL;
	char *entrez_data_base = NULL;
	char *entrez_user_defined = NULL;

	Entrez2BooleanReplyPtr  e2br;
	Entrez2IdListPtr        e2id;
	Entrez2RequestPtr       e2rq;
	Entrez2ReplyPtr         e2ry;
	Int4                    i;
	Int4Ptr                 uids;

	// For now, offer one of a list of choices
	switch(m_type){
		case ENTREZ_GENOME_ACCESSION:
			entrez_search = "ACCN";
			entrez_data_base = "Nucleotide";
			break;
		case ENTREZ_GENOME_GI:
			gi_list.push_back(atoi(m_search_key.c_str()));
			return gi_list;
		case ENTREZ_GENOME_TAXA_NAME:
			entrez_search = "ORGN";
			entrez_data_base = "Genome";
			break;
		case ENTREZ_GENOME_USER_DEFINED:
			entrez_user_defined = (char*)m_search_key.c_str();
			entrez_data_base = "Nucleotide";
			break;
		default:
			// Unknown search key!
			return gi_list;
	};

	e2rq = EntrezCreateBooleanRequest (TRUE, FALSE, entrez_data_base, entrez_user_defined,
				0, 0, NULL, 0, 0);
			
	if (e2rq == NULL){
		return gi_list;
	}
	
	if(entrez_search != NULL){
		EntrezAddToBooleanRequest (e2rq, NULL, 0, entrez_search, (char*)m_search_key.c_str(),
			NULL, 0, 0, NULL, NULL, TRUE, TRUE);
	}

	e2ry = EntrezSynchronousQuery (e2rq);

	e2rq = Entrez2RequestFree (e2rq);

	if (e2ry == NULL){
		return gi_list;
	}

	e2br = EntrezExtractBooleanReply (e2ry);

	if (e2br == NULL){
		return gi_list;
	}

	if (e2br->count > 0) {
		e2id = e2br->uids;


		if (e2id != NULL && e2id->num > 0 && e2id->uids != NULL) {
			
			uids = (Int4Ptr) BSMerge (e2id->uids, NULL);
			if (uids != NULL) {
				for (i = 0; i < e2id->num; i++) {
					gi_list.push_back((unsigned int)(uids[i]));
				}

				MemFree (uids);
			}
		}
	}


	Entrez2BooleanReplyFree (e2br);

	return gi_list;
}

list< pair<unsigned int, string> > docsum_list_from_search_key(const string &m_search_key, 
															   const int &m_type)
{
	list<unsigned int> gi_list;
	list< pair<unsigned int, string> > doc_sum_list;
	
	// First get a list of uids
	gi_list = gi_list_from_search_key(m_search_key, m_type);

	unsigned int num_uid = gi_list.size();
	int *ptr_uids = new int [num_uid];

	list<unsigned int>::iterator gi_iter;
	unsigned int i;

	for(i = 0, gi_iter = gi_list.begin();gi_iter != gi_list.end();gi_iter++, i++){
		ptr_uids[i] = *gi_iter;
	}

	char *entrez_data_base = NULL;

	// For now, offer one of a list of choices
	switch(m_type){
		case ENTREZ_GENOME_ACCESSION:
			entrez_data_base = "Nucleotide";
			break;
		case ENTREZ_GENOME_GI:
			entrez_data_base = "Nucleotide";
			break;
		case ENTREZ_GENOME_TAXA_NAME:
			entrez_data_base = "Genome";
			break;
		case ENTREZ_GENOME_USER_DEFINED:
			entrez_data_base = "Nucleotide";
			break;
		default:
			entrez_data_base = "Nucleotide";
			break;
	};

	// Then look up the corresponding docsums
	Entrez2DocsumDataPtr  e2ddp;
	Entrez2DocsumListPtr  e2dlp;
	Entrez2DocsumPtr      e2dp;
	Entrez2RequestPtr     e2rq;
	Entrez2ReplyPtr       e2ry;

	e2rq = EntrezCreateDocSumRequest (entrez_data_base, 0, num_uid, ptr_uids, NULL);

	delete [] ptr_uids;

	if(e2rq == NULL){
		return doc_sum_list;
	}

	e2ry = EntrezSynchronousQuery (e2rq);
	e2rq = Entrez2RequestFree (e2rq);

	if(e2ry == NULL){
		return doc_sum_list;
	}

	e2dlp = EntrezExtractDocsumReply (e2ry);
	if(e2dlp == NULL){
		return doc_sum_list;
	}

	for(e2dp = e2dlp->list; e2dp != NULL; e2dp = e2dp->next){
		pair<unsigned int, string> tmp;
		
		// This is our first guess at the correct uid. If the user is
		// querying the "Genome" database, however, the uid will be in 
		// the "Extra" field instead.
		tmp.first = e2dp->uid;


		for(e2ddp = e2dp->docsum_data; e2ddp != NULL; e2ddp = e2ddp->next){
			
			if(StringICmp(e2ddp->field_name, "Title") == 0){
				tmp.second = e2ddp->field_value;
			}

			// Extract the true uid when querying the "Genome" database.
			// There has got to be a better way to do this!
			if((m_type == ENTREZ_GENOME_TAXA_NAME) && 
				StringICmp(e2ddp->field_name, "Extra") == 0){ // Check the "Extra" field, not "Gi"!

				SeqIdPtr sid = SeqIdParse(e2ddp->field_value);
				
				SeqIdPtr tmp_sid = sid;

				while(tmp_sid){
					
					if(tmp_sid->choice == SEQID_GI){
						tmp.first = tmp_sid->data.intvalue;
						break;
					}

					tmp_sid = tmp_sid->next;
				}

				sid = SeqIdSetFree(sid);
			}
		}

		doc_sum_list.push_back(tmp);
  }

	Entrez2DocsumListFree (e2dlp);

	return doc_sum_list;
}
